/* cfua.c : CFitsio Utility Archive ...    -tab 2012 */

#include "nsx.h"
#include "mua.h"
#include "fitsio.h"


/* ----------------------------------------------------------------------
 (cfitsio) Check an error.
*/
/*@@*/
void cfua_error( int status )
{
fits_report_error(stderr,status);
if (status != 0) exit(1);
return;
}


/* ----------------------------------------------------------------------
 Fix NAN and big or large values, set them to -999.
*/
/*@@*/
double cfua_fix_value( double rr )
{
/**/
int ok;
double rro;
/**/
ok = 1;
rro = rr;
if (rr != rr) ok = 0;
if (rr < -1.e+37) ok = 0;
if (rr >  1.e+37) ok = 0;
if (ok == 0) rro = -999.0;
return(rro);
}


/* ----------------------------------------------------------------------
  Given the RA and DEC, find the column and row position within a cfitsio image 
  given the FITS World Pix parameters.   -tab 12jun2012
  Input: FWP      : FITS World Pix parameters.
       : ra,dec   : RA and Dec (decimal degrees).
 Output: col,row  : column and row value of center of pixel
                    (icol = col + 0.5;  irow = row + 0.5 )
*/
/*@@*/
void cfua_world_to_pix( FWPtype FWP, double ra, double dec, double *col, double *row )
{
/**/
int ii;
int status = 0;
double xpix,ypix,xpos,ypos;
/**/
xpos = ra; 
ypos = dec;
ii=fits_world_to_pix( ra, dec, FWP.xrefval, FWP.yrefval, 
                      FWP.xrefpix, FWP.yrefpix, FWP.xinc, FWP.yinc, 
                      FWP.rot, FWP.coordtype, &xpix, &ypix, &status );
cfua_error(status);
*col = xpix - 1.;
*row = ypix - 1.;
return;
}


/* ----------------------------------------------------------------------
  Give the column and row position for a cfitsio image given the
  FITS World Pix parameters and the RA and Dec.   -tab 12jun2012
  Input: FWP      : FITS World Pix parameters.
       : col,row  : column and row value of center of pixel
                    (icol = col + 0.5;  irow = row + 0.5 )
 Output: ra,dec   : RA and Dec (decimal degrees).
*/
/*@@*/
void cfua_pix_to_world( FWPtype FWP, double col, double row, double *ra, double *dec )
{
/**/
int ii;
int status = 0;
double xpix,ypix,xpos,ypos;
/**/
xpix = col + 1.;
ypix = row + 1.;
ii=fits_pix_to_world( xpix, ypix, FWP.xrefval, FWP.yrefval, 
                      FWP.xrefpix, FWP.yrefpix, FWP.xinc, FWP.yinc, 
                      FWP.rot, FWP.coordtype, &xpos, &ypos, &status );
*ra = xpos;
*dec= ypos;
cfua_error(status);
return;
}


/* ----------------------------------------------------------------------
 (cfitsio) Write an integer header card.
*/
/*@@*/
void cfua_inheadset( fitsfile *fptr, char card[], int value, char comment[] )
{
int status = 0;
int ii;
ii = value;
fits_update_key( fptr, TINT, card, &ii, comment, &status );
cfua_error(status);
return;
}
/* ----------------------------------------------------------------------
 (cfitsio) Write a double header card.
*/
/*@@*/
void cfua_fheadset( fitsfile *fptr, char card[], double value, char comment[] )
{
int status = 0;
double rr;
rr = cfua_fix_value( value );
fits_update_key( fptr, TDOUBLE, card, &rr, comment, &status );
cfua_error(status);
return;
}
/* ----------------------------------------------------------------------
 This version writes a long (extended) character string header card.
 If the string is short it uses the regular character string routine.
 (Note that for blank strings the whole card itself will not be written 
  with _longstr routine).
*/
/*@@*/
void cfua_cheadset(fitsfile *fptr, char card[], char value[], char comment[])
{
int status = 0;
if (clc(value) < 20) {
  fits_update_key( fptr, TSTRING, card, value, comment, &status );
  cfua_error(status);
/* int fits_update_key(fitsfile *fptr, int datatype, char *keyname, void *value, char *comment, int *status) */
} else {
  fits_update_key_longstr( fptr, card, value, comment, &status );
  cfua_error(status);
}
return;
}

/* ----------------------------------------------------------------------
 This version ONLY writes short string header cards.
 (cfitsio) Write a character header card.
*/
/*@@*/
void cfua_cheadset_short(fitsfile *fptr,char card[],char value[],char comment[])
{
int status = 0;
fits_update_key( fptr, TSTRING, card, value, comment, &status );
cfua_error(status);
return;
}
/* ----------------------------------------------------------------------
 (cfitsio) Write a character header card.  Value is 1 (T) or 0 (F).
*/
/*@@*/
void cfua_lheadset( fitsfile *fptr, char card[], int value, char comment[] )
{
int status = 0;
int ii;
ii = value;
fits_update_key( fptr, TLOGICAL, card, &ii, comment, &status );
cfua_error(status);
return;
}
/* ----------------------------------------------------------------------
 (cfitsio) See if a FITS card exists.
  Input: fptr  : FITS file pointer.
         card  : FITS header card.
  Returns "1" is card exists, otherwise "0".
*/
/*@@*/
int cfua_card_exist( fitsfile *fptr, char card[] )
{
int status = 0;
char wrd[200];
fits_read_card( fptr, card, wrd, &status );
if (status != 0) return(0);
return(1);
}
/* ----------------------------------------------------------------------
 (cfitsio) Read a whole header card with the given name.
  Input: fptr  : FITS file pointer.
         card  : FITS header card.
         value : whole card string (blank "" if card not found).
*/
/*@@*/
void cfua_cardhead( fitsfile *fptr, char card[], char value[] )
{
int status = 0;
fits_read_card( fptr, card, value, &status );
if (status != 0) strcpy(value,"");
return;
}
/* ----------------------------------------------------------------------
 This version reads a long (extended) character string header card.
 (cfitsio) Read a character header card.
  Input: fptr  : FITS file pointer.
         card  : FITS header card.
         value : card string value (blank "" if card not found).
*/
/*@@*/
void cfua_chead_longstr( fitsfile *fptr, char card[], char value[] )
{
int status = 0;
char comment[200];
char *ss;
fits_read_key_longstr( fptr, card, &ss, comment, &status );
if (status != 0) { strcpy(value,""); } else { strcpy(value,ss); }
free(ss);
if (clc(value) < 0) { value[0]='\0'; }  /* blank strings */
return;
}
/* ----------------------------------------------------------------------
 This version returns comment also.
 This version reads a long (extended) character string header card.
 (cfitsio) Read a character header card.
  Input: fptr  : FITS file pointer.
         card  : FITS header card.
         value : card string value (blank "" if card not found).
*/
/*@@*/
void cfua_chead_longstr2( fitsfile *fptr, char card[], char value[], char comment[] )
{
int status = 0;
char *ss;
fits_read_key_longstr( fptr, card, &ss, comment, &status );
if (status != 0) { strcpy(value,""); } else { strcpy(value,ss); }
free(ss);
if (clc(value) < 0) { value[0]='\0'; }  /* blank strings */
return;
}
/* ----------------------------------------------------------------------
 This just calls cfua_chead_longstr()...
 (cfitsio) Read a character header card.
  Input: fptr  : FITS file pointer.
         card  : FITS header card.
         value : card string value (blank "" if card not found).
*/
/*@@*/
void cfua_chead( fitsfile *fptr, char card[], char value[] )
{
cfua_chead_longstr( fptr, card, value );
return;
}
/* ----------------------------------------------------------------------
 (cfitsio) Read a double header card.
  Input: fptr  : FITS file pointer.
         card  : FITS header card.
 Returns: card value (-1.e+38 if card not found).
*/
/*@@*/
double cfua_fhead( fitsfile *fptr, char card[] )
{
char comment[200];
double rr;
int status = 0;
fits_read_key( fptr, TDOUBLE, card, &rr, comment, &status );
if (status != 0) rr = -1.e+38;
return(rr);
}
/* ----------------------------------------------------------------------
 This version returns the comment also.
 (cfitsio) Read a double header card.
  Input: fptr  : FITS file pointer.
         card  : FITS header card.
 Returns: card value (-1.e+38 if card not found).
*/
/*@@*/
double cfua_fhead2( fitsfile *fptr, char card[], char comment[] )
{
double rr;
int status = 0;
fits_read_key( fptr, TDOUBLE, card, &rr, comment, &status );
if (status != 0) rr = -1.e+38;
return(rr);
}

/* ----------------------------------------------------------------------
 (cfitsio) Read an integer header card.
  Input: fptr  : FITS file pointer.
         card  : FITS header card.
 Returns: card value (-1234567 if card not found).
*/
/*@@*/
int cfua_inhead( fitsfile *fptr, char card[] )
{
char comment[200];
int ii;
int status = 0;
fits_read_key( fptr, TINT, card, &ii, comment, &status );
if (status != 0) ii = -1234567;
return(ii);
}
/* ----------------------------------------------------------------------
 This version returns the comment string also.
 (cfitsio) Read an integer header card.
  Input: fptr  : FITS file pointer.
         card  : FITS header card.
 Returns: card value (-1234567 if card not found).
*/
/*@@*/
int cfua_inhead2( fitsfile *fptr, char card[], char comment[] )
{
int ii;
int status = 0;
fits_read_key( fptr, TINT, card, &ii, comment, &status );
if (status != 0) ii = -1234567;
return(ii);
}
/* ----------------------------------------------------------------------
 (cfitsio) Read an integer header card.
  Input: fptr  : FITS file pointer.
         card  : FITS header card.
 Returns: card value (0 if card not found).
*/
/*@@*/
int cfua_inhead0( fitsfile *fptr, char card[] )
{
char comment[200];
int ii;
int status = 0;
fits_read_key( fptr, TINT, card, &ii, comment, &status );
if (status != 0) ii = 0;
return(ii);
}
/* ----------------------------------------------------------------------
 (cfitsio) Read a logical header card.  Returns 1 (T) or 0 (F).
*/
/*@@*/
int cfua_lhead( fitsfile *fptr, char card[] )
{
char comment[200];
int ii = 0;
int status = 0;
fits_read_key( fptr, TLOGICAL, card, &ii, comment, &status );
return(ii);
}



/* ----------------------------------------------------------------------
  Add a comment to a FITS header with 'LONGSTRN' comments.
     Input: fptr     : FITS file pointer.
*/
/*@@*/
void cfua_add_longstrn_comment( fitsfile *fptr )
{
/**/
char comment[300];
int status = 0;
/**/
cfua_cheadset( fptr, "LONGSTRN", "OGIP 1.0", "The OGIP long string convention may be used.");
strcpy(comment,"This FITS file may contain long string keyword values that are continued over multiple keywords. This convention uses the character at the end   of a string which is then continued on subsequent keywords whose name = 'CONTINUE'.");
fits_write_comment( fptr, comment, &status );  cfua_error(status);
return;
}


/* ----------------------------------------------------------------------
  Add a comment to a "TTYPE" card with the given value.
     Input: fptr     : FITS file pointer.
            tfields  : number of fields (columns) in the binary FITS table.
            value    : value or name of binary FITS table column.
            comment  : comment string to add.
*/
/*@@*/
void cfua_add_comment(fitsfile *fptr, int tfields, char value[], char comment[])
{
/**/
char card[200];
char vv[200];
int ii,jj;
int status = 0;
/**/
ii=1;
while (ii <= tfields) {
  sprintf(card,"TTYPE%d",ii);
  cfua_chead( fptr, card, vv );
  jj=clc(vv);  vv[(jj+1)]='\0';
  if (strcmp(vv,value) == 0) { break; }
  ++ii;
}
if (ii <= tfields) {
  fits_update_key(fptr,TSTRING,card,value,comment,&status);
  cfua_error(status);
}
return;
}


/* ----------------------------------------------------------------------
  Find column number for a named column value.
  Input: name      : name of requested column.
       : tfields   : number of fields (columns).
       : ttype[99] : names of columns.
       : mode      : If "1" print warning message if column not found.
  Returns column number (first column is "1").
  Returns "0" if no column found.
*/
/*@@*/
int cfua_find_cfitsio_table_column( char name[], int tfields, char *ttype[], int mode )
{
/**/
int ii,kk;
/**/
kk=0;
for (ii=0; ii<tfields; ++ii) {
  if (strcmp(name,ttype[ii]) == 0) { kk = ii + 1; }
}
if ((mode == 1)&&(kk == 0)) {
  printf("WARNING: cfua_fctc: Cannot find %s in FITS source table.\n",name);
  return(0);
}
return(kk);
}



/* ----------------------------------------------------------------------
  See if a column name exists.
  Input: name      : name of requested column.
       : tfields   : number of fields (columns).
       : ttype[]   : names of columns.
  Returns "1" if column name found.
  Returns "0" if column not found.
*/
/*@@*/
int cfua_ColumnExist( char name[], int tfields, char *ttype[] )
{
/**/
int ii,kk;
/**/
kk=0;
for (ii=0; ii<tfields; ++ii) {
  if (strcmp(name,ttype[ii]) == 0) { kk = ii + 1; }
}
if (kk == 0) { return(0); }
return(1);
}


/* ----------------------------------------------------------------------
  Read in or write out part or all of a (long) integer column for named column.
  Input: fptr      : pointer for FITS file unit.
       : colnum    : number of requested column.
                     ( use cfua_find_cfitsio_table_column() )
       : tfields   : number of fields (columns).
       : ttype[]   : names of columns.
       : frow      : start reading at this row number (first row is 1).
       : nelem     : number of table entry elements to read in 
                     (all rows in the file is (naxis2) * (elements per row).
       : mode      : 1 = read column.
                     2 = write column.
Output : aa_long   : Long array.
Returns "1" if ok, "0" if no column found.
*/
/*@@*/
int cfua_long_column_partial( fitsfile *fptr, int colnum, int tfields,
                                      char *ttype[], long frow, long nelem,
                                      int mode, long aa_long[] )
{
/**/
int anynull;
long felem,longnull;
int status = 0;
/**/

if((mode < 1)||(mode > 2)){fprintf(stderr,"***ERROR:mclcp %d\n",mode);exit(1);}

/* Set up. */
felem=1;  longnull=0;

if (mode == 1) {                  /* Read column. */
  fits_read_col( fptr, TLONG, colnum, frow, felem, nelem, &longnull,
                 aa_long, &anynull, &status );

} else {                          /* Write column. */
  fits_write_col( fptr, TLONG, colnum, frow, felem, nelem, aa_long, &status );

}
cfua_error(status);

return(1);
}




/* ----------------------------------------------------------------------
  Read in or write out a (long) integer column for the named column.
  Input: fptr      : pointer for FITS file unit.
       : name      : name of requested column.
       : tfields   : number of fields (columns).
       : ttype[]   : names of columns.
       : frow      : start writing with this row.
       : nelem     : number of table entries to write starting with frow.
       : mode      : 1 = read column.
                     2 = write column.
Output : aa_long   : Long array.
Returns "1" if ok, "0" if no column found.
*/
/*@@*/
int cfua_long_column( fitsfile *fptr, char name[], int tfields,
                              char *ttype[], long frow, long nelem, int mode,
                              long aa_long[] )
{
/**/
int kk,anynull;
long felem,longnull;
int status = 0;
/**/
if((mode < 1)||(mode > 2)){fprintf(stderr,"***ERROR:mclc %d\n",mode);exit(1);}
/* Set up. */
felem=1;  longnull=0;
/* Find column number. */
kk = cfua_find_cfitsio_table_column( name, tfields, ttype, 0 );
/* Read or write column if it exists. */
if (kk != 0) {
  if (mode == 1) {
    fits_read_col( fptr, TLONG, kk, frow, felem, nelem, &longnull,
                   aa_long, &anynull, &status );
  } else {
    fits_write_col( fptr, TLONG, kk, frow, felem, nelem, aa_long, &status );
  }
  cfua_error(status);
} else { return(0); }
return(1);
}



/* ----------------------------------------------------------------------
  Read in or write out a character columned for the named column.
  Input: fptr      : pointer for FITS file unit.
       : name      : name of requested column.
       : tfields   : number of fields (columns).
       : ttype[]   : names of columns.
       : frow      : start writing with this row.
       : nelem     : number of table entries to write starting with frow.
       : mode      : 1 = read column.
                     2 = write column.
 In/Out: aa_string : Array of character strings.
Returns "1" if ok, "0" if no column found.
*/
/*@@*/
int cfua_string_column( fitsfile *fptr, char name[], int tfields,
                              char *ttype[], long frow, long nelem, int mode,
                              char *aa_string[] )
{
/**/
int kk,anynull;
long felem,longnull;
int status = 0;
/**/
if((mode < 1)||(mode > 2)){fprintf(stderr,"***ERROR:mclc %d\n",mode);exit(1);}
/* Set up. */
felem=1;  longnull=0;
/* Find column number. */
kk = cfua_find_cfitsio_table_column( name, tfields, ttype, 0 );
/* Read or write column if it exists. */
if (kk != 0) {
  if (mode == 1) {
    fits_read_col( fptr, TSTRING, kk, frow, felem, nelem, &longnull, aa_string, &anynull, &status );
  } else {
    fits_write_col( fptr, TSTRING, kk, frow, felem, nelem, aa_string, &status );
  }
  cfua_error(status);
} else { return(0); }
return(1);
}



/* ----------------------------------------------------------------------
  Read in or write out a unsigned byte (integer) column for the named column.
  Input: fptr      : pointer for FITS file unit.
       : name      : name of requested column.
       : tfields   : number of fields (columns).
       : ttype[]   : names of columns.
       : frow      : start writing with this row.
       : nelem     : number of table entries to write starting with frow.
       : mode      : 1 = read column.
                     2 = write column.
Output : aa_byte   : Byte array.
Returns "1" if ok, "0" if no column found.
*/
/*@@*/
int cfua_byte_column( fitsfile *fptr, char name[], int tfields,
                              char *ttype[], long frow, long nelem, int mode,
                              unsigned char aa_byte[] )
{
/**/
int kk,anynull;
long felem,longnull;
int status = 0;
/**/
if((mode < 1)||(mode > 2)){fprintf(stderr,"***ERROR:mcsc %d\n",mode);exit(1);}
/* Set up. */
felem=1;  longnull=0;
/* Find column number. */
kk = cfua_find_cfitsio_table_column( name, tfields, ttype, 0 );
/* Read or write column if it exists. */
if (kk != 0) {
  if (mode == 1) {
    fits_read_col( fptr, TBYTE, kk, frow, felem, nelem, &longnull, aa_byte, &anynull, &status );
  } else {
    fits_write_col( fptr, TBYTE, kk, frow, felem, nelem, aa_byte, &status );
  }
  cfua_error(status);
} else { return(0); }

return(1);
}



/* ----------------------------------------------------------------------
  Read in or write out a (unsigned short) integer column for the named column.
  Input: fptr      : pointer for FITS file unit.
       : name      : name of requested column.
       : tfields   : number of fields (columns).
       : ttype[]   : names of columns.
       : frow      : start writing with this row.
       : nelem     : number of table entries to write starting with frow.
       : mode      : 1 = read column.
                     2 = write column.
Output : aa_ushort : Unsigned Short array.
Returns "1" if ok, "0" if no column found.
*/
/*@@*/
int cfua_ushort_column( fitsfile *fptr, char name[], int tfields,
                               char *ttype[], long frow, long nelem, int mode,
                               unsigned short int aa_ushort[] )
{
/**/
int kk,anynull;
long felem,longnull;
int status = 0;
/**/
if((mode < 1)||(mode > 2)){fprintf(stderr,"***ERROR:mcsc %d\n",mode);exit(1);}
/* Set up. */
felem=1;  longnull=0;
/* Find column number. */
kk = cfua_find_cfitsio_table_column( name, tfields, ttype, 0 );
/* Read or write column if it exists. */
if (kk != 0) {
  if (mode == 1) {
    fits_read_col( fptr, TUSHORT, kk, frow, felem, nelem, &longnull,
                   aa_ushort, &anynull, &status );
  } else {
    fits_write_col( fptr, TUSHORT, kk, frow, felem, nelem, aa_ushort, &status );
  }
  cfua_error(status);
} else { return(0); }

return(1);
}



/* ----------------------------------------------------------------------
  Read in or write out a (short) integer column for the named column.
  Input: fptr      : pointer for FITS file unit.
       : name      : name of requested column.
       : tfields   : number of fields (columns).
       : ttype[]   : names of columns.
       : frow      : start writing with this row.
       : nelem     : number of table entries to write starting with frow.
       : mode      : 1 = read column.
                     2 = write column.
Output : aa_short  : Short array.
Returns "1" if ok, "0" if no column found.
*/
/*@@*/
int cfua_short_column( fitsfile *fptr, char name[], int tfields,
                               char *ttype[], long frow, long nelem, int mode,
                               short aa_short[] )
{
/**/
int kk,anynull;
long felem,longnull;
int status = 0;
/**/
if((mode < 1)||(mode > 2)){fprintf(stderr,"***ERROR:mcsc %d\n",mode);exit(1);}
/* Set up. */
felem=1;  longnull=0;
/* Find column number. */
kk = cfua_find_cfitsio_table_column( name, tfields, ttype, 0 );
/* Read or write column if it exists. */
if (kk != 0) {
  if (mode == 1) {
    fits_read_col( fptr, TSHORT, kk, frow, felem, nelem, &longnull,
                   aa_short, &anynull, &status );
  } else {
    fits_write_col( fptr, TSHORT, kk, frow, felem, nelem, aa_short, &status );
  }
  cfua_error(status);
} else { return(0); }

return(1);
}

/* ----------------------------------------------------------------------
  Read in or write out a double column for the named column.
  Input: fptr      : pointer for FITS file unit.
       : name      : name of requested column.
       : tfields   : number of fields (columns).
       : ttype[]   : names of columns.
       : frow      : start writing with this row.
       : nelem     : number of table entries to write starting with frow.
       : mode      : 1 = read column.
                     2 = write column.
Output : aa_double : Double array.
Returns "1" if ok, "0" if no column found.
*/
/*@@*/
int cfua_double_column( fitsfile *fptr, char name[], int tfields,
                                char *ttype[], long frow, long nelem, int mode,
                                double aa_double[] )
{
/**/
int kk,anynull;
long felem,longnull;
int status = 0;
/**/
if((mode < 1)||(mode > 2)){fprintf(stderr,"***ERROR:mcdc %d\n",mode);exit(1);}
/* Set up. */
felem=1;  longnull=0;
/* Find column number. */
kk = cfua_find_cfitsio_table_column( name, tfields, ttype, 0 );
/* Read or write column if it exists. */
if (kk != 0) {
  if (mode == 1) {
    fits_read_col( fptr, TDOUBLE, kk, frow, felem, nelem, &longnull,
                   aa_double, &anynull, &status );
  } else {
    fits_write_col( fptr, TDOUBLE, kk, frow, felem, nelem, aa_double, &status );
  }
  cfua_error(status);
} else { return(0); }
return(1);
}

/* ----------------------------------------------------------------------
  Read in or write out a float column for the named column.
  Input: fptr      : pointer for FITS file unit.
       : name      : name of requested column.
       : tfields   : number of fields (columns).
       : ttype[]   : names of columns.
       : frow      : start writing with this row.
       : nelem     : number of table entries to write starting with frow.
       : mode      : 1 = read column.
                     2 = write column.
Output : aa_float  : Float array.
Returns "1" if ok, "0" if no column found.
*/
/*@@*/
int cfua_float_column( fitsfile *fptr, char name[], int tfields,
                               char *ttype[], long frow, long nelem, int mode,
                               float aa_float[] )
{
/**/
int kk,anynull;
long felem,longnull;
int status = 0;
int bad = 0;
/**/
if((mode < 1)||(mode > 2)){fprintf(stderr,"***ERROR:mcfc %d\n",mode);exit(1);}
/* Set up. */
felem=1;  longnull=0;
/* Find column number. */
kk = cfua_find_cfitsio_table_column( name, tfields, ttype, 0 );
/* Read or write column if it exists. */
if (kk != 0) {
  if (mode == 1) {
    fits_read_col( fptr, TFLOAT, kk, frow, felem, nelem, &longnull,
                   aa_float, &anynull, &status );
  } else {
    fits_write_col( fptr, TFLOAT, kk, frow, felem, nelem, aa_float, &status );
  }
  fits_report_error(stderr,status); if (status != 0) return(bad);
} else { return(0); }

return(1);
}



/* ----------------------------------------------------------------------
  Read in FITS columns from a binary FITS table file.

    Input: ffile     : FITS filename.
           ncol      : number of columns to be read.
          *colnames[]: names of the columns to be read.
               (name may be followed by :# to indicate an entry of array)
           srow      : starting row to load in image[].
           hdunum    : number of HDU to read table out of (usually 2).
           rowlimit  : Limit on the number of rows to read in.
           maxcol    : Maximum number of columns in image[].
           maxrow    : Maximum number of rows in image[].
           info      : Show info: info=1 (standard), info=2 (with tunit).
           verbose   : Verbosity.
           maxfield  : maximum number of fields.
           *ttype[]  : TTYPE strings.
           *tform[]  : TFORM strings.

   Output: image[]   : 2-D image containing the rows and columns of data.
           nrow_ret  : return number of rows loaded into image[].
           forms[]   : The form of a column, i.e. I,J,E,D,...

  Allocate ttype and tform like this:
     const int maxfield = 1000;
     char *ttype[maxfield];
     char *tform[maxfield];
     for (ii=0; ii<maxfield; ++ii) { ttype[ii] = (char *)malloc(FLEN_VALUE); }
     for (ii=0; ii<maxfield; ++ii) { tform[ii] = (char *)malloc(FLEN_VALUE); }
*/
/*@@*/
void cfua_read_fits_columns( char ffile[], int ncol, char *colnames[],
                             int srow, int hdunum, int rowlimit, 
                             int maxcol, int maxrow, int info, int verbose, 
                             int maxfield, char *ttype[], char *tform[],
                             double image[], int *nrow_ret, char forms[], 
                             int quiet )
{
/**/
int naxis2,pixno,ii,jj,maxkk,vv,ll,kk,mm,nrow,tfields;
int nelem,maxna,nfound,hdutype;
int status = 0;
/**/
fitsfile *fptr;
/**/
unsigned char *aa_b;
short *aa_s;
long *aa_l;
float *aa_f;
double *aa_d;
/**/
char wrd[200];
char wrd2[200];
char wrd3[200];
char wrd4[200];
char line1[200];
char noc[200];    /* Name Of Column */
char singlequote[9];
/**/
int colnum[600];   /* column number. */
int colna[600];    /* number of entries in column array. */
int colai[600];    /* column array index */
char colform[600]; /* column form letter: I,J,E,D */
/**/

/* Clear. */
aa_b = NULL;
aa_s = NULL;
aa_l = NULL;
aa_f = NULL;
aa_d = NULL;
strcpy(singlequote,"'");

/* Open file. */
fits_open_file( &fptr, ffile, READONLY, &status );
cfua_error( status );
for (ii=1; ii<=hdunum; ++ii) {
  fits_movabs_hdu(fptr, ii, &hdutype, &status); cfua_error(status);
}

/* Check. */
if (hdutype != BINARY_TBL) {
  fprintf(stderr,"***ERROR:1:rfc: Finding binary table.\n");
  exit(1);
}

/* Number of rows. */
naxis2 = cfua_inhead( fptr, "NAXIS2" );
if (rowlimit > 0) { nrow = rowlimit; } else { nrow = naxis2; }
*nrow_ret = nrow;
if (quiet == 0) printf("Found %d rows.\n",nrow);

/* Check. */
if (nrow+srow >= maxrow) { printf("***error: too many rows.\n"); exit(1); }
if (ncol      >= maxcol) { printf("***error: too many columns.\n"); exit(1); }

/* Number of fields (columns). */
tfields = cfua_inhead( fptr, "TFIELDS" );

/* Check. */
if (tfields > maxfield) {
  fprintf(stderr,"***ERROR:rfc: Too many fields (%d), max is %d.\n",tfields,maxfield);
  exit(1);
}


/* Read the column names from the TTYPEn keywords */
fits_read_keys_str(fptr, "TTYPE", 1, tfields, ttype, &nfound, &status);
cfua_error(status);

/* Read the column forms from the TFORMn keywords */
fits_read_keys_str(fptr, "TFORM", 1, tfields, tform, &nfound, &status);
cfua_error(status);


/* Echo. */
if ((info > 0)||(verbose == 1)) {
  if (info != 4) {
    if (quiet == 0) printf("NOTE: %s : srow=%d  nrow=%d  naxis2=%d  tfields=%d  nfound=%d\n",ffile,srow,nrow,naxis2,tfields,nfound);
  }
  if (info > 0) {
    maxkk = 0;
    for (ii=0; ii<nfound; ++ii) { kk = clc(ttype[ii]); if (kk > maxkk) maxkk = kk; }
    for (ii=0; ii<nfound; ++ii) {
      strcpy(wrd3,"");
      if (info > 1) {
        sprintf(wrd,"TUNIT%d",ii+1);
        fits_read_keyword( fptr, wrd, wrd3, line1, &status );
        cfua_error( status );
      }
      sprintf(wrd,"TTYPE%d",ii+1);
      fits_read_keyword( fptr, wrd, wrd2, line1, &status );  /* grab comment */
      cfua_error( status );
      if (info == 2) {
        strcat(wrd3," : ");
        strprefix_wb( line1, wrd3 ); 
      }
      if (info < 3) {
        sprintf(wrd,"TTYPE%d",ii+1);
        jj=clc(wrd);  wrd[jj+1]=' '; wrd[jj+2]=' '; wrd[jj+3]=' '; wrd[jj+4]=' '; wrd[9]='\0';
        sprintf(wrd2,"(%s)",tform[ii]);
        jj=clc(wrd2);  wrd2[jj+1]=' '; wrd2[jj+2]=' '; wrd2[jj+3]=' '; wrd2[jj+4]=' '; wrd2[6]='\0';
        if (maxkk < 10) { printf(" %10s : %s%s: %s\n",ttype[ii],wrd,wrd2,line1); } else {
        if (maxkk < 15) { printf(" %15s : %s%s: %s\n",ttype[ii],wrd,wrd2,line1); } else {
        if (maxkk < 20) { printf(" %20s : %s%s: %s\n",ttype[ii],wrd,wrd2,line1); } else {
        if (maxkk < 25) { printf(" %25s : %s%s: %s\n",ttype[ii],wrd,wrd2,line1); } else {
        if (maxkk < 30) { printf(" %30s : %s%s: %s\n",ttype[ii],wrd,wrd2,line1); } else {
                          printf(" %40s : %s%s: %s\n",ttype[ii],wrd,wrd2,line1);        }}}}}
      }
      if (info == 3) {
        vv=0; 
        for (ll=0; ll<clc(wrd3); ++ll) { 
          if ((wrd3[ll] != ' ')&&(wrd3[ll] != singlequote[0])) { wrd4[vv] = wrd3[ll]; ++vv; }  
        }  
        wrd4[vv]='\0';
        printf("%s	%d	%s	%s	%s\n",ttype[ii],ii+1,tform[ii],wrd4,line1);
      }
      if (info == 4) {
        printf("%s\n",ttype[ii]);
      }
    }
    if (info != 4) {
      if (quiet == 0) printf("NOTE: %s : srow=%d  nrow=%d  naxis2=%d  tfields=%d  nfound=%d\n",ffile,srow,nrow,naxis2,tfields,nfound);
    }
  }
} else {
  if (quiet == 0) printf("NOTE: %s : srow=%d  nrow=%d  naxis2=%d  tfields=%d  nfound=%d\n",ffile,srow,nrow,naxis2,tfields,nfound);
}


/* Any columns? */
if (ncol > 0) {


/* . . . . . . First find column numbers and array lengths. */

/* Read columns. */
  maxna = 1;
  for (ii=0; ii<ncol; ++ii) {

/* Root column name and index. */
    strcpy(noc,colnames[ii]);
    mm = cindex(noc,":");
    if (mm > 0) {
      colai[ii] = cvalread0(noc,mm+1,clc(noc));
      noc[mm] = '\0';
    } else {
      colai[ii] = 1;
    }

/* Find and save column numbers. */
    kk=-1;
    for (jj=0; jj<nfound; ++jj) {
      if (strcmp(ttype[jj],noc) == 0) { kk = jj; }
    }
    if (kk < 0) {
      if (verbose == 1) printf("NOTE: rfc: Cannot find column name=%s\n",noc);
      colnum[ii]  = -9;
      colna[ii]   = 1;
      colform[ii] = 'D';
      forms[ii]   = 'D';
    } else { 
      colnum[ii] = kk;
      strcpy(wrd,tform[kk]);
      mm = clc(wrd);
      colform[ii] = wrd[mm];
      forms[ii] = colform[ii];
      if (mm > 0) {
        wrd[mm] = '\0';
        colna[ii] = GLV(wrd,1);
      } else {
        colna[ii] = 1;
      }
    }

/* Check. */
    if (colai[ii] > colna[ii]) { 
      fprintf(stderr,"***ERROR: rfc: column index to large:  colname=%s  colai=%d  colna=%d\n",colnames[ii],colai[ii],colna[ii]);
      exit(1);
    }

/* Maximum array size. */
    if (colna[ii] > maxna) { maxna = colna[ii]; }

  }

/* Echo. */
  if (maxna != 1) printf("NOTE: Maximum number of elements in a column is %d .\n",maxna);


/* . . . . . . Allocate memory. */
  if (!( aa_b = (unsigned char *)malloc((maxna*nrow)*sizeof(unsigned char)) )) {
    fprintf(stderr,"***ERROR: rfc: Out of memory for aa_b.\n");
    exit(1);
  }
  if (!( aa_s = (short *)malloc((maxna*nrow)*sizeof(short)) )) {
    fprintf(stderr,"***ERROR: rfc: Out of memory for aa_s.\n");
    exit(1);
  }
  if (!( aa_d = (double *)malloc((maxna*nrow)*sizeof(double)) )) {
    fprintf(stderr,"***ERROR: rfc: Out of memory for aa_d.\n");
    exit(1);
  }
  if (!( aa_l = (long *)malloc((maxna*nrow)*sizeof(long)) )) {
    fprintf(stderr,"***ERROR: rfc: Out of memory for aa_l.\n");
    exit(1);
  }
  if (!( aa_f = (float *)malloc((maxna*nrow)*sizeof(float)) )) {
    fprintf(stderr,"***ERROR: rfc: Out of memory for aa_f.\n");
    exit(1);
  }


/* . . . . . . Read data. */


/* Clear image (at and after starting row (srow). */
  for (jj=srow; jj<(nrow+srow); ++jj) {
  for (ii=  0;  ii< ncol      ; ++ii) { 
    pixno = ii + (jj * ncol);  image[pixno]=0.;
  }}

/* Read column data, load each named column. */
  for (ii=0; ii<ncol; ++ii) { if (colnum[ii] != -9) {

/* Check image[] dimensions. */
    pixno = ii + (ncol * (srow + nrow));
    if (pixno > (maxcol * maxrow)) {
      fprintf(stderr,"***ERROR:rfc: image[] not big enough: max=%d, requested=%d\n",maxcol*maxrow,pixno);
      exit(1);
    }

/* Set column number. */
    kk = colnum[ii];
    if (verbose == 1) {
      printf("NOTE: Reading column name=%s,  column number=%d,  form=%s\n",colnames[ii],kk,tform[kk]);
    }
   
/* long. */
    if (colform[ii] == 'J') {
      nelem = nrow * colna[ii];
      if (cfua_long_column(fptr,ttype[kk],tfields,ttype,1,nelem,1,aa_l) == 1) {
        for(jj=srow; jj<(srow+nrow); ++jj){ 
          pixno = ii+(ncol * jj);
          mm = (colai[ii] - 1) + (colna[ii] * (jj - srow));
          image[pixno]= aa_l[ mm ];
        }
      }

    } else {
/* float. */
    if (colform[ii] == 'E') {
      nelem = nrow * colna[ii];
      if (cfua_float_column(fptr,ttype[kk],tfields,ttype,1,nelem,1,aa_f) == 1) {
        for(jj=srow; jj<(srow+nrow); ++jj){
          pixno = ii+(ncol * jj);
          mm = (colai[ii] - 1) + (colna[ii] * (jj - srow));
          image[pixno]= aa_f[ mm ];
        }
      }

    } else {
/* double. */
    if (colform[ii] == 'D') {
      nelem = nrow * colna[ii];
      if (cfua_double_column(fptr,ttype[kk],tfields,ttype,1,nelem,1,aa_d) == 1) {
        for(jj=srow; jj<(srow+nrow); ++jj){
          pixno = ii+(ncol * jj);
          mm = (colai[ii] - 1) + (colna[ii] * (jj - srow));
          image[pixno]= aa_d[ mm ];
        }
      }

    } else {
/* short. */
    if (colform[ii] == 'I') {
      nelem = nrow * colna[ii];
      if (cfua_short_column(fptr,ttype[kk],tfields,ttype,1,nelem,1,aa_s) == 1) {
        for(jj=srow; jj<(srow+nrow); ++jj){
          pixno = ii+(ncol * jj);
          mm = (colai[ii] - 1) + (colna[ii] * (jj - srow));
          image[pixno]= aa_s[ mm ];
        }
      }

    } else {
/* unsigned byte. */
    if (colform[ii] == 'B') {
      nelem = nrow * colna[ii];
      if (cfua_byte_column(fptr,ttype[kk],tfields,ttype,1,nelem,1,aa_b) == 1) {
        for(jj=srow; jj<(srow+nrow); ++jj){
          pixno = ii+(ncol * jj);
          mm = (colai[ii] - 1) + (colna[ii] * (jj - srow));
          image[pixno]= aa_b[ mm ];
        }
      }

    }}}}}

  }}

/* Free. */
  free(aa_b);  aa_b = NULL;
  free(aa_s);  aa_s = NULL;
  free(aa_l);  aa_l = NULL;
  free(aa_f);  aa_f = NULL;
  free(aa_d);  aa_d = NULL;

}

/* Close FITS file. */
fits_close_file( fptr, &status );             cfua_error( status );

return;
}

/* ----------------------------------------------------------------------
  Copy a FITS image clipped with the specified column and row range. -tab 13jun2012
  In this case, the first column and row is 0, so 1 is added to all for values before
  the cfitsio routine call.  In other words, if the sc=100 sr=100, then pixel 100,100
  on the old image will be 0,0 on the new image, and the dimensions will be
  nc=1+ec-sc and nr=1+er-sr .
*/
/*@@*/
void cfua_clip_copy_image( char imgfile[], char imgfile2[], int sc, int ec, int sr, int er )
{
/**/
fitsfile *fptr;
fitsfile *fptr2;
int status = 0;
char wrd[300];
/**/

/*  Open clipped image file. */
sprintf(wrd,"%s[%d:%d,%d:%d]",imgfile,sc+1,ec+1,sr+1,er+1);
printf("reading... '%s'\n",wrd);
fits_open_file( &fptr, wrd, READONLY, &status ); cfua_error( status );

/* Write new image and close files. */
printf("Creating '%s'.\n",imgfile2);
fits_create_file( &fptr2, imgfile2, &status );        cfua_error(status);
fits_copy_file( fptr, fptr2, 1, 1, 1, &status );      cfua_error(status);
fits_close_file( fptr2, &status );                    cfua_error(status);
fits_close_file( fptr, &status );   cfua_error(status);

return;
}



/* ----------------------------------------------------------------------
  Load a FITS image with FITS World Pix parameters.   -tab 12jun2012
*/
/*@@*/
void cfua_load_image( char imgfile[], float image[], FWPtype FWP[] )
{
/**/
fitsfile *fptr;
/**/
long nbuffer;
long firstpixel = 1;
/**/
int anynull,naxis1,naxis2;
int status = 0;
/**/
float nullval = 0;  /* don't check for null values in the image */
/**/
double xrefval,yrefval,xrefpix,yrefpix,xinc,yinc,rot;
/**/
char coordtype[200];
/**/

/*  Open image file. */
fits_open_file( &fptr, imgfile, READONLY, &status ); cfua_error( status );

/* Read header card values. */
naxis1 = cfua_inhead( fptr, "NAXIS1" );
naxis2 = cfua_inhead( fptr, "NAXIS2" );

      /* for reference...
      exptime= cfua_fhead( fptr, "EXPTIME" );
      if (exptime < 1.) exptime = 1.;
      cdelt1 = cfua_fhead( fptr,"CDELT1") ;
      if (cdelt1 < -100000.) cdelt1 = 0.;
      /x Bad value limit. x/
      if (exptime > 10.) { limit= 10. * exptime;  } else { limit= 999000.; }
       .. WCS cards.  for reference only, could be done like this:
      xrefval = cfua_fhead( fptr, "CRVAL1" );
      yrefval = cfua_fhead( fptr, "CRVAL2" );
      xrefpix = cfua_fhead( fptr, "CRPIX1" );
      yrefpix = cfua_fhead( fptr, "CRPIX2" );
      xinc    = cfua_fhead( fptr, "CDELT1" );
      yinc    = cfua_fhead( fptr, "CDELT2" );
      rot     = cfua_fhead( fptr, "CROTA2" );
      strcpy(coordtype,"-TAN");
      */

/* Read in WCS (World Coordinate System) header card values. */
fits_read_img_coord( fptr, &xrefval, &yrefval, &xrefpix, &yrefpix,
                     &xinc, &yinc, &rot, coordtype, &status );
cfua_error( status );

/* Save. */
FWP[0].xrefval = xrefval;
FWP[0].yrefval = yrefval;
FWP[0].xrefpix = xrefpix;
FWP[0].yrefpix = yrefpix;
FWP[0].xinc    = xinc;
FWP[0].yinc    = yinc;
FWP[0].rot     = rot;
FWP[0].ncol    = naxis1;
FWP[0].nrow    = naxis2;
strcpy(FWP[0].coordtype,coordtype);
/* Read in image data. */
nbuffer = naxis1 * naxis2;
fits_read_img( fptr, TFLOAT, firstpixel, nbuffer, &nullval, image, &anynull, &status );
cfua_error(status);

/* Close image FITS file. */
fits_close_file( fptr, &status );
cfua_error(status);

return;
}




