/* galactic.c - various astronomical routines. */

#include "nsx.h"
#include "mua.h"


/* ----------------------------------------------------------------------
  Get Position of Sun.
  Fundamental arguments (Van Flandern & Pulkkinen 1979).
  Converted from FORTRAN. -tab 06aug2012
*/
/*@@*/
void gal_GetSunPosition( double T, double TT, double *RA, double *DEC )
{
/**/
double L,G,V,U,W,S;
const double P2 = 2. * 3.141592654;
/**/
L = 0.779072 + (0.00273790931 * T);
G = 0.993126 + (0.0027377785  * T);
L = L - (double)((int)L);
G = G - (double)((int)G);
L = L * P2;
G = G * P2;
V = 0.39785 * sin(L);
V = V - (0.01000 * sin(L-G));
V = V + (0.00333 * sin(L+G));
V = V - (0.00021 * TT * sin(L));
U = 1.0 - (0.03349 * cos(G));
U = U - (0.00014 * cos(2.*L));
U = U + (0.00008 * cos(L));
W = (-0.00010) - (0.04129 * sin(2.*L));
W = W + (0.03211 * sin(G));
W = W + (0.00104 * sin((2.*L)-G));
W = W - (0.00035 * sin((2.*L)+G));
W = W - (0.00008 * TT * sin(G));
/* Compute Sun's RA and DEC. */
S   = W / sqrt(U - (V * V));
*RA = L + atan(S / sqrt(1.0-(S*S)));
S   = V / sqrt(U);
*DEC= atan(S / sqrt(1.0-(S*S)));
*RA = *RA * 24.0 / P2;
if (*RA < 0.) *RA = *RA + 24.0;
*DEC= *DEC * 360. / P2;
return;
}

/* ----------------------------------------------------------------------
  Convert a calendar date to a Julian date.
  Converted from FORTRAN. -tab 06aug2012
*/
/*@@*/
void gal_calendar_to_julian( double Y, double M, double DY, double *JJ, double *FF)
{
/**/
double D1,G,S,ATEMP,J3,J,F;
/**/
J3 = 0;
G  = 1.0;
if (Y < 1583.0) { G = 0.0; }
D1 = (double)((int)DY);
F  = DY - D1 - 0.5;
J = -1.0 * (double)((int)( 7.0 * ( (double)((int)(  (M + 9.0) / 12.0 )) + Y  ) / 4.0  ));
if (G != 0.) {
  if ((M - 9.0) < 0.) {
    S = -1.0;
  } else {
     S = +1.0;
  }
  ATEMP = ABS((M - 9.0));
  J3  = (double)((int)(  Y + ( S *  (double)((int)( ATEMP / 7.0 ))  )  ));
  J3  = -1.0 * (double)((int)(  ((double)((int)( J3 / 100.0 ))  +  1.0)  *  0.75 ));
}
J = J + (double)((int)( 275.0 * M / 9.0 )) + D1 + (G * J3);
J = J + 1721027.0 + (2.0 * G) + (367.0 * Y);
if (F < 0.) {
  F = F + 1.0;
  J = J - 1.0;
}
*JJ = J;
*FF = F;
printf("*JJ=%f  *FF=%f\n",*JJ,*FF);
return;
}


/* --------------------------------------------------------------------------
  Sun position calculations.                         tab  July 1994
  Taken from Sky and Telescope, August 1994 (page 84), by Roger W. Sinnott.
  Calculations the RA and DEC of the Sun at the given UT.  UT is given in year,
  month, day, and hour.  Accurate to within about 1 minute of arc.
  Whether the coordinates are 1950 or 2000 or something else is not known.
  It does not say in the Sky and Telescope article.
  Converted from FORTRAN. -tab 06aug2012
*/
/*@@*/
void ORIG_gal_Sun_RA_Dec( int year, int month, int day, double hour, double *ra, double *dec )
{
/**/
double T,TT,Z0,H,Y,M,DY,J,F;
/**/
/* Set time variables. */
H  = hour;
Z0 = H / 24.0;
Y  = (double)year;
M  = (double)month;
DY = (double)day;
gal_calendar_to_julian(Y,M,DY,&J,&F);
T  = (J - 2451545.0) + F;
/* TT is in centuries from 1900.0 */
TT = (T / 36525.0) + 1.0;
T  = T + Z0;
/* Get Sun's position. */
gal_GetSunPosition(T,TT,ra,dec);
return;
}


/* --------------------------------------------------------------------------
  Sun position calculations.                         tab  July 1994
  Taken from Sky and Telescope, August 1994 (page 84), by Roger W. Sinnott.
  Converted to work on Julian Day rather than year,month,day,hour. -tab 06aug2012
  Calculations the RA and DEC of the Sun at the given Julian Day.
  Accurate to within about 1 minute of arc.
  Whether the coordinates are 1950 or 2000 or something else is not known.
  It does not say in the Sky and Telescope article.
  Converted from FORTRAN. -tab 06aug2012
*/
/*@@*/
void gal_Sun_RA_Dec( double jd, double *ra, double *dec )
{
/**/
double T,TT,Z0,J;
/**/
/* Fraction of a day. */
Z0 = (jd - 0.5) - (double)((int)(jd - 0.5));
/* TT is in centuries from 1900.0 */
J = (double)((int)(jd - 0.5));
T  = (J - 2451545.0) + 0.5;
TT = (T / 36525.0) + 1.0;
T  = T + Z0;
/* Get Sun's position. */
gal_GetSunPosition(T,TT,ra,dec);
if (*ra <  0.) { *ra = *ra + 24.; } else {
if (*ra > 24.) { *ra = *ra - 24.; }      }
return;
}


/* ----------------------------------------------------------------------
  Get Moon RA and DEC given T.    Fundamental Arguments.
  The distance "R5" is in earth radii (about 6378 km).
*/
/*@@*/
void gal_GetMoonPosition( double T, double *ra, double *dec, double *R5 )
{
/**/
double A5,D5,S,F,L,M,D,N,G,U,V,W;
const double PI = 3.141592654;
const double PI2= 2. * 3.141592654;
/**/
L = 0.606434 + (0.03660110129*T);
M = 0.374897 + (0.03629164709*T);
F = 0.259091 + (0.03674819520*T);
D = 0.827362 + (0.03386319198*T);
N = 0.347343 - (0.00014709391*T);
G = 0.993126 + (0.00273777850*T);
L = L - (double)((int)(L));
M = M - (double)((int)(M));
F = F - (double)((int)(F));
D = D - (double)((int)(D));
N = N - (double)((int)(N));
G = G - (double)((int)(G));
L = L * PI2;
M = M * PI2;
F = F * PI2;
D = D * PI2;
N = N * PI2;
G = G * PI2;
V =     ( 0.39558 * sin(F+N) );
V = V + ( 0.08200 * sin(F) );
V = V + ( 0.03257 * sin((M-F)-N) );
V = V + ( 0.01092 * sin(M+F+N) );
V = V + ( 0.00666 * sin(M-F) );
V = V - ( 0.00644 * sin((M+F-(2.*D))+N) );
V = V - ( 0.00331 * sin((F-(2.*D))+N) );
V = V - ( 0.00304 * sin(F-(2.*D)) );
V = V - ( 0.00240 * sin(((M-F)-(2.*D))-N) );
V = V + ( 0.00226 * sin(M+F) );
V = V - ( 0.00108 * sin(M+F-(2.*D)) );
V = V - ( 0.00079 * sin(F-N) );
V = V + ( 0.00078 * sin(F+(2.*D)+N) );
U = 1.- ( 0.10828 * cos(M) );
U = U - ( 0.01880 * cos(M-(2.*D)) );
U = U - ( 0.01479 * cos(2.*D) );
U = U + ( 0.00181 * cos((2.*M)-(2.*D)) );
U = U - ( 0.00147 * cos(2.*M) );
U = U - ( 0.00105 * cos((2.*D)-G) );
U = U - ( 0.00075 * cos((M-(2.*D))+G) );
W =     ( 0.10478 * sin(M) );
W = W - ( 0.04105 * sin((2.*F)+(2.*N)) );
W = W - ( 0.02130 * sin(M-(2.*D)) );
W = W - ( 0.01779 * sin((2.*F)+N) );
W = W + ( 0.01774 * sin(N) );
W = W + ( 0.00987 * sin(2.*D) );
W = W - ( 0.00338 * sin((M-(2.*F))-(2.*N)) );
W = W - ( 0.00309 * sin(G) );
W = W - ( 0.00190 * sin(2.*F) );
W = W - ( 0.00144 * sin(M+N) );
W = W - ( 0.00144 * sin((M-(2.*F))-N) );
W = W - ( 0.00113 * sin(M+(2.*F)+(2.*N)) );
W = W - ( 0.00094 * sin((M-(2.*D))+G) );
W = W - ( 0.00092 * sin((2.*M)-(2.*D)) );

/* Compute RA, Dec, Distance. */
S  = W / sqrt(U - (V * V));
A5 = L + atan(S / sqrt(1.0 - (S * S)));
S  = V / sqrt(U);
D5 = atan(S / sqrt(1.0 - (S * S)));
*R5= 60.40974 * sqrt(U);

/* Convert to standard units. */
*ra = A5 * 12.0 / PI;
if (*ra < 0.) *ra = 24.0 + *ra;
*dec= D5 * 180.0 / PI;

return;
}


/* ----------------------------------------------------------------------
  Moon position calculations.                         tab  July 1994
  Taken from Sky and Telescope July 1989, page 78 by Roger W. Sinnott
  Converted to work on Julian Day rather than year,month,day,hour. -tab 06aug2012
  Calculations the RA and DEC of the Moon at the given Julian Day.
  Accurate to within about 40 minutes of arc.
  The distance to the moon "R" is in earth radii (about 6378 km).
  Converted from FORTRAN. -tab 06aug2012
*/
/*@@*/
void gal_Moon_RA_Dec( double jd, double *ra, double *dec, double *R )
{
/**/
double T,Z0,J;
/**/
/* Fraction of a day. */
Z0= (jd - 0.5) - (double)((int)(jd - 0.5));
J = (double)((int)(jd - 0.5));
T = (J - 2451545.0) + 0.5;
T = T + Z0;
/* Get Moon's position. */
gal_GetMoonPosition(T,ra,dec,R);
if (*ra <  0.) { *ra = *ra + 24.; } else {
if (*ra > 24.) { *ra = *ra - 24.; }      }
return;
}


/* ----------------------------------------------------------------------
  Moon illumination... -tab October 2012
  Give a julian date.  Returns lunar illumination fraction.
*/
/*@@*/
double gal_moon_illumination( double jd )
{
/**/
double ill,theta,sra,sdec,mra,mdec,mdist;
/**/
gal_Sun_RA_Dec( jd, &sra, &sdec ); sra = sra * 15.;
gal_Moon_RA_Dec( jd, &mra, &mdec, &mdist );  mra = mra * 15.;
theta = cangsep(sra,sdec,mra,mdec);
if (theta < 90.) {
  ill = (1.0 - cos(degtorad(theta))) / 2.;
} else { 
  ill = (1.0 + cos(degtorad(180.-theta))) / 2.;
}
return(ill);
}


/* ----------------------------------------------------------------------
  Horizon...                         tab  October 2012
  Angle downward towards the sea level horizon from altitude...
  Input is elevation in meters.
  Returns downward angle in degrees.
*/
/*@@*/
double gal_horizon( double elevation )
{
/**/
/* Earth's average radius is about 6370 km, which is also the radius at
   about latitude of 30 degrees.  The radius near the equator is 6378. */
const double EarthRadiusAverage = 6370.;
/**/
double rat,x,y,r,theta;
/**/
x = EarthRadiusAverage;
r = x + (elevation / 1000.);
y = sqrt((r * r) - (x * x));  /* distance to sealevel horizon in km */
rat = y / r;
theta = radtodeg( asin(rat) );
return(theta);
}



/* ------------------------------------------------------------------- */
double copysign(double,double);

/* ------------------------------------------------------------------- */
/* Galactic to Equatorial                                              */
/*                                                                     */
/*@@*/
int pop_galeqd(double xl,double xb,double *rad,double *decd)
{
  /* galeqd,equgad,sglgal,galsgl subroutines moved from jcnvcd.f file to galeqd.f */
  /*   on 14sep92 - no changes to code, itself, made (code same as 17jan91) */
  /* galeqd (modifcation of galequ) requires double prec. args (22oct90) */
  /*ccc  entry equcor(xl,xb,rad,decd)   obsolete entry pt equcor. */
  /* 22mar85 for hepvax (changed darsin to asin;removed entry equcor) */
  /* 17mar82 updated so has more sensible name- galequ, but still */
  /*   retains alias for equcor. */
  /* 02oct80  jdb   29jan81 to input dtor,rtod to more places. 22apr81 */
  /* subroutine to compute earth equatorial coords from galactic coords */
  /* inputs: xl l in degrees, xb b in degrees. */
  /* b = galactic latitude (-90. to + 90. deg);l=galactic longitude(0-360) */
  /* outputs: rad ra in degrees, decd dec in degrees. */
  /*  x = cos(b)*cos(l) */
  /*  y = sin(l)*cos(b) */
  /*  z = sin(b) */
  /*  gal <==> equ equations from tom soifer.  see also: */
  /*   Classical Mechanics by Herbert Goldstein (Addison-Wesley Publ.Co., */
  /*   c.1950, 7th printing 1965) Section 4-4 on Eulerian Angles (p.107-109) */

  /* assumes gal north pole at 12h 49.0m, +27d24' (192.25d,+27.4d) */
  /*         gal (0, 0) at     17h 42.4m, -28d55' (265.36d,-28.91d) */
  /*   also: delta=lat=0 at gal (33d,0d) ; equ (282.25d,0d). */

  double xlr,xbr,cosb,cosl,sinl,x,y,z,sind,
       cosd,decr,rar,eq2,eq3,cosa,sina;

  static double a[4],b[4],c[4];
  static double cosph,sinph,cosps,sinps,costh,sinth;
  static double dtor, rtod;
  static double psi=-33.0, theta=62.6, phi=282.25;
  static int nthru=0;

  if(nthru == 0) {
               dtor = atan(1.0) / 45.0;
               rtod = 1.0 / dtor;
               /*         compute matrix for gal ==> equ conversion: */
               cosps = cos(psi*dtor);
               sinps = sin(psi*dtor);
               cosph = cos(phi*dtor);
               sinph = sin(phi*dtor);
               costh = cos(theta*dtor);
               sinth = sin(theta*dtor);
       
               /* |  b[1]  b[2]  b[3]  | */
               /* |  c[1]  c[2]  c[3]  | */
               /* |  a[1]  a[2]  a[3]  | */
       
               b[1] = cosps*cosph - costh*sinph*sinps;
               b[2] = -(sinps*cosph) - costh*sinph*cosps;
               b[3] = sinth*sinph;
       
               c[1] = cosps*sinph + costh*cosph*sinps;
               c[2] = -(sinps*sinph) + costh*cosph*cosps;
               c[3] = -(sinth*cosph);
       
               a[1] = sinth*sinps;
               a[2] = sinth*cosps;
               a[3] = costh;
       
               nthru = 1;
             }
                                                   
                                                   
  xlr = xl * dtor;
  xbr = xb * dtor;
  cosb = cos(xbr);
  cosl = cos(xlr);
  sinl = sin(xlr);
  z    = sin(xbr);
  x = cosb*cosl;
  y = sinl*cosb;
  sind = a[1]*x + a[2]*y + a[3]*z;
  if(fabs(sind)>=1.) {
                   sind = copysign(1.,sind);
                   decr =  asin(sind);
                   rar = 0.;
                 }
  else {
       eq2  = b[1]*x + b[2]*y + b[3]*z;
       eq3  = c[1]*x + c[2]*y + c[3]*z;
       decr =  asin(sind);
       cosd = sqrt(1.-sind*sind);
       cosa = eq2/cosd;
       sina = eq3/cosd;
       if(fabs(cosa)>1.) cosa = copysign(1.,cosa);
       if(fabs(sina)>1.) sina = copysign(1.,sina);
       rar = atan2(sina,cosa);
       }
  *rad  = rar * rtod;
  if(*rad < 0.0) *rad = 360.0 + *rad;
  *decd = decr * rtod;
  if(fabs(*decd) >= 90.0) {
                      *rad = 0.0;
                      if(*decd >  90.0) *decd =  90.0;
                      if(*decd < -90.0) *decd = -90.0;
                    }
  return(0);                                             
}

/* ------------------------------------------------------------------- */
/* Equatorial to Galactic                                              */
/*                                                                     */
/*@@*/
int pop_equgad(double rad,double decd,double *xl,double *xb)
{
  /* equgad (modification of equgal) requires double prec. args (22oct90) */

  /* 22mar85 for hepvax (darsin changed to asin;entry galcor removed) */
  /*c    entry galcor(rad,decd,xl,xb) */
  /* 17mar82 updated for more sensible name- equgal, while retaining */
  /*     alias for old name of galcor. */
  /* 02oct80 jdb:subroutine to compute galactic coords from earth */
  /*  equatorial coords.29jan81 for more places in dtor,rtod.22apr81 */
  /*  gal <==>equ equations from tom soifer.  see also: */
  /*   Classical Mechanics by Herbert Goldstein (Addison-Wesley Publ.Co., */
  /*     c.1950, 7th printing 1965) Section 4-4 on Eulerian Angles (p.107-109) */
  /* assumes gal north pole at 12h 49.0m, +27d24' (192.25d,+27.4d) */
  /*         gal (0, 0) at     17h 42.4m, -28d55' (265.36d,-28.91d) */
  /*   also: delta=lat=0 at gal (33d,0d) ; equ (282.25d,0d). */

  /* inputs: rad ra in degrees, decd is dec in degrees. */
  /* b = galactic latitude (-90. to + 90. deg);l=galactic longitude(0-360) */

  /*  x = cos(alpha)*cos(delta) */
  /*  y = sin(alpha)*cos(delta) */
  /*  z = sin(delta) */
  /*  eq1 = sin(b) = -.8676*x - .1884*y + .4602*z */
  /*  eq2 = cos(b)*cos(l) = -.0669*x - .8728*y - .4835*z */
  /*  eq3 = cos(b)*sin(l) = +.4927*x - .4503*y + .7446*z */
  /*    b = arsin(sin(b)) */
  /*  cos(b) = sqrt(1.-sin(b)*sin(b)) */
  /*  cos(l) = eq2/cos(b) */
  /*  sin(l) = eq3/cos(b) */
  /*  l = atan2(sin(l)/cos(l): if(l<0) l = l+2*pi if radians. */

  double rar,decr,cosa,sina,cosd,x,y,z,cosb,
         sinb,eq2,eq3,xbr,xlr,cosl,sinl;

  static double a[4],b[4],c[4];
  static double cosph,sinph,cosps,sinps,costh,sinth;
  static double dtor, rtod;
  static double psi=-33.0, theta=62.6, phi=282.25;
  static int nthru=0;

  if(nthru == 0) {
               dtor = atan(1.0) / 45.0;
               rtod = 1.0 / dtor;
               /* compute matrix for equ ==> gal conversion: */
               cosps = cos(psi*dtor);
               sinps = sin(psi*dtor);
               cosph = cos(phi*dtor);
               sinph = sin(phi*dtor);
               costh = cos(theta*dtor);
               sinth = sin(theta*dtor);

               /* |  b[1]  b[2]  b[3]  | */
               /* |  c[1]  c[2]  c[3]  | */
               /* |  a[1]  a[2]  a[3]  | */

               b[1] = cosps*cosph - costh*sinph*sinps;
               b[2] = cosps*sinph + costh*cosph*sinps;
               b[3] = sinps*sinth;

               c[1] = -(sinps*cosph) - costh*sinph*cosps;
               c[2] = -(sinps*sinph) + costh*cosph*cosps;
               c[3] =  cosps*sinth;

               a[1] = sinth*sinph;
               a[2] = -(sinth*cosph);
               a[3] = costh;

               nthru = 1;
             }


  rar  = rad * dtor;
  decr = decd * dtor;
  cosa = cos(rar);
  sina = sin(rar);
  cosd = cos(decr);
  z    = sin(decr);
  x = cosa*cosd;
  y = sina*cosd;
  sinb = a[1]*x + a[2]*y + a[3]*z;
  if(fabs(sinb)>=1.) {
                   sinb = copysign(1.,sinb);
                   xbr =  asin(sinb);
                   xlr = 0.;
                 }
  else {
       eq2  = b[1]*x + b[2]*y + b[3]*z;
       eq3  = c[1]*x + c[2]*y + c[3]*z;
       xbr =  asin(sinb);
       cosb = sqrt(1.-sinb*sinb);
       cosl = eq2/cosb;
       sinl = eq3/cosb;
       if(fabs(cosl)>1.) cosl = copysign(1.,cosl);
       if(fabs(sinl)>1.) sinl = copysign(1.,sinl);
       xlr = atan2(sinl,cosl);
       }
  *xl  = xlr * rtod;
  if(*xl < 0.0) *xl = 360.0 + *xl;
  *xb = xbr * rtod;
  if(fabs(*xb) >= 90.0) {
                    *xl = 0.0;
                    if(*xb >  90.0) *xb =  90.0;
                    if(*xb < -90.0) *xb = -90.0;
                  }

  return(0);
}

/* ------------------------------------------------------------------- */
/*@@*/
int pop_galsgl(double rad,double decd,double *xl,double *xb)
{
  /* galsgl (modification of equgad) to compute supergalactic coordinates */
  /*  from galactic coordinates.  j.bennett  30oct-05nov90 */

  /* inputs: rad =  gal lon (l) in degrees. double. */
  /*         decd = gal lat (b) in degrees. double. */
  /* returned: */
  /* xb = supergalactic latitude  (SGB)  (-90. to + 90. deg). double. */
  /* xl = supergalactic longitude (SGL)  (0-360. deg). double. */

  /*  Computed values of rotation used for the Eulerian angles: */
  /*   phi = 137.37 deg.;  theta = 83.68 deg.;  psi = 0. deg. */
  /*    Based on gal l,b = 137.37, 0 deg. at sg SGL,SGB = 0., 0. and */
  /*         SG North Pole (0,+90) at gal l,b = 47.37, +6.32 deg. */

  /* References: */
  /* de Vaucouleurs,G., A. de Vaucouleurs, and G.H.Corwin, */
  /*  Second Reference Catalog of Bright Galaxies, (1976), p.8 */
  /*  */
  /* Tully,B., Nearby Galaxies Catalog, (1988) p. 1,4-5 */
  /*   */
  /*    Note: de Vaucouleurs gives gal l,b 137.29, 0 for SGL,SGB= 0,0; */
  /*          this is not 90 degrees from Galactic North Pole.- used */
  /*          Tully's value of 137.37, 0 deg. */

  /*  gal <==> equ equations from tom soifer.  see also: */
  /*   Classical Mechanics by Herbert Goldstein (Addison-Wesley Publ.Co., */
  /*     c.1950, 7th printing 1965) Section 4-4 on Eulerian Angles  */
  /*     (p.107-109) */
  /*    note: such equations also appropriate for gal <==> sgl . */
  /*          See code below for def. of arrays a,b,c. */
  /*  x = cos(lon)*cos(lat) */
  /*  y = sin(lon)*cos(lat) */
  /*  z = sin(lat) */
  /*  eq1 = sin(sgb)          =   a[1]*x +  a[2]*y +  a[3]*z */
  /*  eq2 = cos(sgb)*cos(sgl) =   b[1]*x +  b[2]*y +  b[3]*z */
  /*  eq3 = cos(sgb)*sin(sgl) =   c[1]*x +  c[2]*y +  c[3]*z */
  /*  sgb = arsin(sin(sgb)) */
  /*  cos(sgb) = sqrt(1.-sin(sgb)*sin(sgb)) */
  /*  cos(sgl) = eq2/cos(sgb) */
  /*  sin(sgl) = eq3/cos(sgb) */
  /*  sgl = atan2(sin(sgl)/cos(sgl): if(sgl<0) sgl = sgl+2*pi if radians. */

  double rar,decr,cosa,sina,cosd,x,y,z,cosb,
         sinb,eq2,eq3,xbr,xlr,cosl,sinl;

  static double a[4],b[4],c[4];
  static double cosph,sinph,cosps,sinps,costh,sinth;
  static double dtor, rtod;
  static double psi=  0.0, theta=83.68, phi=137.37;
  static int nthru=0;

  if(nthru == 0) {
               dtor = atan(1.0) / 45.0;
               rtod = 1.0 / dtor;
               /* compute matrix for equ ==> gal conversion: */
               cosps = cos(psi*dtor);
               sinps = sin(psi*dtor);
               cosph = cos(phi*dtor);
               sinph = sin(phi*dtor);
               costh = cos(theta*dtor);
               sinth = sin(theta*dtor);

               /* |  b[1]  b[2]  b[3]  | */
               /* |  c[1]  c[2]  c[3]  | */
               /* |  a[1]  a[2]  a[3]  | */

               b[1] = cosps*cosph - costh*sinph*sinps;
               b[2] = cosps*sinph + costh*cosph*sinps;
               b[3] = sinps*sinth;

               c[1] = -(sinps*cosph) - costh*sinph*cosps;
               c[2] = -(sinps*sinph) + costh*cosph*cosps;
               c[3] =  cosps*sinth;

               a[1] = sinth*sinph;
               a[2] = -(sinth*cosph);
               a[3] = costh;

               nthru = 1;
             }


  rar  = rad * dtor;
  decr = decd * dtor;
  cosa = cos(rar);
  sina = sin(rar);
  cosd = cos(decr);
  z    = sin(decr);
  x = cosa*cosd;
  y = sina*cosd;
  sinb = a[1]*x + a[2]*y + a[3]*z;
  if(fabs(sinb)>=1.) {
                   sinb = copysign(1.,sinb);
                   xbr =  asin(sinb);
                   xlr = 0.;
                 }
  else {
       eq2  = b[1]*x + b[2]*y + b[3]*z;
       eq3  = c[1]*x + c[2]*y + c[3]*z;
       xbr =  asin(sinb);
       cosb = sqrt(1.-sinb*sinb);
       cosl = eq2/cosb;
       sinl = eq3/cosb;
       if(fabs(cosl)>1.) cosl = copysign(1.,cosl);
       if(fabs(sinl)>1.) sinl = copysign(1.,sinl);
       xlr = atan2(sinl,cosl);
       }
  *xl  = xlr * rtod;
  if(*xl < 0.0) *xl = 360.0 + *xl;
  *xb = xbr * rtod;
  if(fabs(*xb) >= 90.0) {
                    *xl = 0.0;
                    if(*xb >  90.0) *xb =  90.0;
                    if(*xb < -90.0) *xb = -90.0;
                  }

  return(0);
}

/* ------------------------------------------------------------------- */
/*@@*/
int pop_sglgal(double xl,double xb,double *rad,double *decd)
{
  /* sglgal (modification of galeqd) to compute galactic coordinates */
  /*  from supergalactic coordinates.  j.bennett  30oct-05nov90 */

  /* inputs: */
  /* xb = supergalactic latitude  (SGB)  (-90. to + 90. deg). double. */
  /* xl = supergalactic longitude (SGL)  (0-360. deg). double. */
  /* returned: rad =  gal lon (l) in degrees. double. */
  /*           decd = gal lat (b) in degrees. double. */

  /*  Computed values of rotation used for the Eulerian angles: */
  /*   phi = 137.37 deg.;  theta = 83.68 deg.;  psi = 0. deg. */
  /*    Based on gal l,b = 137.37, 0 deg. at sg SGL,SGB = 0., 0. and */
  /*         SG North Pole (0,+90) at gal l,b = 47.37, +6.32 deg. */

  /* References: */
  /* de Vaucouleurs,G., A. de Vaucouleurs, and G.H.Corwin, */
  /*  Second Reference Catalog of Bright Galaxies, (1976), p.8 */
  /*  */
  /* Tully,B., Nearby Galaxies Catalog, (1988) p. 1,4-5 */
  /*   */
  /*    Note: de Vaucouleurs gives gal l,b 137.29, 0 for SGL,SGB= 0,0; */
  /*          this is not 90 degrees from Galactic North Pole.- used */
  /*          Tully's value of 137.37, 0 deg. */

  /*  gal <==> equ equations from tom soifer.  see also: */
  /*   Classical Mechanics by Herbert Goldstein (Addison-Wesley Publ.Co., */
  /*     c.1950, 7th printing 1965) Section 4-4 on Eulerian Angles  */
  /*     (p.107-109) */
  /*    note: such equations also appropriate for gal <==> sgl . */
  /*          See code below for def. of arrays a,b,c. */
  /*  x = cos(sgb)*cos(sgl) */
  /*  y = sin(sgl)*cos(sgb) */
  /*  z = sin(sgb) */
  /*  eq1 = sin(lat)          =   a[1]*x +  a[2]*y +  a[3]*z */
  /*  eq2 = cos(lat)*cos(lon) =   b[1]*x +  b[2]*y +  b[3]*z */
  /*  eq3 = cos(lat)*sin(lon) =   c[1]*x +  c[2]*y +  c[3]*z */
  /*  sgb = arsin(sin(lat)) */
  /*  cos(lat) = sqrt(1.-sin(lat)*sin(lat)) */
  /*  cos(lon) = eq2/cos(lat) */
  /*  sin(lon) = eq3/cos(lat) */
  /*  lon = atan2(sin(lon)/cos(lon): if(lon<0) lon = lon+2*pi if radians. */

  double xlr,xbr,cosb,cosl,sinl,x,y,z,sind,
         cosd,decr,rar,eq2,eq3,cosa,sina;

  static double a[4],b[4],c[4];
  static double cosph,sinph,cosps,sinps,costh,sinth;
  static double dtor, rtod;
  static double psi=  0.0, theta=83.68, phi=137.37;
  static int nthru=0;

  if(nthru == 0) {
               dtor = atan(1.0) / 45.0;
               rtod = 1.0 / dtor;

               /* compute matrix for gal ==> equ conversion: */
               cosps = cos(psi*dtor);
               sinps = sin(psi*dtor);
               cosph = cos(phi*dtor);
               sinph = sin(phi*dtor);
               costh = cos(theta*dtor);
               sinth = sin(theta*dtor);

               /* |  b[1]  b[2]  b[3]  | */
               /* |  c[1]  c[2]  c[3]  | */
               /* |  a[1]  a[2]  a[3]  | */

               b[1] = cosps*cosph - costh*sinph*sinps;
               b[2] = -(sinps*cosph) - costh*sinph*cosps;
               b[3] = sinth*sinph;

               c[1] = cosps*sinph + costh*cosph*sinps;
               c[2] = -(sinps*sinph) + costh*cosph*cosps;
               c[3] = -(sinth*cosph);

                 a[1] = sinth*sinps;
               a[2] = sinth*cosps;
               a[3] = costh;

               nthru = 1;
             }


  xlr = xl * dtor;
  xbr = xb * dtor;
  cosb = cos(xbr);
  cosl = cos(xlr);
  sinl = sin(xlr);
  z    = sin(xbr);
  x = cosb*cosl;
  y = sinl*cosb;
  sind = a[1]*x + a[2]*y + a[3]*z;
  if(fabs(sind)>=1.) {
                   sind = copysign(1.,sind);
                   decr =  asin(sind);
                   rar = 0.;
                 }
  else {
       eq2  = b[1]*x + b[2]*y + b[3]*z;
       eq3  = c[1]*x + c[2]*y + c[3]*z;
       decr =  asin(sind);
       cosd = sqrt(1.-sind*sind);
       cosa = eq2/cosd;
       sina = eq3/cosd;
       if(fabs(cosa)>1.) cosa = copysign(1.,cosa);
       if(fabs(sina)>1.) sina = copysign(1.,sina);
       rar = atan2(sina,cosa);
       }
  *rad  = rar * rtod;
  if(*rad < 0.0) *rad = 360.0 + *rad;
  *decd = decr * rtod;
  if(fabs(*decd) >= 90.0) {
                      *rad = 0.0;
                      if(*decd >  90.0) *decd =  90.0;
                      if(*decd < -90.0) *decd = -90.0;
                    }

  return(0);
}

/* ------------------------------------------------------------------- */
/* Direct copy from Judy Bennet's ECLEQX.F routine, translated to C.   */
/*  ecleqx.f  contains subs jgtobq,ecleqx,equecx                       */
/*    30jul92  j.bennett                                               */
/*                                                                     */
/*@@*/
void jgtobq( int iaus, double eqx, double *ep)
{
/*  find value of epsilon (obliquity of ecliptic) for eqx */
/*   iaus is iau system : 1=Besselian; 2 (or anything but 1) = Julian. */
/*   eqx is equinox year (double precision) in terms of system designated */
/*        by iaus.  e.g. if B1950 wanted, then iau=1,eqx=1950.0d0; */
/*        if J2000 wanted, then iau=2,eqx=2000.0d0 */
/*   ep is epsilon in degrees (returned)  (double precision) */
/*  */
/*     iaus integer; all other arguments double precision. */
/*  if iaus = 2 (Julian): */
/*  ep   = epsilon = obliquity of the ecliptic of date (with respect to the */
/*         mean equator of date.(value from p.S26 of 1984 Astr.Alman.Supp) */
/*         epsilon = 84381.448" - 46.8150"*t - 0.00059"t*t + 0.001813"*t*t*t */
/*      where t is time measured in Julian centuries from epoch J2000.0 */
/*           note: the resulting ep when  t = -.5000021 (for B1950) is used */
/*                 differs by .015" of value usually */
/*                 used for B1950 of 23d26'44.84"(the value usually used */
/*                 for B1950). This discrepancy should cause no problems in */
/*                 in our precession conversion calculations. */
/*                 ep is used only in the E-term removal portion of conversion  */
/*                 at poles.  However, for real work, where the actually */
/*                 position is required, use the following formula for Besselian. */
/*  if iaus = 1 (Besselian): */
/*  ep   = epsilon = obliquity of the ecliptic (p.B18 of 1981-1983Astr.Alman.) */
/*         epsilon = 84404.84"  - 46.850"*t - 0.0033t*t + 0.00182"*t*t*t */
/*      where t is time measured in Besselian centuries from epoch B1950.0 */


  double t, t2, t3;

  if(iaus != 1) {
    t = (eqx - 2000.0) * 0.01;
    t2 = t*t;
    t3 = t*t2;
    
    *ep = (84381.448 - 46.8150*t - 0.00059*t2 + 0.001813*t3)/
      3600.0;
    
  }
  
  else {
    
    t = (eqx - 1950.0) * 0.01;
    t2 = t*t;
    t3 = t*t2;
    
    *ep = (84404.84  - 46.850*t  - 0.0033*t2  + 0.00182*t3)/
      3600.0;

  }

  return;
}


/* ------------------------------------------------------------------- */
/* Ecliptic to Equatorial                                              */
/*                                                                     */
/*@@*/
int pop_ecleqx(int iaus, double eqx, double xlam, double beta, 
          double *rad, double *decd)
{
/*  version of ecleqd that adds args.iaus and  eqx so epsilon may be  */
/*        determined for that date.  Calls jgtobq. */
/*  ecliptic to  equatorial coordinate subroutine */

/*  equations from j.fowler (jan 1981) */

/*     iaus integer; all other arguments double precision. */
/*  inputs: */
/*   iaus is iau system : 1=Besselian; 2 (or anything but 1) = Julian. */
/*       iaus must apply to both input and output positions (ecleqx does */
/*       no Besselian to/from Julian conversions. */
/*   eqx is equinox year (double precision) in terms of system designated */
/*        by iaus.  e.g. if B1950 wanted, { iau=1,eqx=1950.0d0; */
/*        if J2000 wanted, { iau=2,eqx=2000.0d0 */
/*        Note: eqx applies to both input and output (ecleqx does no precession). */
/*   xlam=lambda=ecliptic longitude in decimal degrees (0.0d0 - 360.0d0 deg) */

/*   beta= ecliptic latitude in decimal degrees (-90.0d0 to +90.0d0 deg) */

/*   xl = sin(beta)              xe = xl*cos(e) - yl*sin(e) */
/*   yl = -cos(beta)*sin(lambda) ye = xl*sin(e) + yl*cos(e) */
/*   zl = cos(beta)*cos(lambda)  ze = zl */

/*   outputs: rad ra in decimal degrees; decd dec in decimal degrees. */

  double cosb,cosl,sinb,sinl,xl,yl,zl,xe,ye,ze,
         xlamr,betar,rar,decr,copysign();
  double e;
  static double leqx=-1.0, dtor, rtod, cose, sine;
  static int  nthru=0, laus=-99;

  if(nthru == 0) {
    dtor = atan(1.0) / 45.0;
    rtod = 1.0 / dtor;
    nthru = 1;
  }

  if(eqx != leqx || iaus != laus) {
    jgtobq(iaus,eqx,&e);
    e = e * dtor;
    cose = cos(e);
    sine = sin(e);
    leqx = eqx;
    laus = iaus;
  }

  xlamr = xlam*dtor;
  betar = beta*dtor;
  
  cosb = cos(betar);
  cosl = cos(xlamr);
  sinb = sin(betar);
  sinl = sin(xlamr);
  
  xl = sinb;
  yl = -(cosb*sinl);
  zl = cosb*cosl;

  xe = cose*xl - sine*yl;
  ye = sine*xl + cose*yl;
  ze = zl;

  rar = atan2(-ye,ze);
  *rad = rar * rtod;
  if(*rad < 0.0) *rad = 360.0 + *rad;

/*  try to catch pole on any machine (& return ra=0 {) */
  if(fabs(xe) > 1.0) {
    *decd = copysign(90.0,xe);
    *rad = 0.0;
  }
  else {
    decr =  asin(xe);
    *decd = decr * rtod;
    if(fabs(*decd) > 90.0) {
      *rad = 0.0;
      if(*decd >  90.0) *decd =  90.0;
      if(*decd > -90.0) *decd = -90.0;
    }
  }
  
  return(0);
}

/* ------------------------------------------------------------------- */
/* Equatorial to Ecliptic                                              */
/*                                                                     */
/*@@*/
int pop_equecx(int iaus, double eqx, double rad, double decd, 
          double *xlam, double *beta)
{
/*  version of equecd that adds args. iaus and  eqx so epsilon may be */
/*         determined for that date.  Calls jgtobq. */
/*   equatorial to ecliptic conversion coordinate subroutine */

/*   equations from j.fowler (jan 1981) */

/*  inputs: */
/*   iaus is iau system : 1=Besselian; 2 (or anything but 1) = Julian. */
/*       iaus must apply to both input and output positions (equecx does */
/*       no Besselian to/from Julian conversions. */
/*   eqx is equinox year (double precision) in terms of system designated */
/*        by iaus.  e.g. if B1950 wanted, { iau=1,eqx=1950.0d0; */
/*        if J2000 wanted, { iau=2,eqx=2000.0d0 */
/*        Note: eqx applies to both input and output (equecx does no precession). */
/*   rad ra in deciaml degrees; decd dec in decimal degrees. */

/*  returned: xlam=lambda=ecliptic longitude in decimal deg.(0.0d0 - 360.0d0 deg) */

/*            beta= ecliptic latitude in decimal degrees (-90.0d0 to +90.0d0 deg) */

/*   xe = sin(dec)              xl = xe*cos(e) - ye*sin(e) */
/*   ye = -cos(dec)*sin(ra)     yl = xe*sin(e) + ye*cos(e) */
/*   ze = cos(dec)*cos(ra)      zl = ze */

  double cosd,cosr,sind,sinr,xl,yl,zl,xe,ye,ze,
         xlamr,betar,rar,decr,copysign();
  double e;

  static double leqx=-1.0, dtor, rtod, cose, sine;
  static int  nthru=0, laus=-99;

  if(nthru == 0) {
    dtor = atan(1.0) / 45.0;
    rtod = 1.0 / dtor;
    nthru = 1;
  }

  if(eqx != leqx || iaus != laus) {
    jgtobq(iaus,eqx,&e);
    e = e * dtor;
    cose = cos(e);
    sine = sin(e);
    leqx = eqx;
    laus = iaus;
  }

  rar = rad*dtor;
  decr = decd*dtor;

  cosd = cos(decr);
  cosr = cos(rar);
  sind = sin(decr);
  sinr = sin(rar);

  xe = sind;
  ye = -(cosd*sinr);
  ze = cosd*cosr;

  xl = cose*xe + sine*ye;
  yl = -sine*xe + cose*ye;
  zl = ze;

  if(fabs(xl) > 1.0) {
    *beta = copysign(90.0,xl);
    *xlam = 0.0;
    return(0);
  }
  else {
    betar =  asin(xl);
    xlamr = atan2(-yl,zl);
    *xlam = xlamr * rtod;
    if(*xlam < 0.0) *xlam = 360.0 + *xlam;
    *beta = betar * rtod;
    if(fabs(*beta) > 90.0) {
      if(*beta >  90.0) *beta =  90.0;
      if(*beta < -90.0) *beta = -90.0;
      *xlam = 0.0;
    }
  }

  return(0);
}


/* ----------------------------------------------------------------------
  Compute true zenith and azimuth angles given the hour angle and
  declination in decimal degrees.  Returns tzd, taz in decimal degrees.
*/
/*@@*/
void gal_true_altaz( double ha, double dec, double latitude, double *tzd, double *taz )
{
/**/
double xp,yp,zp,x,y,z,decrad,harad,latrad;
/**/
const double PI    = 3.1415926535898;
const double HALFPI= 1.5707963267949;
const double  TWOPI= 6.2831853071796;
/**/
/* Set. */
decrad = degtorad(dec);
harad  = degtorad(ha);
latrad = degtorad(latitude);
/* Now calculate azimuth and True Zenith Distance (tzd) */
zp = sin(decrad);
xp = cos(decrad) * sin(harad);
yp = cos(decrad) * cos(harad);
x  = xp;
z  = (zp * sin(latrad)) + (yp * cos(latrad));
y  = (yp * sin(latrad)) - (zp * cos(latrad));
if (ABS(z) > 1.e-20) {
  *tzd = atan(( sqrt((x*x) + (y*y)) / z ));
} else {
  *tzd = HALFPI;
}
if (ABS(y) > 1.e-20) {
  *taz = atan(x/y);
} else {
  if (x >= 0.) {
    *taz = HALFPI + PI;
  } else {
    *taz = PI;
  }
}
if (y > 0.) { *taz = PI + *taz; }
if (*taz < 0.) { *taz = TWOPI + *taz; }
*tzd = radtodeg(( *tzd ));
*taz = radtodeg(( *taz ));
return;
}


/* ----------------------------------------------------------------------
  Compute apparent zenith angle at 4000 Angstroms, given the
  true zenith angle in degrees.  Return -99. if the true angle is  
  too large (>80) to do an accurate calculation.  Assume pressure=760 mm Hg
  and T=15 deg Celius, dry air, from CRC and other sources
  Returns apparent zenith angle in degrees.            -tab 10oct2012
*/
/*@@*/
double gal_apparent_zenith_angle( double tzd )
{
/**/
int i;
double r,azd,oldazd,differr;
/**/
if ((tzd < 0.)||(tzd > 80.)) return(-99.);
azd=degtorad(tzd); i=0 ; differr=1.0;
while ((i < 40)&&(differr > 1.e-10)) {
  i = i + 1; 
  oldazd = azd;
  r = tan(azd);
  r = (58.16 * r) - (0.067 * r * r * r);
  azd = degtorad(tzd) - degtorad(r / 3600.);
  differr = ABS(( oldazd - azd ));
}
if (i > 39) {
  printf("===warning: gal_apparent_zenith_angle: too many iterations.\n");
}
return(radtodeg(azd));
}


/* ----------------------------------------------------------------------
  Compute airmass....    taken from ast-util.c in MAKEE -tab 10oct2012
  The Earth's average radius is about 6370 km.  The Atmosphere's thickness is
  about 11.6 km (level of tropopause at about 41 degrees latitude).
  Divide the Earth's average radius by the atmosphere's thickness.
 
      r = 6370.d0 / 11.6d0

  This formula is very inaccurate for zenith angles greater than 80.
  Note that the apparent zenith angle is used here.
  Returns airmass given the apparent zenith angle in degrees.
  Returns -99. if problem (like azd > 80).
*/
/*@@*/
double gal_airmass( double azd )
{
/**/
const double r = 6370. / 11.6;
const double PI    = 3.1415926535898;
double rr,airmass;
/**/
if ((azd < 0.)||(azd > 80.)) return(-9.);
rr = r * cos((PI - degtorad(azd) ));
airmass = rr + sqrt( (rr*rr) + 1. + (2. * r) );
return(airmass);
}


/* ----------------------------------------------------------------------
  Differential refraction between two wavelengths at a given apparent
  zenith angle.  From CRC (za<50 deg,760mm Hg,T=15C,dry).. 
  See astro_lib.f in MAKEE package.. 
  Returns differential refraction in arcseconds.
*/
/*@@*/
double gal_diff_refrac( double w1, double w2, double azd )
{
/**/
double n1,n2,dr;
/**/
n1 = 2726.43 + (12.288 / (w1*w1*1.e-8))+ (0.3555/(w1*w1*w1*w1*1.e-16));
n1 = n1 * (1.e-7);
n2 = 2726.43 + (12.288 / (w2*w2*1.e-8))+ (0.3555/(w2*w2*w2*w2*1.e-16));
n2 = n2 * (1.e-7);
dr = (2.06265e+5) * (n1 - n2) * tan(degtorad(azd));
/* FORTRAN:
N1=2726.43+(12.288/(W1*W1*1.0D-8))+ (0.3555/(W1*W1*W1*W1*1.0D-16))
N1=N1*(1.0D-7)
N2=2726.43+(12.288/(W2*W2*1.0D-8))+ (0.3555/(W2*W2*W2*W2*1.0D-16))
N2=N2*(1.0D-7)
R = (2.06265D+5)*(N1-N2)*tandd(ZA)
*/
return(dr);
}



/* ----------------------------------------------------------------------
  Given a range of two JDs, return moonrise and moonset times (in JD)...
  (if they exist).  -tab 22nov2012
*/
/*@@*/
void gal_Moon_RiseSet( double jd1, double jd2, double *jd_rise, double *jd_set )
{
/**/
double tzd,taz,alt,ha,lst,jd,ra,dec,R,last_alt;
/**/
/* Initial point. */
jd=jd1-0.0001;
lst = local_sidereal( jd, P48LONG );   lst= lst* 15.;
gal_Moon_RA_Dec( jd, &ra, &dec, &R );  ra = ra * 15.;
ha = ra - lst;
gal_true_altaz( ha, dec, P48LAT, &tzd, &taz );
if (tzd < 0.) { last_alt = -90. - tzd; } else { last_alt = 90. - tzd; }
*jd_rise=-1.; *jd_set=-1.; 
for (jd=jd1; jd<jd2; jd=jd+0.0001) {
  lst = local_sidereal( jd, P48LONG );   lst= lst* 15.;
  gal_Moon_RA_Dec( jd, &ra, &dec, &R );  ra = ra * 15.;
  ha = ra - lst;
  gal_true_altaz( ha, dec, P48LAT, &tzd, &taz );
  if (tzd < 0.) { alt = -90. - tzd; } else { alt = 90. - tzd; }
  if (ABS((alt)) < 0.1) {
    if ((alt > last_alt)&&(*jd_rise < 0.)) { *jd_rise = jd; } else {
    if ((alt < last_alt)&&(*jd_set  < 0.)) { *jd_set  = jd; }      }
  }
  last_alt = alt;
}
return;
}

