/* mua.c : Miscellaneous Utility Archive      -tab 2012 */

#include "nsx.h"
#include "galactic.h"



/* ----------------------------------------------------------------------
   Finds the position of the end-of-string character '\0'.
   The first character position is 0.
   Returns -1 is no end-of-string character.
   The returned value is also number of characters in string preceding '\0'.
   The last real character in the string is this value minus 1.
*/
/*@@*/
int ceos(char a[])
{
int ii,ia,eos;
eos = -1;
ia = strlen(a);
ii=0; while (ii <= ia)  { if (a[ii] == '\0') eos=ii;  ++ii;  }
return(eos);
}


/* ---------------------------------------------------------------------- */
/* Convert degrees to radians.  */
/*@@*/
double degtorad(double theta)
{
  const double pio180 = 0.017453293;   /* PI over 180 degrees. */
  return(theta*pio180);
}

/* ---------------------------------------------------------------------- */
/* Convert radians to degrees.  */
/*@@*/
double radtodeg(double theta)
{
  const double pio180 = 0.017453293;   /* PI over 180 degrees. */
  return(theta/pio180);
}

/* ----------------------------------------------------------------------
  RA angle separation (absolute).
  Assume both RA values are >= 0. and <360. 
*/
/*@@*/
double misc_rasep( double ra1, double ra2)
{
double rr;
if (ra2 > ra1) { rr = ra2 - ra1; } else { rr = ra1 - ra2; }
if (rr > 180.) rr = 360. - rr;
return(rr);
}


/* -----------------------------------------------------------------------
  Convert x,y,z unit sphere to RA,DEC.
  Input: x,y,z (coordinates on unit sphere).
  Output: ra,dec (RA and DEC in decimal degrees).
*/
/*@@*/
void xyz2radec(double x, double y, double z, double *ra, double *dec)
{
*dec = radtodeg( asin(z) );
*ra  = radtodeg( atan2(y,x) );
if ( *ra < 0. ) *ra += 360.;
return;
}

/* -----------------------------------------------------------------------
  Convert RA,DEC to x,y,z unit sphere.
  Input: ra,dec (RA and DEC in decimal degrees).
  Output: x,y,z (coordinates on unit sphere).
*/
/*@@*/
void radec2xyz(double ra, double dec, double *x, double *y, double *z)
{
*x = cos(degtorad(dec)) * cos(degtorad(ra));
*y = cos(degtorad(dec)) * sin(degtorad(ra));
*z = sin(degtorad(dec));
return;
}

/* ---------------------------------------------------------------------
  Convert x,y,z on unit sphere to primed coordinate frame using
  rotation matrix.  Set matrix using SetRotationMatrix().
  The prime system is set up such that the x-prime axis points in the direction
  of the RA,DEC origin, and the z-prime axis lies in the plane of the x-prime
  axis and the original z axis.
  On the new prime image plane, up-down (row#) will be zp, and left-right
  (col#) will be yp, where yp and zp are in radians.
  Input:  x,y,z     (on unit sphere)
          RM        (3x3 rotation matrix)
  Output: xp,yp,zp  (in primed frame)
*/
/*@@*/
void xyz2prime(double x, double y, double z, double RM[3][3],
               double *xp, double *yp, double *zp)
{
*xp= RM[0][0]*x + RM[1][0]*y + RM[2][0]*z;
*yp= RM[0][1]*x + RM[1][1]*y + RM[2][1]*z;
*zp= RM[0][2]*x + RM[1][2]*y + RM[2][2]*z;
return;
}

/* ---------------------------------------------------------------------
  Convert x,y,z on unit sphere to primed coordinate frame using
  rotation matrix.  Set matrix using SetRotationMatrix().

  The prime system is set up such that the x-prime axis points in the direction
  of the RA,DEC origin, and the z-prime axis lies in the plane of the x-prime
  axis and the original z axis.

  Input:  xp,yp,zp  (in prime system)
          RMinv     (3x3 inverse rotation matrix)
  Output: x,y,z     (on unit sphere)
*/
/*@@*/
void prime2xyz(double xp, double yp, double zp, double RMinv[3][3],
               double *x, double *y, double *z)
{
*x= RMinv[0][0]*xp + RMinv[1][0]*yp + RMinv[2][0]*zp;
*y= RMinv[0][1]*xp + RMinv[1][1]*yp + RMinv[2][1]*zp;
*z= RMinv[0][2]*xp + RMinv[1][2]*yp + RMinv[2][2]*zp;
return;
}


/* ---------------------------------------------------------------------
  Convert ra and dec (in decimal degrees) to prime coordinates (in degrees ypd,zpd only).
  Input:  ra,dec   (in decimal degrees)
          RMinv    (3x3 inverse rotation matrix)
  Output: ypd,zpd  (in degrees, in prime system)
*/
/*@@*/
void dd_radec2prime( double ra, double dec, double RMinv[3][3], double *ypd, double *zpd )
{
double xp,yp,zp,x,y,z;
radec2xyz(ra,dec,&x,&y,&z);
xyz2prime(x,y,z,RMinv,&xp,&yp,&zp);
*ypd = radtodeg(yp);
*zpd = radtodeg(zp);
return;
}

/* ---------------------------------------------------------------------
  Convert prime coordinates (in degrees ypd,zpd only) to ra,dec.
  Input:  ypd,zpd  (in degrees, in prime system)
          RMinv    (3x3 inverse rotation matrix)
  Output: ra,dec   (in decimal degrees)
*/
/*@@*/
void dd_prime2radec( double ypd, double zpd, double RMinv[3][3], double *ra, double *dec )
{
double xp,yp,zp,x,y,z;
yp = degtorad(ypd);
zp = degtorad(zpd);
xp = sqrt( 1. - ((yp*yp) + (zp*zp)) );
prime2xyz(xp,yp,zp,RMinv,&x,&y,&z);
xyz2radec(x,y,z,ra,dec);
return;
}




/* ---------------------------------------------------------------------
  Returns the determinant of a 2x2 matrix
  Input:  aa     (2x2 matrix).
*/
/*@@*/
double MatDet2D(double aa[2][2])
{
return( ( aa[0][0] * aa[1][1] ) - ( aa[1][0] * aa[0][1] ) );
}

/* ---------------------------------------------------------------------
  Returns the sub-determinant within a 3x3 matrix.
  The determinant of the 2x2 matrix obtained by deleting the
  "col" column and the "row" row.
  Input:  col,row  : column and row to be deleted.
          aa       : a 3x3 matrix.
*/
/*@@*/
double MatSubDet3D(int col, int row, double aa[3][3])
{
/**/
double bb[2][2];
int xx,yy,ii,jj;
/**/
/* Initialize. */
/* Form sub matrix */
xx=-1;
for (ii=0; ii<3; ++ii) {
if (ii != col) {
  ++xx;
  yy=-1;
  for (jj=0; jj<3; ++jj) {
  if (jj != row) {
    ++yy;
    bb[xx][yy] = aa[ii][jj];
  }
  }
}
}
return( MatDet2D(bb) );
}

/* ---------------------------------------------------------------------
  Returns the determinant of a 3x3 matrix
  Input:  aa     (3x3 matrix).
*/
/*@@*/
double MatDet3D(double aa[3][3])
{
return(
     aa[0][0] * MatSubDet3D(0,0,aa)
   - aa[1][0] * MatSubDet3D(1,0,aa)
   + aa[2][0] * MatSubDet3D(2,0,aa)
      );
}

/* ---------------------------------------------------------------------
  Returns the co-factor within a 3x3 matrix for column "col" and row "row".
  Input: col,row    (column and row).
         aa         (3x3 matrix).
*/
/*@@*/
double MatCof3D(int col, int row, double aa[3][3])
{
double prefix;
/* Prefix. */
if ( (col+row)%2 == 0 ) { prefix = 1.; } else { prefix = -1.; }
/* Sub-determinant. */
return( prefix * MatSubDet3D(col,row,aa) );
}

/* ---------------------------------------------------------------------
  Invert a 3x3 matrix.
  Input:  mm     (3x3 matrix).
  Output: mminv  (inverse 3x3 matrix).
*/
/*@@*/
void MatInv3D(double mm[3][3], double mminv[3][3] )
{
/**/
double aa[3][3],temp;
int col,row;
/**/
/* First create cofactor matrix. */
for (col=0; col<3; ++col) {
  for (row=0; row<3; ++row) {
     aa[col][row] = MatCof3D(col,row,mm);
  }
}
/* Transpose cofactor matrix (classical adjoint). */
for (col=0; col<3; ++col) {
  for (row=0; row<3; ++row) {
    if (col > row) {
      temp         = aa[col][row];
      aa[col][row] = aa[row][col];
      aa[row][col] = temp;
    }
  }
}
/* Determinant of 3x3 matrix. */
temp = MatDet3D(mm);
if (ABS(temp) < 1.e-50) {
  fprintf(stderr,"***ERROR: MatInv3D: matrix non-invertible.\n");
  exit(1);
}
/* Inverse matrix: classical adjoint divided by determinant. */
for (col=0; col<3; ++col) {
  for (row=0; row<3; ++row) {
    mminv[col][row] = aa[col][row]  / temp;
  }
}
return;
}

/* ---------------------------------------------------------------------
  Set rotation matrices (RM[0].*, see phomax.h) to convert x,y,z on unit sphere
  to primed coordinate frame and vice versa.
  The prime system is set up such that the x-prime axis points in the direction
  of the RA,DEC origin, and the z-prime axis lies in the plane of the x-prime
  axis and the original z axis.

  Input: alpha,theta  : Origin on unit sphere, alpha is RA and theta is DEC
                          in RADIANS.  Use xyz2radec to compute from x,y,z.
 Output: RM           : 3x3 and 2x2 rotation matrices (structure).
*/
/*@@*/
void SetRotationMatrix( double alpha, double theta, RMtype RM[1] )
{
/**/
const double pi         =           3.141592654;
const double halfpiclip = (pi/2.) - 0.000000005; /* decrease by milli-arcsec */
/**/
/* Clip DEC, just in case. */
theta = MIN( theta , halfpiclip        );
theta = MAX( theta , halfpiclip * -1.0 );
/* Set up forward matrix: x,y,z to xp,yp,zp. */
RM[0].mat[0][0] =     cos(theta) * cos(alpha);
RM[0].mat[0][1] =-1.*              sin(alpha);
RM[0].mat[0][2] =-1.* sin(theta) * cos(alpha);
RM[0].mat[1][0] =     cos(theta) * sin(alpha);
RM[0].mat[1][1] =                  cos(alpha);
RM[0].mat[1][2] =-1.* sin(theta) * sin(alpha);
RM[0].mat[2][0] =     sin(theta);
RM[0].mat[2][1] = 0.;
RM[0].mat[2][2] =     cos(theta);
/* Invert matrix. */
MatInv3D(RM[0].mat,RM[0].inv);
return;
}




/* ----------------------------------------------------------------------
  Remove a file.
*/
/*@@*/
void misc_delete_file( char wrd[] )
{
/**/
char wrd2[2000];
/**/
sprintf(wrd2,"/bin/rm -f %s",wrd);
system(wrd2);
return;
}

/* ----------------------------------------------------------------------
  Take an input from terminal to create a pause.
*/
/*@@*/
void cpauseit()
{
  char dummy[3];
  printf("Hit return.");
  fgets(dummy,sizeof(dummy),stdin);
  return;
}

/* ----------------------------------------------------------------------
  Execute system command and print it out first.
*/
/*@@*/
void misc_system( char wrd[] )
{
printf("SYSTEM: %s \n",wrd);
system(wrd);
return;
}


/* ----------------------------------------------------------------------
 Bin data using whole pixels.  Ignores negative pixels in mode 1.

 Input:  bin : binning factor (>1).
        mode : binning mode: 0 = normal averaging
                             1 = averaging error array.
                             2 = summing.

In/Out:   np : number of points in array.
        xx[] : array of values to be binned.
*/
/*@@*/
void cDoBin4(int bin, int *np, float xx[], int mode)
{
/**/
int nn,ii,pix,ok;
float xt;
/**/

/* Normal, averaging rebinning. */
if (mode == 0) {
  nn =-1;
  pix= 0;
  while((pix+bin-1) < *np) {
    xt=0.;
    for (ii=pix; ii<=(pix+bin-1); ++ii) {
      xt = xt + xx[ii];
    }
    ++nn;
    xx[nn]= xt / bin;
    pix = pix + bin;
  }
  *np = nn;

/* Error array rebinning. */
} else {
if (mode == 1) {
  nn =-1;
  pix= 0;
  while((pix+bin-1) < *np) {
    ok=1;
    xt=0.;
    for (ii=pix; ii<=(pix+bin-1); ++ii) {
      xt = xt + (xx[ii] * xx[ii]);
      if (xx[ii] < 0.) ok=0;
    }
    ++nn;
    if (ok == 1) {
      xx[nn] = sqrt( xt ) / bin;
    } else {
      xx[nn] = -1.;
    }
    pix = pix + bin;
  }
  *np = nn;

/* Summing. */
} else {
  nn =-1;
  pix= 0;
  while((pix+bin-1) < *np) {
    xt=0.;
    for (ii=pix; ii<=(pix+bin-1); ++ii) {
      xt = xt + xx[ii];
    }
    ++nn;
    xx[nn]= xt;
    pix = pix + bin;
  }
  *np = nn;

}}
return;
}


/* ----------------------------------------------------------------------
  Get seconds since 1970 (UTC).
*/
/*@@*/
int gmtimeget()
{
/**/
int ii;
time_t tp;
/**/
gmtime(&tp);
ii = tp;
return(ii);
}



/* ----------------------------------------------------------------------
  Get a time string.
*/
/*@@*/
void timegetstring(char uline[])
{
time_t tp;
struct tm *ptr;
time(&tp);
ptr = localtime(&tp);
strftime(uline,25,"%Y-%m-%dT%H:%M:%S", ptr);
return;
}

/* ----------------------------------------------------------------------
  Get a time string in GMT.  Form: yyyy-mm-ddTHH:MM:SS
*/
/*@@*/
void timegetstring_gmt(char uline[])
{
time_t tp;
struct tm *ptr;
time(&tp);
ptr = gmtime(&tp);
strftime(uline,25,"%Y-%m-%dT%H:%M:%S", ptr);
return;
}


/* ----------------------------------------------------------------------
  Returns truncated integer.
*/
/*@@*/
int cint( double rr )
{
int ii;
ii = rr;
return(ii);
}

/* ----------------------------------------------------------------------
  Returns nearest integer.
*/
/*@@*/
int cnint( double rr )
{
int ii;
if (rr > 0.) {
  ii = rr + 0.5;
} else {
  ii = rr - 0.5;
}
return(ii);
}


/* ----------------------------------------------------------------------
  Write a double precision value into a 9 characters in >=10 character string.
  Input:  value : value to write to string.
  Output:    ss : string, must be at least 10 characters long.
  (Like rf9 in misc.f.)
*/
/*@@*/
void WriteString9(double value, char ss[])
{
if (ABS((value)) < 0.0001) { sprintf(ss,"%9.2e",value);
} else {
if ((value < 99.999999)&&(value > -9.999999)) { sprintf(ss,"%9.6f",value);
} else {
if ((value < 999.99999)&&(value > -99.99999)) { sprintf(ss,"%9.5f",value);
} else {
if ((value < 9999.9999)&&(value > -999.9999)) { sprintf(ss,"%9.4f",value);
} else {
if ((value < 99999.999)&&(value > -9999.999)) { sprintf(ss,"%9.3f",value);
} else {
if ((value < 999999.99)&&(value > -99999.99)) { sprintf(ss,"%9.2f",value);
} else {
if ((value < 9999999.9)&&(value > -999999.9)) { sprintf(ss,"%9.1f",value);
} else { sprintf(ss,"%9.2e",value);
}}}}}}}
ss[9]='\0';
return;
}


/* ----------------------------------------------------------------------
  Write a double precision value into a 7 characters in >=8 character string.
  Input:  value : value to write to string.
  Output:    ss : string, must be at least 8 characters long.
  (Like rf7 in misc.f.)
*/
/*@@*/
void WriteString7(double value, char ss[])
{
if (ABS((value)) < 1.e-20) { strcpy(ss,"0.");
} else {
if (ABS((value)) < 0.001) { sprintf(ss,"%7.0e",value);
} else {
if ((value < 99.9999)&&(value > -9.9999)) { sprintf(ss,"%7.4f",value);
} else {
if ((value < 999.999)&&(value > -99.999)) { sprintf(ss,"%7.3f",value);
} else {
if ((value < 9999.99)&&(value > -999.99)) { sprintf(ss,"%7.2f",value);
} else {
if ((value < 99999.9)&&(value > -9999.9)) { sprintf(ss,"%7.1f",value);
} else {
if ((value < 999999.)&&(value > -99999.)) { sprintf(ss,"%7.0f",value);
} else { sprintf(ss,"%7.0e",value);
}}}}}}}
ss[7]='\0';
return;
}



/* ---------------------------------------------------------------------- */
/* Finds the first occurence of a non-blank character in string "a".
   The first character position is 0.
   Returns -1 if string is completely blank.
*/
/*@@*/
int cfc(char a[])
{
int ii,ia;
ia = strlen(a);
ii=0;
while ((ii <= ia)&&(a[ii] == ' ')) { ++ii; }
if (ii == ia) ii=-1;
return(ii);
}

/* ----------------------------------------------------------------------
   Finds the last occurence of a non-blank, non-carraige return
   character in string "a".
   The first character position is 0.
   Returns -1 if string is completely blank.
*/
/*@@*/
int clc(char a[])
{
int ii,ia;
ia = strlen(a);
ii=ia-1;
while ( (ii >= 0) && ((a[ii] == ' ')||(a[ii] == '\012')||(a[ii] == '\013')) ) { --ii; }
return(ii);
}


/* ----------------------------------------------------------------------
   Prefixes the string 'b' in the front of the string 'a'.
   This is essentially the opposite of 'strcat()'.
   This will REMOVE or TRUNCATE any trailing blanks in either string.
*/
/*@@*/
void strprefix( char a[], char b[] )
{
int ii,jj,kk;
jj = clc( a );
kk = clc( b );
for (ii=jj; ii>=0; --ii) { a[(ii+kk+1)] = a[ii]; }
a[(jj+kk+2)]='\0';
for (ii=0; ii<=kk; ++ii) { a[ii] = b[ii]; }
return;
}

/* ----------------------------------------------------------------------
   "Prefix With Blanks" --  this version will INCLUDE all trailing blanks.
   Prefixes the string 'b' in the front of the string 'a'.
   This is essentially the opposite of 'strcat()'.
*/
/*@@*/
void strprefix_wb( char a[], char b[] )
{
int ii,jj,kk;
jj = ceos( a ) - 1;
kk = ceos( b ) - 1;
for (ii=jj; ii>=0; --ii) { a[(ii+kk+1)] = a[ii]; }
a[(jj+kk+2)]='\0';
for (ii=0; ii<=kk; ++ii) { a[ii] = b[ii]; }
return;
}


/* ----------------------------------------------------------------------
   Copies characters a1 to a2 in string "a" into the string "b" starting
   at the position "b1".  First character is number 0.
*/
/*@@*/
void substrcpy( char a[], int a1, int a2, char b[], int b1 )
{
int ii;
for (ii=a1; ii<=a2; ++ii) { b[ b1 + (ii-a1) ] = a[ ii ]; }
return;
}

/* ----------------------------------------------------------------------
   Just like substrcpy, except add a terminating NULL character.
*/
/*@@*/
void substrcpy_terminate( char a[], int a1, int a2, char b[], int b1 )
{
substrcpy(a,a1,a2,b,b1);  b[(b1+1+a2-a1)] = '\0';
return;
}

/* ----------------------------------------------------------------------
  Read a line from a file.
  Places end-of-string character (NULL) after last
  non-blank,non-carraige-return  character.
  Return 0 if EOF, 1 if ok.
*/
/*@@*/
int fgetline(char userline[], FILE *fu)
{
  char *ptr;
  char line[19000];
  int ii;
  ptr = fgets(line,sizeof(line),fu);
  line[strlen(line)] = '\0';
  ii = clc(line);
  line[ii+1] = '\0';
  strcpy(userline,line);
  if (ptr == '\0') { return(0); } else { return(1); }
}

/* ----------------------------------------------------------------------
  Return "1" if the character is a numeral 0-9 (else return "0").
*/
/*@@*/
int numeral(char cc)
{
if (cc == '0') return(1);
if (cc == '1') return(1);
if (cc == '2') return(1);
if (cc == '3') return(1);
if (cc == '4') return(1);
if (cc == '5') return(1);
if (cc == '6') return(1);
if (cc == '7') return(1);
if (cc == '8') return(1);
if (cc == '9') return(1);
return(0);
}


/* ---------------------------------------------------------------------- */
/* Compute the angular separation in degrees between two ra and dec
   positions.  ra and dec are in decimal degrees.
*/
/*@@*/
double cangsep(double ra1, double dec1, double ra2, double dec2)
{
/**/
  double a1,t1,a2,t2,x1,y1,z1,x2,y2,z2,rr;
  const double pi = 3.141592654;
  const double pio180 = pi/180.;   /* PI over 180 degrees. */
/**/
/* Set alpha and theta (in radians). */
  a1 = ra1 * pio180;
  t1 = ( 90. - dec1 ) * pio180;
  a2 = ra2 * pio180;
  t2 = ( 90. - dec2 ) * pio180;
/* Computations. */
  x1 = sin(t1) * cos(a1);
  y1 = sin(t1) * sin(a1);
  z1 = cos(t1);
  x2 = sin(t2) * cos(a2);
  y2 = sin(t2) * sin(a2);
  z2 = cos(t2);
  rr = (x1*x2) + (y1*y2) + (z1*z2);
  if ((1. - rr) > 1.e-20) { rr = acos( rr ); } else { rr = 0.; }
  rr = ( rr / pio180 );
  return(rr);
}


/* ----------------------------------------------------------------------
  Open a file for reading and exit if problem opening file.
*/
/*@@*/
FILE *fopen_read( char file[] )
{
FILE *infu;
infu = fopen(file,"r");
if (infu == NULL) {
  fprintf(stderr,"***ERROR:fopen_read: opening file for reading: %s \n",file);
  exit(1);
}
return(infu);
}


/* ----------------------------------------------------------------------
  Open a file for writing and exit with error if cannot create file.
*/
/*@@*/
FILE *fopen_write( char file[] )
{
FILE *outfu;
outfu = fopen(file,"w");
if (outfu == NULL) {
  fprintf(stderr,"***ERROR:fopen_write: opening file for writing: %s \n",file);
  exit(1);
}
return(outfu);
}


/* ----------------------------------------------------------------------
  Does a file exist?  If so return "1".
*/
/*@@*/
int FileExist( char filename[] )
{
FILE *fu;
int ii;
fu = fopen(filename,"r");
if (fu == NULL) { ii=0; } else { ii=1; fclose(fu); }
return(ii);
}



/* ----------------------------------------------------------------------
  Return "1" or "2" if the character is a numeric: 0-9, ".",
     "E", "e", "+", or "-".   Returns 0 otherwise.
*/
/*@@*/
int numeric(char cc)
{
if (numeral(cc) == 1) return(1);
if (cc == '.') return(2);
if (cc == '-') return(2);
if (cc == '+') return(2);
if (cc == 'E') return(2);
if (cc == 'e') return(2);
return(0);
}


/* ----------------------------------------------------------------------
  Return "1" if all the characters in a string are numeric.
*/
/*@@*/
int numeric_string(char wrd[])
{
int ok,kk,ii;
kk = clc(wrd);
ok = 1;
for (ii=0; ii<=kk; ++ii) {
  if (numeric(wrd[ii]) == 0) ok = 0;
}
return(ok);
}

/* ----------------------------------------------------------------------
  Return "1" if character is NOT a deliminator: space(), comma(,), or pipe(|).
    Input: cc     : character.
           mode   : 0= normal (space,comma,pipe).
                    1= pipe only (pipe).
*/
/*@@*/
int nondelim( char cc, int mode )
{
if (mode == 0) {
  if (cc == ' ') return(0);
  if (cc == ',') return(0);
  if (cc == '|') return(0);
} else {
  if (cc == '|') return(0);
}
return(1);
}


/* ----------------------------------------------------------------------
  Converts a string to lower case.
*/
/*@@*/
void clower_case(char a[])
{
/**/
int ii,ia;
char aa;
/**/
ia = strlen(a);
for (ii=0; ii<ia; ++ii) { aa = a[ii];
  if (aa == 'A') aa='a';
  if (aa == 'B') aa='b';
  if (aa == 'C') aa='c';
  if (aa == 'D') aa='d';
  if (aa == 'E') aa='e';
  if (aa == 'F') aa='f';
  if (aa == 'G') aa='g';
  if (aa == 'H') aa='h';
  if (aa == 'I') aa='i';
  if (aa == 'J') aa='j';
  if (aa == 'K') aa='k';
  if (aa == 'L') aa='l';
  if (aa == 'M') aa='m';
  if (aa == 'N') aa='n';
  if (aa == 'O') aa='o';
  if (aa == 'P') aa='p';
  if (aa == 'Q') aa='q';
  if (aa == 'R') aa='r';
  if (aa == 'S') aa='s';
  if (aa == 'T') aa='t';
  if (aa == 'U') aa='u';
  if (aa == 'V') aa='v';
  if (aa == 'W') aa='w';
  if (aa == 'X') aa='x';
  if (aa == 'Y') aa='y';
  if (aa == 'Z') aa='z';  a[ii] = aa;
}
return;
}

/* ----------------------------------------------------------------------
  Converts a string to upper case.
*/
/*@@*/
void cupper_case(char a[])
{
/**/
int ii,ia;
char aa;
/**/
ia = strlen(a);
for (ii=0; ii<ia; ++ii) { aa = a[ii];
  if (aa == 'a') aa='A';
  if (aa == 'b') aa='B';
  if (aa == 'c') aa='C';
  if (aa == 'd') aa='D';
  if (aa == 'e') aa='E';
  if (aa == 'f') aa='F';
  if (aa == 'g') aa='G';
  if (aa == 'h') aa='H';
  if (aa == 'i') aa='I';
  if (aa == 'j') aa='J';
  if (aa == 'k') aa='K';
  if (aa == 'l') aa='L';
  if (aa == 'm') aa='M';
  if (aa == 'n') aa='N';
  if (aa == 'o') aa='O';
  if (aa == 'p') aa='P';
  if (aa == 'q') aa='Q';
  if (aa == 'r') aa='R';
  if (aa == 's') aa='S';
  if (aa == 't') aa='T';
  if (aa == 'u') aa='U';
  if (aa == 'v') aa='V';
  if (aa == 'w') aa='W';
  if (aa == 'x') aa='X';
  if (aa == 'y') aa='Y';
  if (aa == 'z') aa='Z';  a[ii] = aa;
}
return;
}


/* ----------------------------------------------------------------------
   Finds the first occurence of substring "b" in string "a" going in the
   reverse direction.  Returns this position of the substring within "a",
   or will return "-1" if the substring is not found.
   The first character is number 0.
*/
/*@@*/
int cindex_reverse(char a[], char b[] )
{
/**/
int ii,jj,ia,ib,pos;
/**/
/* Get and check lengths of strings. */
ia = strlen(a);
ib = strlen(b);
/* Return immediately if string a is too small or either string empty. */
if (ia < ib) { return(-1); }
if (ia < 1)  { return(-1); }
if (ib < 1)  { return(-1); }
/* Find substring. */
pos= -1;
ii = ia - ib;
while (ii >= 0) {
  pos=ii;
  for (jj=0; jj<ib; ++jj) { if (a[ii+jj] != b[jj]) pos = -1; }
  if (pos > -1)  break;
  --ii;
}
return(pos);
}

/* ----------------------------------------------------------------------
   Finds the first occurence of substring "b" in string "a" at or after
   position "ptr".  Returns this position of the substring within "a", or
   will return "-1" if the substring is not found. First character is number 0.
*/
/*@@*/
int cindex2(char a[], char b[], int ptr )
{
/**/
int ii,jj,ia,ib,pos;
/**/
/* Get and check lengths of strings. */
ia = strlen(a);
ib = strlen(b);
/* Return immediately if string a is too small or either string empty. */
if (ia < ib) { return(-1); }
if (ia < ptr+1) { return(-1); }
if (ia < 1)  { return(-1); }
if (ib < 1)  { return(-1); }
/* Find substring. */
pos=-1;
ii=ptr;
while (ii <= (ia-ib)) {
  pos=ii;
  for (jj=0; jj<ib; ++jj) { if (a[ii+jj] != b[jj]) pos = -1; }
  if (pos > -1)  break;
  ++ii;
}
return(pos);
}


/* ----------------------------------------------------------------------
   Finds the first occurence of substring "b" in string "a".
   Returns the position of the substring within "a", or will return "-1"
   if the substring is not found.  First character is number 0.
*/
/*@@*/
int cindex(char a[], char b[] )
{
/**/
int ii,jj,ia,ib,pos;
/**/
/* Get and check lengths of strings. */
ia = strlen(a);
ib = strlen(b);
/* Return immediately if string a is too small or either string empty. */
if (ia < ib) { return(-1); }
if (ia < 1)  { return(-1); }
if (ib < 1)  { return(-1); }
/* Find substring. */
pos=-1;
ii=0;
while (ii <= (ia-ib)) {
  pos=ii;
  for (jj=0; jj<ib; ++jj) { if (a[ii+jj] != b[jj]) pos = -1; }
  if (pos > -1)  break;
  ++ii;
}
return(pos);
}


/* ----------------------------------------------------------------------
  Returns '1' (true) is a string 'a' contains the given substring 'b'. 
*/
/*@@*/
int sck(char a[], char b[])
{
/**/
int ii;
/**/
if (cindex(a,b) == -1) { ii=0; } else { ii=1; }
return(ii);
}


/* ----------------------------------------------------------------------
   Compares the first "n" characters of two strings.  If equal return "0",
   if not, return "1".  (same convention as "strcmp").  */
/*@@*/
int substrcmp0(char a[], char b[], int n )
{
int ii,ok;
if (strlen(a) < n) return(1);
if (strlen(b) < n) return(1);
ok=0;
for (ii=0; ii<n; ++ii) { if (a[ii] != b[ii]) ok = 1; }
return(ok);
}


/* ----------------------------------------------------------------------
  Convert a string segment into a double precision number.
  The string segment is defined by up1 and up2.
  (Here up1 must be >= 1, it cannot be 0.  In other words, the routine
   thinks that the string starts with number 1 not number 0 (as usual).)
  Returns "-1.e+40" if there is an error.
*/
/*@@*/
double cvalread(char userline[], int up1, int up2)
{
/**/
  int eos,dp,i1,i2,ii,jj,kk,ee,jjj,p1,p2,ok;
  double neg,sum,r1,r2,exponent;
  char cc[2];
  char line[1000];
/**/

/* Maximum string length. */
  kk = strlen(userline)-1;

/* Limit pointers. Convert to zero origin. */
  p1 = up1-1;  if (p1<0) p1=0;  if (p1>kk) p1=kk;
  p2 = up2-1;  if (p2<0) p2=0;  if (p2>kk) p2=kk;

/* Load line string. */
  jj=0;
  for (ii=p1; ii<=p2; ++ii) {
    line[jj] = userline[ii];
    ++jj;
  }
  line[jj] = '\0';

/* Last real character of string. */
  eos = strlen(line)-1;

/* Default. */
   exponent = 1.;

/* Look for exponent. */
  ee=0;
  while ( (line[ee] != 'E') && (line[ee] != 'e') ) {
    ++ee;
    if (ee>eos) break;
  }

/* Exponent exists? */
  if (ee <= eos) {

/* First numeral. */
    ii=ee+1;

/* End of string? */
    if (ii <= eos) {

/* The sign. */
      neg = +1.;
      if (line[ii] == '+') { ++ii; }
        else { if (line[ii] == '-') { ++ii; neg = -1.; } }

/* End of string? */
      if (ii <= eos) {

/* Find next non-numeral. */
        jj=ii;
        while ( (jj <= eos) && (numeral(line[jj]) == 1) ) ++jj;
        jj=jj-1;
/* Numbers? */
        if (jj >= ii ) {
          sum = 0.;
          for (kk=jj; kk>=ii; kk=kk-1) {
            cc[0] = line[kk];
            cc[1] = '\0';
            sscanf(cc,"%d",&jjj);
            r1 = jjj;
            r2 = jj-kk;
            sum= sum + ( r1 * pow(10.,r2) );
          }
          if (neg > 0.) { exponent = pow(10.,sum); }
                   else { exponent = 1./pow(10.,sum); }
        }
      }
    }
    line[ee] = '\0';  /* Cut off string at "E" or "e". */
  }

/* Last real character of string. */
  eos = strlen(line)-1;

/* See if any numerals exist in string. */
  ok=0;
  for (ii=0; ii<strlen(line); ++ii) { if (numeral(line[ii]) == 1) ok=1; }
  if (ok == 0) {
    return(-1.e+40);     /* ERROR: cvalread: reading string. */
  }

/* Initialize. */
  neg = +1.;

/* Find decimal point (or end of string). */
  dp=0;
  while ((line[dp] != '.')&&(line[dp] != '\0')) ++dp;
/* If no decimal point, find first numeral, then find next non-numeral. */
  if (line[dp] == '\0') {
    ii=0;
    while ( numeral(line[ii]) == 0 ) ++ii;
    while ( numeral(line[ii]) == 1 ) ++ii;
    dp = ii;
  }

/* Go left from decimal point to find first non-numeral. */
  i1=dp-1;
  while ( (i1>=0) && (numeral(line[i1]) == 1) ) i1=i1-1;
/* Set negative if non-numeral character is "-" */
  if (i1>=0) { if (line[i1] == '-') neg = -1.; };
/* Offset from non-numeral character. */
  i1=i1+1;
/* Go right from decimal point to find first non-numeral. */
  i2=dp+1;
  while ( (i2<=eos) && (numeral(line[i2]) == 1) ) i2=i2+1;
  i2=i2-1;

/* Left side of decimal point. */
  sum=0.;
  for (ii=dp-1; ii>=i1; ii=ii-1) {
    cc[0] = line[ii];
    cc[1] = '\0';
    sscanf(cc,"%d",&jj);
    r1 = jj;
    r2 = (dp-ii)-1.;
    sum = sum + ( r1 * pow(10.,r2) );
  }

/* Right side of decimal point. */
  for (ii=dp+1; ii<=i2; ii=ii+1) {
    cc[0] = line[ii];
    cc[1] = '\0';
    sscanf(cc,"%d",&jj);
    r1 = jj;
    r2 = -1.*(ii-dp);
    sum = sum + ( r1 * pow(10.,r2) );
  }
  return(neg*sum*exponent);
}



/* ----------------------------------------------------------------------
  Call cvalread routine, but with string starting with column 0 rather
  than 1.
*/
/*@@*/
double cvalread0( char userline[], int up1, int up2 )
{
double rr;
rr = cvalread( userline, up1+1, up2+1 );
return(rr);
}




/* ----------------------------------------------------------------------
  Find the starting and ending positions of the next word from a line,
  starting from character position "pp".  In this case a word is deliminated
  by a space( ), comma(,), or pipe(|).

   Input:  line[]  : whole string.
           pp      : consider only string from this point forward.
                   : use "0" for whole string.
   Output: sow     : start of word position.
           eow     : end of word position.
           ok      : 1 if successful, 0 if error.
    Input: mode    : 0= normal (space,comma,pipe).
                     1= pipe only (pipe).
*/
/*@@*/
void FindWord( char line[], int pp, int *sow, int *eow, int *ok, int mode )
{
/**/
int ii,kk,eos,repeat,nondelimflag;
/**/

/* Defaults. */
*sow=-1;
*eow=-1;

/* End of string. */
eos = strlen(line)-1;
if (pp>eos) {
  *ok=0;
  return;
}

/* At end of string and last character is a deliminator. */
if ((nondelim(line[pp],mode) == 0)&&(pp == eos)) {
  *ok=0;
  return;
}

/* Special case, where there are two adjacent pipe deliminators. */
if (mode == 1) {
  if ((nondelim(line[pp],mode) == 0)&&(nondelim(line[pp+1],mode) == 0)) {
    *ok=1;
    *sow = pp + 1;   /* sow > eow tells caller that it is a null string */
    *eow = pp;
    return;
  }
}

/* Starting point. */
ii=pp;
/* Find next non-deliminator character. */
repeat=1;
while (repeat == 1) {
  nondelimflag = 0;
  while (ii <= eos) {
    kk = nondelim(line[ii],mode);
    if (kk == 1) nondelimflag=1;
    if (kk > 0) break;
    ++ii;
  }
  if (ii > eos ) {
    *ok=0;
    return;
  }
  *sow = ii;
/* Find next deliminator character. */
  ++ii;
  while (ii <= eos) {
    kk = nondelim(line[ii],mode);
    if (kk == 1) nondelimflag = 1;
    if (kk == 0) break;
    ++ii;
  }
/* Do not repeat if a nondelim was found. */
  if (nondelimflag == 1) repeat=0;
}
/* Back up one position, unless beyond end-of-string. */
if (ii > eos ) ii = eos;  else ii=ii-1;
*eow = ii;
*ok=1;
return;
}



/* ----------------------------------------------------------------------
  Find the starting and ending positions of the next numeric word from a line,
  starting from character position "pp".  In this case a numeric word
  (0-9,.,E,e,+,-) must contain at least one numeral (0-9).

   Input:  line[]      (whole string.)
           pp          (consider only string from this point forward.)
                       (use "0" for whole string.)
   Output: sow         (start of word position.)
           eow         (end of word position.)
           ok          (1 if successful, 0 if error.)
*/
/*@@*/
void FindNumericWord(char line[], int pp, int *sow, int *eow, int *ok)
{
/**/
int ii,kk,eos,repeat,numeralflag;
/**/
/* Defaults. */
*sow=-1;
*eow=-1;
/* End of string. */
eos = strlen(line)-1;
if (pp>eos) {
  *ok=0;
  return;
}
/* Starting point. */
ii=pp;
/* Find next numeric character. */
repeat=1;
while (repeat == 1) {
  numeralflag = 0;
  while (ii <= eos) {
    kk = numeric(line[ii]);
    if (kk == 1) numeralflag=1;
    if (kk > 0) break;
    ++ii;
  }
  if (ii > eos ) {
    *ok=0;
    return;
  }
  *sow = ii;
/* Find next non-numeric character. */
  ++ii;
  while (ii <= eos) {
    kk = numeric(line[ii]);
    if (kk == 1) numeralflag = 1;
    if (kk == 0) break;
    ++ii;
  }
/* Do not repeat if a numeral (0-9) was found. */
  if (numeralflag == 1) repeat=0;
}
/* Back up one position, unless beyond end-of-string. */
if (ii > eos ) ii = eos;  else ii=ii-1;
*eow = ii;
*ok=1;
return;
}



/* ----------------------------------------------------------------------
  Read the nn'th number from a line string.
  Returns "-1.e+40" if there is a problem.
*/
/*@@*/
double GetLineValue(char line[], int nn)
{
/**/
int sow,eow,ii,jj,ok;
double rr;
char wrd[10000];
/**/
/* Fix NANs. */
strcpy(wrd,line);
cupper_case(wrd);
while( cindex(wrd,"NAN") > -1) {
  ii = cindex(wrd,"NAN");
  jj = clc(wrd);
  wrd[ii+0] = '0';
  if (ii+1 <= jj) wrd[ii+1] = ' ';
  if (ii+2 <= jj) wrd[ii+2] = ' ';
}
/* Get bounds of nn'th number. */
eow=-1;
for (ii=0; ii<nn; ++ii) {
  FindNumericWord(wrd,eow+1,&sow,&eow,&ok);
  if (ok == 0) {
    printf("***error: GetLineValue: getting value number %d.\n",nn);
    return(-1.e+40);
  }
}
/* Read value from "wrd". */
rr = cvalread(wrd,sow+1,eow+1);
return(rr);
}


/* ----------------------------------------------------------------------
  Exits on Error...
  Read the nn'th number from a line string.
  Aborts if there is a problem.
*/
/*@@*/
double GetLineValueError(char line[], int nn)
{
/**/
int sow,eow,ii,jj,ok;
double rr;
char wrd[10000];
/**/
/* Fix NANs. */
strcpy(wrd,line);
cupper_case(wrd);
while( cindex(wrd,"NAN") > -1) {
  ii = cindex(wrd,"NAN");
  jj = clc(wrd);
  wrd[ii+0] = '0';
  if (ii+1 <= jj) wrd[ii+1] = ' ';
  if (ii+2 <= jj) wrd[ii+2] = ' ';
}
/* Get bounds of nn'th number. */
eow=-1;
for (ii=0; ii<nn; ++ii) {
  FindNumericWord(wrd,eow+1,&sow,&eow,&ok);
  if (ok == 0) {
    printf("#@# sow=%d  eow=%d  wrd='%s'\n",sow,eow,wrd);
    fprintf(stderr,"***ERROR: GetLineValueError: getting value number %d.\n",nn);
    fprintf(stderr,"***ERROR: ABORT: line='%s'\n",line);
    exit(1);
  }
}
/* Read value from "wrd". */
rr = cvalread(wrd,sow+1,eow+1);
return(rr);
}

/* ----------------------------------------------------------------------
  Read the nn'th number from a line string.
  No error messages will be printed.
  Returns "-1.e+40" if there is a problem.
*/
/*@@*/
double GetLineValueQuiet(char line[], int nn)
{
/**/
int sow,eow,ii,jj,ok;
double rr;
char wrd[10000];
/**/
/* Fix NANs. */
strcpy(wrd,line);
cupper_case(wrd);
while( cindex(wrd,"NAN") > -1) {
  ii = cindex(wrd,"NAN");
  jj = clc(wrd);
  wrd[ii+0] = '0';
  if (ii+1 <= jj) wrd[ii+1] = ' ';
  if (ii+2 <= jj) wrd[ii+2] = ' ';
}
/* Get bounds of nn'th number. */
eow=-1;
for (ii=0; ii<nn; ++ii) {
  FindNumericWord(wrd,eow+1,&sow,&eow,&ok);
  if (ok == 0) { return(-1.e+40); }
}
/* Read value from "wrd". */
rr = cvalread(wrd,sow+1,eow+1);
return(rr);
}


/* ----------------------------------------------------------------------
  Calls GetLineValueQuiet()...
  Returns "-1.e+40" if there is a problem.
*/
/*@@*/
double GLVQ(char line[], int nn)
{
return( GetLineValueQuiet(line,nn) );
}

/* ----------------------------------------------------------------------
  Calls GetLineValueError()...
  Returns "-1.e+40" if there is a problem AND it EXITs.
*/
/*@@*/
double GLVE(char line[], int nn)
{
return( GetLineValueError(line,nn) );
}

/* ----------------------------------------------------------------------
  Calls GetLineValue()...
  Returns "-1.e+40" if there is a problem.
*/
/*@@*/
double GLV(char line[], int nn)
{
return( GetLineValue(line,nn) );
}



/* ---------------------------------------------------------------------- */
/* Finds an argument "parm" in argument series and extracts the word
   which follows the "parm" string and loads it into "wrd".  It also decrements
   narg by 1, and reloads the arg[][] array to delete the found entry.
   This function will "delete and load" only the first matching argument.
   Returns "1" if the argument was found, otherwise "0".  If "d" is "s",
   then this is a switch and the parm must match the argument exactly.
     Input:  arg[][], narg, parm, d.
     Output: narg,wrd,n.  (n is always 1).
*/
/*@@*/
int cfindarg(char arg[202][200], int *narg, char parm[], char d, char wrd[] )
{
/**/
int ii,jj,kk,ok,parmlen,arglen;
/**/
char pt[200],ptt[200];
/**/

/* Initialize. */
parmlen = strlen(parm);
ok=0;
ii=1;
strcpy(wrd,"");

/* Switch only. */
if (d == 's') {

/* Find the parameter. Exact match. */
  while ( (ii <= (*narg)) && (ok == 0) ) {
    if ( (substrcmp0(arg[ii],parm,parmlen) == 0) &&
         (strlen(parm) == strlen(arg[ii])) ) {
      ok=1;
    }
    else {
      ++ii;
    }
  }

/* Value parameter, not a switch. */
} else {

/* Find the parameter. */
  while ( (ii <= (*narg)) && (ok == 0) ) {
    if (substrcmp0(arg[ii],parm,parmlen) == 0) {
/* Load the word. */
      arglen = strlen(arg[ii]);
      for (jj=parmlen; jj<arglen; ++jj) {
        wrd[jj-parmlen] = arg[ii][jj];
      }
      wrd[arglen-parmlen] = '\0';
      ok=1;
    }
    else {
      ++ii;
    }
  }

}

/* If not found, and "parm" has trailing "=" look for switch. */
if ((ok == 0)&&(d != 's')&&(parm[parmlen - 1] == '=')) {

/* Look for -parm ... */
  ok=0;
  ii=1;
  strcpy(ptt,parm); ptt[parmlen - 1] = '\0';
  strcpy(pt,"-");   strcat(pt,ptt);
  while ( (ii <= (*narg)) && (ok == 0) ) {
    if ( (substrcmp0(arg[ii],pt,parmlen) == 0) &&
         (strlen(pt) == strlen(arg[ii])) ) {
      ok=1;
    }
    else {
      ++ii;
    }
  }

/* If found, load next argument as word. */
  if (ok == 1) {
    if (*narg > ii) { strcpy(wrd,arg[ii+1]); } else { strcpy(wrd,""); }
/* Delete the "-parm" argument. */
    for (kk=ii; kk<(*narg); ++kk) { strcpy(arg[kk],arg[kk+1]); }
    *narg = *narg - 1;
/* And the next word (if exists). */
    if (*narg >= ii) {
      for (kk=ii; kk<(*narg); ++kk) { strcpy(arg[kk],arg[kk+1]); }
      *narg = *narg - 1;
    }
  }

} else {

/* Reload the arg[] array (delete the parameter). */
  if (ok == 1) {
    for (kk=ii; kk<(*narg); ++kk) { strcpy(arg[kk],arg[kk+1]); }
    *narg = *narg - 1;
  }

}

return(ok);
}



/* ---------------------------------------------------------------------- */
/* Copy the argv[] strings into char arg[202][200]. */
/*@@*/
void cargcopy(char *argv[], int argc, int *narg, char arg[202][200])
{
  int ii;
  *narg = argc - 1;
  for (ii=0; ii<=(*narg); ++ii) {
    strcpy(&arg[ii][0],&argv[ii][0]);
  }
return;
}


/* ----------------------------------------------------------------------
  Parse comma or space separated words in a string.
  Allocate wrds like this: for (ii=0; ii<maxwrds; ++ii) { wrds[ii] = (char *)calloc(maxlen,1); }
  where 'maxlen' is the maximum length for an individual word.
*/
/*@@*/
void misc_word_parse( char uline[], char line[], int maxwrds, char *wrds[maxwrds], int *nwrds )
{
/**/
int ii,ii1,ii2,nn,ptr;
/**/
/* Copy string, erase commas. */
strcpy(line,uline);
ii=1; while(ii > 0) { ii=cindex(line,","); line[ii]=' '; }
/* Separate. */
*nwrds = 0;
ii1 = cfc(line);
ii2 = clc(line);
if ((ii1 < 0)||(ii2 < 0)) { return; }
ptr = -1;
nn = 0;
for (ii=ii1; ii<=ii2; ++ii) {
  if (line[ii] == ' ') {
    if (ptr > -1) {
      substrcpy_terminate(line,ptr,ii-1,wrds[nn],0);
      ++nn;
      ptr = -1;
      if (nn > maxwrds-2) { printf("***error:mwp: too many words (max=%d).\n",maxwrds); exit(1); }
    }
  } else {
    if (ptr < 0) ptr = ii;
  }
}
/* Last word. */
if (ptr > -1) {
  substrcpy_terminate(line,ptr,ii2,wrds[nn],0);
  ++nn;
}
*nwrds = nn;
return;
}


/* ----------------------------------------------------------------------
  Range of an array.
*/
/*@@*/
void misc_ranges( int nn, double aa[], double *lo, double *hi )
{
int ii;
*lo = aa[0];
*hi = aa[0];
for (ii=1; ii<nn; ++ii) {
  if (aa[ii] < *lo) *lo = aa[ii];
  if (aa[ii] > *hi) *hi = aa[ii];
}
return;
}


/* ----------------------------------------------------------------------
  Get number of IPAC columns using the header line.
      Input: ipachead[]  : A string of an IPAC table header line.
  Returns number of columns.
*/
/*@@*/
int misc_get_number_of_IPAC_columns( char ipachead[] )
{
/**/
int jj,eoh;
int count = 0;
/**/
/* Check. */
if (ipachead[0] != '|') return(count);
/* Find number of columns. */
eoh = clc(ipachead);
jj = 1;
while (1) {
  while((jj <= eoh)&&(ipachead[jj] != '|')) { ++jj; }
  if (jj > eoh) break;
  ++count;
  ++jj;
}
return(count);
}

/* ----------------------------------------------------------------------
  Load in the names of all the IPAC table header columns.
  The boundaries are defined as the left pipe of a given column and
  the column one pixel left of the right pipe of a given column.  The
  line is assumed to start at column#1 (this syncs with cvalread()).

    Input: ipachead[]  : A string of an IPAC table header line.
   Output: ITC[]       : IPAC Table Columns structure array.
*/
/*@@*/
void misc_get_IPAC_columns( char ipachead[], ITCtype ITC[] )
{
/**/
int ii,jj,nn,eoh,kk1,kk2;
char wrd[200],wrd2[200];
/**/
/* Check. */
if (ipachead[0] != '|') return;
/* Load column names. */
eoh = clc(ipachead);
jj = 1;
nn = 0;
while (1) {

  ii = jj;

  while((jj <= eoh)&&(ipachead[jj] != '|')) { ++jj; }

  if (jj > eoh) break;

  jj = jj - 1;

  if (jj >= ii) {

    substrcpy_terminate(ipachead,ii,jj,wrd,0);   /* between the pipes. */ 

    kk1= cfc(wrd);

    kk2= clc(wrd);

    substrcpy_terminate(wrd,kk1,kk2,wrd2,0);  /*No leading or trailing blanks.*/

    strcpy(ITC[nn].name,wrd2);

    ITC[nn].b1 = ii;        /* where first line character is #1  (-1 +1) */
    ITC[nn].b2 = jj + 1;    /* where first line character is #1 */

  } else {

    strcpy(ITC[nn].name,"");
    ITC[nn].b1 = ii;
    ITC[nn].b2 = ii;

  }

  ++nn;

  jj = jj + 2;                         /* jump ahead of pipe. */

}

ITC[0].ncol = nn;
return;
}


/* ----------------------------------------------------------------------
  Returns column number (starting with 0) of a given named column.
  Returns -1 if not found.
  Case sensitive.
    Input:  username[]  : Name of column to find.
            ITC[]       : IPAC Table Column information structure array.
            mode        : 1=case sensitve. 0=case insensitive
*/
/*@@*/
int misc_find_IPAC_column( char username[], ITCtype ITC[], int mode )
{
/**/
int ii,kk1,kk2,ucn;
char wrd[40],wrd2[40];
/**/
/* Default. */
ucn = -1;
/* No leading or trailing blanks. */
kk1 = cfc(username);
kk2 = clc(username);
if (kk2 < 0) return(ucn);
substrcpy_terminate(username,kk1,kk2,wrd,0);
if (mode == 0) { cupper_case(wrd); }
/* Find the column. */
ii=0;
while ((ucn == -1)&&(ii < ITC[0].ncol)) {
  strcpy(wrd2,ITC[ii].name);
  if (mode == 0) { cupper_case(wrd2); }
  if (strcmp(wrd,wrd2) == 0) { ucn = ii; }
  ++ii;
}
return(ucn);
}




/* ----------------------------------------------------------------------
  Load in the names, positions, and types of all the IPAC table header columns.
    Input:  ipachead1[]  : First IPAC table column header line.
            ipachead2[]  : Second IPAC table column header line with types (optional).
   Output:  ITC[]        : IPAC table column information structure.
*/
/*@@*/
void misc_load_IPAC_columns( char ipachead1[], char ipachead2[], ITCtype ITC[] )
{
/**/
int ii;
char wrd[200];
/**/
/* Get names of columns and boundaries. */
misc_get_IPAC_columns( ipachead1, ITC );
/* Default types. */
for (ii=0; ii<ITC[0].ncol; ++ii) { ITC[ii].type = 'd'; }
/* Find types if second header exists. */
if (strcmp(ipachead2,"") != 0) {
  for (ii=0; ii<ITC[0].ncol; ++ii) {
    substrcpy_terminate(ipachead2,ITC[ii].b1,ITC[ii].b2,wrd,0);
    if (cindex(wrd,"d") > -1) { ITC[ii].type = 'd'; } else {
    if (cindex(wrd,"i") > -1) { ITC[ii].type = 'i'; } else {
    if (cindex(wrd,"c") > -1) { ITC[ii].type = 'c'; } else {
    if (cindex(wrd,"f") > -1) { ITC[ii].type = 'f'; } else {
    if (cindex(wrd,"s") > -1) { ITC[ii].type = 's'; } else {
                                ITC[ii].type = ' ';        }}}}}
  }
}
return;
}


/* ----------------------------------------------------------------------
  Create a time + pid unique string.
*/
/*@@*/
void misc_timepidstring( char timepid[] )
{
/**/
char wrd[80];
int ii;
/**/
timegetstring( wrd );
ii = getpid();
sprintf(timepid,"%s_%d",wrd,ii);
return;
}

/* ----------------------------------------------------------------------
  Create a time + pid unique string.
  Blank out hours, minutes, and seconds.
*/
/*@@*/
void misc_timepidstring2( char timepid[] )
{
/**/
char wrd[80];
int ii;
/**/
timegetstring( wrd );
ii = getpid();
sprintf(timepid,"%s_%d",wrd,ii);
timepid[11]='x';
timepid[12]='x';
timepid[14]='x';
timepid[15]='x';
timepid[17]='x';
timepid[18]='x';
return;
}

/* ----------------------------------------------------------------------
  Get a double value from a IPAC line given the name of the column and
  the column header data from: 'misc_get_IPAC_columns()'.
    Input:  line[]    : Data line from IPAC file.
            colname[] : Name of column to get value from.
            ITC[]     : IPAC Table Column information structure array.
            mode      : 1=case sensitve. 0=case insensitive
*/
/*@@*/
double misc_get_double_IPAC( char line[], char colname[], ITCtype ITC[], int mode )
{
int ii;
double rr;
ii = misc_find_IPAC_column( colname, ITC, mode );
rr = cvalread(line,ITC[ii].b1,ITC[ii].b2);
return(rr);
}


/* ----------------------------------------------------------------------
  Pad a string with blanks until end-of-word (eow).
*/
/*@@*/
void misc_pad_blanks( char wrd0[], int eow, char wrd[] )
{
/**/
int kk,ii;
/**/
strcpy(wrd,wrd0);
kk = clc(wrd) + 1;
if (kk < eow) {
  for (ii=kk; ii<eow; ++ii) { wrd[ii]=' '; }
}
wrd[eow] = '\0';
return;
}

/* ---------------------------------------------------------------------- */
/* Trim blanks preceding or following a string.
*/
/*@@*/
void misc_trim_blanks(char wrd[])
{
/**/
int ii,kk1,kk2;
/**/
kk1 = cfc(wrd);   kk2 = clc(wrd);
if (kk2 <  0) { wrd[0]       ='\0'; return; }
if (kk1 == 0) { wrd[(1+kk2)] ='\0'; return; }
for (ii=kk1; ii<=kk2; ++ii) { wrd[(ii-kk1)] = wrd[ii]; }
wrd[(1+kk2-kk1)] = '\0';
return;
}


/* ----------------------------------------------------------------------
 Append a line to a file, opening and closing the file.
*/
/*@@*/
void misc_append_to_file( char userfile[], char userline[] )
{
/**/
char wrd[200];
char tempfile[200];
char timepid[60];
char line[2000];
FILE *infu;
FILE *outfu;
/**/
/* Set temporary filename. */
misc_timepidstring( timepid );
sprintf(tempfile ,"temp-atf-%s",timepid);
misc_delete_file(tempfile);
/* Open, copy, append, move. */
if (FileExist(userfile) == 0) { sprintf(wrd,"touch %s",userfile); system(wrd); }
infu = fopen_read(userfile);
outfu= fopen_write(tempfile);
while (fgetline(line,infu)) { fprintf(outfu,"%s\n",line); }
fclose(infu);
fprintf(outfu,"%s\n",userline);
fclose(outfu);
sprintf(line,"mv %s %s",tempfile,userfile);
system(line);
return;
}


/* ----------------------------------------------------------------------
  Get a character string from an IPAC line given the name of the column and
  the column header data from: 'misc_get_IPAC_columns()'.
  Trims any leading or trailing blanks from the string.
    Input:  line[]    : Data line from IPAC file.
            colname[] : Name of column to get value from.
            ITC[]     : IPAC Table Column information structure array.
            mode      : 1=case sensitve. 0=case insensitive
   Output: wrd[]      : Value from column.
*/
/*@@*/
void misc_get_character_IPAC( char line[], char colname[], ITCtype ITC[], int mode, char wrd[] )
{
int ii;
ii = misc_find_IPAC_column( colname, ITC, mode );
substrcpy_terminate(line,ITC[ii].b1,ITC[ii].b2,wrd,0);
misc_trim_blanks( wrd );
return;
}


/* ----------------------------------------------------------------------
  Load in a ".." series of numbers from a string.
  For example: 3..6  would be loaded as 3,4,5,6.
          and: -4..2 would be loaded as -4,-3,-2,-1,0,1,2
          and: 7     would be loaded as 7.
  Input: wrd   : string to read number(s) from.
 In/Out: nkey  : current position in array of integers to load more numbers.
 In/Out: key[] : updated list of numbers.
  Input: maxkey: Maximum number of entries.
*/
/*@@*/
void misc_load_integer_series( char wrd[], int *nkey, int key[], int maxkey )
{
/**/
int ii,ii1,ii2,kk;
char tw[200];
/**/
strcpy(tw,wrd);
kk = cindex(tw,"..");
if (kk < 0) kk = cindex(tw,"--");
if (kk < 0) { 
  if (*nkey >= maxkey) {
    fprintf(stderr,"===WARNING:mlis:1: Too many entries in list (%s) [maxkey=%d].\n",wrd,maxkey);
    return;
  }
  key[(*nkey)] = GLV(tw,1);
  *nkey = *nkey + 1;
} else {
  tw[(kk  )] = ' ';
  tw[(kk+1)] = ' ';
  ii1 = GLV(tw,1);
  ii2 = GLV(tw,2);
  for (ii=ii1; ii<=ii2; ++ii) {
    if (*nkey >= maxkey) {
      fprintf(stderr,"===WARNING:mlis:2: Too many entries in list (%s) [maxkey=%d].\n",wrd,maxkey);
      return;
    }
    key[(*nkey)] = ii;
    *nkey = *nkey + 1;
  }
}
return;
}


/* ----------------------------------------------------------------------
  Load in an integer sequence of numbers from a string.
  For example: 1,2,3..6  would be loaded as 1,2,3,4,5,6.
          and: -4..2,5,6 would be loaded as -4,-3,-2,-1,0,1,2,5,6
          and: 7     would be loaded as 7.
  Input: wrd   : string to read number(s) from.
 In/Out: nkey  : current position in array of integers to load more numbers.
 In/Out: key[] : updated list of numbers.
  Input: maxkey: Maximum number of entries.
*/
/*@@*/
void misc_load_integer_sequence( char wrd[], int *nkey, int key[], int maxkey )
{
/**/
int ii,nn,bwl,jj,kk;
/**/
const int maxsubkey=300;
int subkey[maxsubkey];
int nsubkey;
/**/
char sw[200];
/**/

/* Now divide 'wrd' into sub-words. */
bwl = clc(wrd);
nn = 0;
kk = 0;
for (jj=0; jj<=bwl; ++jj) {
  if (wrd[jj] == ',') {
    sw[kk] = '\0';
    nsubkey = 0;
    if (clc(sw) > -1) { misc_load_integer_series( sw, &nsubkey, subkey, maxsubkey ); }
    for (ii=0; ii<nsubkey; ++ii) {
      key[nn] = subkey[ii];
      ++nn;
      if (nn >= maxkey) {
        fprintf(stderr,"===WARNING: mliseq:1: Too many entries in list (%s) [maxkey=%d].\n",wrd,maxkey);
        *nkey = nn;
        return;
      }
    }
    kk = 0;
  } else {
    sw[kk] = wrd[jj];
    ++kk;
  }
}
sw[kk] = '\0';
nsubkey = 0;
if (clc(sw) > -1) { misc_load_integer_series( sw, &nsubkey, subkey, maxsubkey ); }
for (ii=0; ii<nsubkey; ++ii) {
  key[nn] = subkey[ii];
  ++nn;
  if (nn >= maxkey) {
    fprintf(stderr,"===WARNING: mliseq:2: Too many entries in list (%s) [maxkey=%d].\n",wrd,maxkey);
    *nkey = nn;
    return;
  }
}
*nkey = nn;
return;
}


/* ----------------------------------------------------------------------
  Convert a date string 'YYYY-MM-DDTHH:MM:SS" into seconds since 1970 (unix time).
  New version. -tab 09oct2011
  Reads format 'YYYY-MM-DDTHH:MM:SS' instead of 'YYMMDDTHHMMSSZ'..
*/
/*@@*/
double misc_datestring_to_unixtime_2( char ds[] )
{
/**/
const int nd[12]  = { 0,31,59,90,120,151,181,212,243,273,304,334 };
const int ndly[12]= { 0,31,60,91,121,152,182,213,244,274,305,335 };
/**/
double tt,yy,mm,dd,HH,MM,SS,doy;
/**/
int ii,iyy,imm;
/**/
/*   yyyy-mm-ddTHH:MM:SS
     12345678901234567890
     ----+----1----+----2  */
/**/
/* Get numbers. */
yy=cvalread(ds, 1, 4);   mm=cvalread(ds, 6, 7);   dd=cvalread(ds, 9,10);
HH=cvalread(ds,12,13);   MM=cvalread(ds,15,16);   SS=cvalread(ds,18,19);
iyy = yy;
imm = mm;

/* Check. */
if ((yy < 1970.)||(mm < 1.)||(mm > 12.)||(dd < 1.)||(dd > 31.)) {
  fprintf(stderr,"***ERROR: mcds: problem with date '%s'.\n",ds);
  return(0.);
}
if ((HH < 0.)||(HH > 23.)||(MM < 0.)||(MM > 59.)||(SS < 0.)||(SS > 59.)) {
  fprintf(stderr,"***ERROR: mcds: problem with time '%s'.\n",ds);
  return(0.);
}

/* 
  Count up all years from 1970 to yy-1.
  Account for leap years, remember that 2000 IS a leap year (but 1900 was not).
  Ignore any leap second corrections.
*/
tt = 0.;
for (ii=1970; ii<=(iyy-1); ++ii) {
  if ((ii%4) == 0) { tt = tt + (366. * 24. * 3600.); }   /* leap year     */
              else { tt = tt + (365. * 24. * 3600.); }   /* non-leap year */
}
/* Days in the year up to the month given. */
if ((iyy%4) == 0) { doy = ndly[(imm - 1)]; }     /* leap year     */
             else { doy = nd[(imm - 1)];   }     /* non-leap year */

/* Remaining time within the given month. */
tt = tt + SS + (MM * 60.) + (HH * 3600.) + ((doy + dd - 1.) * 24. * 3600.);

return(tt);
}



/* ----------------------------------------------------------------------
  Convert a date string 'YYMMDDTHHMMSSZ" into seconds since 1970 (unix time).
*/
/*@@*/
double misc_datestring_to_unixtime( char ds[] )
{
/**/
const int nd[12]  = { 0,31,59,90,120,151,181,212,243,273,304,334 };
const int ndly[12]= { 0,31,60,91,121,152,182,213,244,274,305,335 };
/**/
double tt,yy,mm,dd,HH,MM,SS,doy;
/**/
int ii,iyy,imm;
/**/
/*   yymmddTHHMMSSZ
     123456 890123 
     ----+----1----+----2  */
/**/
/* Get numbers. */
yy=cvalread(ds, 1, 2);   mm=cvalread(ds, 3, 4);   dd=cvalread(ds, 5, 6);
HH=cvalread(ds, 8, 9);   MM=cvalread(ds,10,11);   SS=cvalread(ds,12,13);
if (yy > 50) { yy = yy + 1900.; } else { yy  = yy + 2000.; }
iyy = yy;
imm = mm;

/* Check. */
if ((yy < 1970.)||(mm < 1.)||(mm > 12.)||(dd < 1.)||(dd > 31.)) {
  fprintf(stderr,"***ERROR: mcds: problem with date '%s'.\n",ds);
  return(0.);
}
if ((HH < 0.)||(HH > 23.)||(MM < 0.)||(MM > 59.)||(SS < 0.)||(SS > 59.)) {
  fprintf(stderr,"***ERROR: mcds: problem with time '%s'.\n",ds);
  return(0.);
}

/* 
  Count up all years from 1970 to yy-1.
  Account for leap years, remember that 2000 IS a leap year (but 1900 was not).
  Ignore any leap second corrections.
*/
tt = 0.;
for (ii=1970; ii<=(iyy-1); ++ii) {
  if ((ii%4) == 0) { tt = tt + (366. * 24. * 3600.); }   /* leap year     */
              else { tt = tt + (365. * 24. * 3600.); }   /* non-leap year */
}
/* Days in the year up to the month given. */
if ((iyy%4) == 0) { doy = ndly[(imm - 1)]; }     /* leap year     */
             else { doy = nd[(imm - 1)];   }     /* non-leap year */

/* Remaining time within the given month. */
tt = tt + SS + (MM * 60.) + (HH * 3600.) + ((doy + dd - 1.) * 24. * 3600.);

return(tt);
}




/* ----------------------------------------------------------------------
  Convert a "timegetstring()" time string (yyyy-mm-ddTHH:MM:SS) into 
  seconds since 1970 (unix time).
*/
/*@@*/
double misc_convert_time_string( char ts[] )
{
/**/
const int nd[12]  = { 0,31,59,90,120,151,181,212,243,273,304,334 };
const int ndly[12]= { 0,31,60,91,121,152,182,213,244,274,305,335 };
/**/
double tt,yy,mm,dd,HH,MM,SS,doy;
/**/
int ii,iyy,imm;
/**/
/*   yyyy-mm-ddTHH:MM:SS
     1234 67 90 23 56 89
     ----+----1----+----2  */
/**/
/* Get numbers. */
yy=cvalread(ts, 1, 4);   mm=cvalread(ts, 6, 7);   dd=cvalread(ts, 9,10);
HH=cvalread(ts,12,13);   MM=cvalread(ts,15,16);   SS=cvalread(ts,18,19);
iyy = yy;
imm = mm;

/* Check. */
if ((yy < 1970.)||(mm < 1.)||(mm > 12.)||(dd < 1.)||(dd > 31.)) {
  fprintf(stderr,"***ERROR: mcts: problem with date '%s'.\n",ts);
  return(0.);
}
if ((HH < 0.)||(HH > 23.)||(MM < 0.)||(MM > 59.)||(SS < 0.)||(SS > 59.)) {
  fprintf(stderr,"***ERROR: mcts: problem with time '%s'.\n",ts);
  return(0.);
}

/* 
  Count up all years from 1970 to yy-1.
  Account for leap years, remember that 2000 IS a leap year (but 1900 was not).
*/
tt = 0.;
for (ii=1970; ii<=(iyy-1); ++ii) {
  if ((ii%4) == 0) { tt = tt + (366. * 24. * 3600.); }   /* leap year     */
              else { tt = tt + (365. * 24. * 3600.); }   /* non-leap year */
}
/* Days in the year up to the month given. */
if ((iyy%4) == 0) { doy = ndly[(imm - 1)]; }     /* leap year     */
             else { doy = nd[(imm - 1)];   }     /* non-leap year */

/* Remaining time within the given month. */
tt = tt + SS + (MM * 60.) + (HH * 3600.) + ((doy + dd - 1.) * 24. * 3600.);

return(tt);
}


/* ----------------------------------------------------------------------
  Find file date of a file given a filename.
  Returns seconds since 1970 in Universal Time (UTC).
  Returns -1 if problem.
*/
/*@@*/
int misc_file_unixtime_UTC( char file[] )
{
/**/
int unixtime = 0;
struct stat Stat;
/**/
if (stat(file,&Stat) != 0) { return(-1); }
unixtime = Stat.st_mtime;        /* Modification time in seconds since 1970. */
return(unixtime);
}

/* ----------------------------------------------------------------------
 Find file size of a file given a filename.  -tab 01feb2011
 Returns size of file in bytes.
 Returns -1 if problem.
*/
/*@@*/
int misc_file_size( char file[] )
{
/**/
int filesize = 0;
struct stat Stat;
/**/
if (stat(file,&Stat) != 0) { return(-1); }
filesize = Stat.st_size;        /* total size, in bytes. */
return(filesize);
}


/* ----------------------------------------------------------------------
  Grab a date from a 'ls -laF --time-style=long-iso' type command.
  Put the string in (yyyy-mm-ddTHH:MM:SS) format.
  Returns blank string if problem.  Example input line:
-rw-r--r--  1 tab galex    1609 2012-01-21 21:54 hmake.c
*/
/*@@*/
void misc_ls_date( char line[], char ts[] )
{
int more,ii;
strcpy(ts,""); more=1; ii=4;
while ((more)&&(ii < (clc(line)-11))) {
  if (line[ii] == '-') { 
    if ((line[ii+3] == '-')&&(line[ii+9] == ':')) { 
      if ((line[ii-4] == '1')||(line[ii-4] == '2')) {
        substrcpy_terminate(line,ii-4,ii+11,ts,0);
        ts[10]='T'; strcat(ts,":00"); more=0;
      }
    }
  }
  ++ii;
}
return;
}


/* ----------------------------------------------------------------------
  Convert julian date into a unix time (seconds since 1970).
  Uses the julian date of "1970-01-01T00:00:00.0" ==  2440587.500000 .
*/
/*@@*/
double misc_julian_to_unixtime( double jd )
{
double unixtime;
unixtime = (jd - 2440587.500 ) * (24. * 3600.);
return(unixtime);
}

/* ----------------------------------------------------------------------
  Convert unix time (seconds since 1970) into a julian date.
  Uses the julian date of "1970-01-01T00:00:00.0" ==  2440587.500000 .
*/
/*@@*/
double misc_unixtime_to_julian( double unixtime )
{
double jd;
jd = 2440587.500 + ( unixtime / (24. * 3600.) );
return(jd);
}


/* ----------------------------------------------------------------------
  Convert unix time (seconds since 1970) into a "timegetstring()" time 
  string (yyyy-mm-ddTHH:MM:SS).  Look at "misc_convert_time_string()" to
  go the other direction.
         " unixtime_to_datestring "
*/
/*@@*/
void misc_reverse_unix_time( double unixtime, char ts[] )
{
/**/
const int nd[13]  = { 0,31,59,90,120,151,181,212,243,273,304,334,365 };
const int ndly[13]= { 0,31,60,91,121,152,182,213,244,274,305,335,366 };
/**/
double tt,year_seconds,whole_month_seconds;
/**/
int iyear,imonth,iday,ihour,iminute,isecond;
/**/

/* Default. */
strcpy(ts,"");
tt = unixtime;

/* Check. */
if (tt < 1.) {
  fprintf(stderr,"===WARNING: mrut: unix time is too small [%f].\n",tt);
  return;
}

/* Find iyear and subtract all complete years from unix time. */
year_seconds = 0.;
iyear = 1970;
while (tt > 0.) {
  if ((iyear%4) == 0) { year_seconds = 366. * 24. * 3600.; }   /* leap year     */
                 else { year_seconds = 365. * 24. * 3600.; }   /* non-leap year */
  tt = tt - year_seconds;
  ++iyear;
}
if (tt < 0.) {
  tt = tt + year_seconds;
  --iyear;
} else {
  tt = tt + 0.001;
}

/* Find imonth. */
if ((iyear%4) == 0) {
  imonth = 1;
  whole_month_seconds = ( (double)ndly[imonth] ) * 24. * 3600.;
  while (tt > whole_month_seconds) {
    ++imonth; 
    whole_month_seconds = ( (double)ndly[imonth] ) * 24. * 3600.;
  }
  whole_month_seconds = ( (double)ndly[(imonth-1)] ) * 24. * 3600.;
  tt = tt - whole_month_seconds;
} else {
  imonth = 1;
  whole_month_seconds = ( (double)nd[imonth] ) * 24. * 3600.;
  while (tt > whole_month_seconds) {
    ++imonth; 
    whole_month_seconds = ( (double)nd[imonth] ) * 24. * 3600.;
  }
  whole_month_seconds = ( (double)nd[(imonth-1)] ) * 24. * 3600.;
  tt = tt - whole_month_seconds;
}

/* Find iday. */
iday = tt / (24. * 3600.);
iday = iday + 1;
tt = tt - ( (double)(iday - 1) * 24. * 3600. );

/* Find ihour. */
ihour = tt / 3600.;
tt = tt - ( (double)ihour * 3600. );

/* Find iminute. */
iminute = tt / 60.;
tt = tt - ( (double)iminute * 60. );

/* Find isecond (round off). */
isecond = tt + 0.5;

/* Write time string: yyyy-mm-ddTHH:MM:SS */
sprintf(ts,"%4.4d-%2.2d-%2.2dT%2.2d:%2.2d:%2.2d",iyear,imonth,iday,ihour,iminute,isecond);

return;
}


/* ----------------------------------------------------------------------
  Sort an array using "QuickSort" method.
  Originally from NUMERICAL RECIPES (tab July 1988)
  Converted to C September 1999.
*/
/*@@*/
void cqcksrt4(int nn, float arr[] )
{
/**/
const float fm = 7875.;
const float fa = 211.;
const float fc = 1663.;
const float fmi= 1.2698413e-4;
const int mm = 7;
const int nstack = 100;
int istack[nstack];
int jstack,ll,ir,j,i,iq,ii,jj,flag1;
float fx,aa;
/**/
/* Initialize. */
jstack=0; ll=1; ir=nn; fx=0.;
/* Big loop. */
while(1) {
if (ir-ll < mm) {
  for (j=ll+1; j<=ir; ++j) {
    aa=arr[j -1];
    flag1=0;
    for (i=j-1; i>=1; i=i-1) {
      if (arr[i -1] <= aa) { flag1=1; break; }
      arr[i+1 -1]=arr[i -1];
    }
    if (flag1 == 0) i=0;
    arr[i+1 -1]=aa;
  }
  if (jstack == 0) return;   /* Exit routine here. */
  ir= istack[jstack -1];
  ll= istack[jstack-1 -1];
  jstack = jstack-2;
} else {
  i=ll; j=ir;
  ii= ((fx*fa)+fc);
  jj= (fm);
  fx= ii%jj;
  iq= ll + ( ((ir-ll)+1) * (fx*fmi) );
  aa = arr[iq -1];
  arr[iq -1] = arr[ll -1];
  while (1) {
    while (j > 0) {
      if (aa >= arr[j -1]) break;
      j=j-1;
    }
    if (j <= i) {
      arr[i -1] = aa;
      break;
    }
    arr[i -1] = arr[j -1];
    i=i+1;
    while (i <= nn) {
      if (aa <= arr[i -1]) break;
      i=i+1;
    }
    if (j <= i) {
      arr[j -1]=aa;
      i=j;
      break;
    }
    arr[j -1]=arr[i -1];
    j=j-1;
  }
  jstack=jstack+2;
  if (jstack > nstack) fprintf(stderr,"***ERROR: qcksrt: NSTACK must be made larger.\n");
  if ((ir-i) >= (i-ll)) {
    istack[jstack -1]=ir;
    istack[jstack-1 -1]=i+1;
    ir=i-1;
  } else {
    istack[jstack -1]  = i-1;
    istack[jstack-1 -1]= ll;
    ll=i+1;
  }
}     /* if block */
}     /* while loop */
return;
}



/* ----------------------------------------------------------------------
  THIS VERSION USES A double precision ARRAY.
  Sort an array using "QuickSort" method.
  Originally from NUMERICAL RECIPES (tab July 1988)
  Converted to C September 1999.
*/
/*@@*/
void cqcksrt8(int nn, double arr[] )
{
/**/
const double fm = 7875.;
const double fa = 211.;
const double fc = 1663.;
const double fmi= 1.2698413e-4;
const int mm = 7;
const int nstack = 100;
int istack[nstack];
int jstack,ll,ir,j,i,iq,ii,jj,flag1;
double fx,aa;
/**/
/* Initialize. */
jstack=0; ll=1; ir=nn; fx=0.;
/* Big loop. */
while(1) {
if (ir-ll < mm) {
  for (j=ll+1; j<=ir; ++j) {
    aa=arr[j -1];
    flag1=0;
    for (i=j-1; i>=1; i=i-1) {
      if (arr[i -1] <= aa) { flag1=1; break; }
      arr[i+1 -1]=arr[i -1];
    }
    if (flag1 == 0) i=0;
    arr[i+1 -1]=aa;
  }
  if (jstack == 0) return;   /* Exit routine here. */
  ir= istack[jstack -1];
  ll= istack[jstack-1 -1];
  jstack = jstack-2;
} else {
  i=ll; j=ir;
  ii= ((fx*fa)+fc);
  jj= (fm);
  fx= ii%jj;
  iq= ll + ( ((ir-ll)+1) * (fx*fmi) );
  aa = arr[iq -1];
  arr[iq -1] = arr[ll -1];
  while (1) {
    while (j > 0) {
      if (aa >= arr[j -1]) break;
      j=j-1;
    }
    if (j <= i) {
      arr[i -1] = aa;
      break;
    }
    arr[i -1] = arr[j -1];
    i=i+1;
    while (i <= nn) {
      if (aa <= arr[i -1]) break;
      i=i+1;
    }
    if (j <= i) {
      arr[j -1]=aa;
      i=j;
      break;
    }
    arr[j -1]=arr[i -1];
    j=j-1;
  }
  jstack=jstack+2;
  if (jstack > nstack) fprintf(stderr,"***ERROR: qcksrt: NSTACK must be made larger.\n");
  if ((ir-i) >= (i-ll)) {
    istack[jstack -1]=ir;
    istack[jstack-1 -1]=i+1;
    ir=i-1;
  } else {
    istack[jstack -1]  = i-1;
    istack[jstack-1 -1]= ll;
    ll=i+1;
  }
}     /* if block */
}     /* while loop */
return;
}



/* ----------------------------------------------------------------------
  Sort an array using "QuickSort" method.
  This version sorts a key[] array.  The original arr[] is not changed.
  This version uses a float precision array.
  Originally from NUMERICAL RECIPES (tab July 1988), converted to C Sept. 1999.
  The values of key[] are array element positions within arr[].  WARNING: Be
  sure that all values of key[] are within the array dimension bounds of arr[].
    Input: arr[], nkey, key[].    Output: key[] (sorted).
*/
/*@@*/
void cqcksrtkey4(float arr[], int nkey, int key[] )
{
/**/
const float fm = 7875.;
const float fa = 211.;
const float fc = 1663.;
const float fmi= 1.2698413e-4;
const int mm = 7;
const int nstack = 100;
int istack[nstack];
int jstack,ll,ir,j,i,iq,ii,jj,kk,flag1;
float fx;
/**/
/* Initialize. */
jstack=0; ll=1; ir=nkey; fx=0.;
/* Big loop. */
while(1) {
if (ir-ll < mm) {
  for (j=ll+1; j<=ir; ++j) {
    kk=key[j -1];
    flag1=0;
    for (i=j-1; i>=1; i=i-1) {
      if (arr[key[i -1]] <= arr[kk]) { flag1=1; break; }
      key[i+1 -1]=key[i -1];
    }
    if (flag1 == 0) i=0;
    key[i+1 -1]=kk;
  }
  if (jstack == 0) return;   /* Exit routine here. */
  ir= istack[jstack -1];
  ll= istack[jstack-1 -1];
  jstack = jstack-2;
} else {
  i=ll; j=ir;
  ii= ((fx*fa)+fc);
  jj= (fm);
  fx= ii%jj;
  iq= ll + ( ((ir-ll)+1) * (fx*fmi) );
  kk = key[iq -1];
  key[iq -1] = key[ll -1];
  while (1) {
    while (j > 0) {
      if (arr[kk] >= arr[key[j -1]]) break;
      j=j-1;
    }
    if (j <= i) {
      key[i -1] = kk;
      break;
    }
    key[i -1] = key[j -1];
    i=i+1;
    while (i <= nkey) {
      if (arr[kk] <= arr[key[i -1]]) break;
      i=i+1;
    }
    if (j <= i) {
      key[j -1]=kk;
      i=j;
      break;
    }
    key[j -1]=key[i -1];
    j=j-1;
  }
  jstack=jstack+2;
  if (jstack > nstack) fprintf(stderr,"***ERROR: qcksrt: NSTACK must be made larger.\n");
  if ((ir-i) >= (i-ll)) {
    istack[jstack -1]=ir;
    istack[jstack-1 -1]=i+1;
    ir=i-1;
  } else {
    istack[jstack -1]  = i-1;
    istack[jstack-1 -1]= ll;
    ll=i+1;
  }
}     /* if block */
}     /* while loop */
return;
}



/* ----------------------------------------------------------------------
  Sort an array using "QuickSort" method.
  This version sorts a key[] array.  The original arr[] is not changed.
  This version uses a double precision array.
  Originally from NUMERICAL RECIPES (tab July 1988), converted to C Sept. 1999.
  The values of key[] are array element positions within arr[].  WARNING: Be
  sure that all values of key[] are within the array dimension bounds of arr[].
    Input: arr[], nkey, key[].    Output: key[] (sorted).
*/
/*@@*/
void cqcksrtkey8(double arr[], int nkey, int key[] )
{
/**/
const double fm = 7875.;
const double fa = 211.;
const double fc = 1663.;
const double fmi= 1.2698413e-4;
const int mm = 7;
const int nstack = 100;
int istack[nstack];
int jstack,ll,ir,j,i,iq,ii,jj,kk,flag1;
double fx;
/**/
/* Initialize. */
jstack=0; ll=1; ir=nkey; fx=0.;
/* Big loop. */
while(1) {
if (ir-ll < mm) {
  for (j=ll+1; j<=ir; ++j) {
    kk=key[j -1];
    flag1=0;
    for (i=j-1; i>=1; i=i-1) {
      if (arr[key[i -1]] <= arr[kk]) { flag1=1; break; }
      key[i+1 -1]=key[i -1];
    }
    if (flag1 == 0) i=0;
    key[i+1 -1]=kk;
  }
  if (jstack == 0) return;   /* Exit routine here. */
  ir= istack[jstack -1];
  ll= istack[jstack-1 -1];
  jstack = jstack-2;
} else {
  i=ll; j=ir;
  ii= ((fx*fa)+fc);
  jj= (fm);
  fx= ii%jj;
  iq= ll + ( ((ir-ll)+1) * (fx*fmi) );
  kk = key[iq -1];
  key[iq -1] = key[ll -1];
  while (1) {
    while (j > 0) {
      if (arr[kk] >= arr[key[j -1]]) break;
      j=j-1;
    }
    if (j <= i) {
      key[i -1] = kk;
      break;
    }
    key[i -1] = key[j -1];
    i=i+1;
    while (i <= nkey) {
      if (arr[kk] <= arr[key[i -1]]) break;
      i=i+1;
    }
    if (j <= i) {
      key[j -1]=kk;
      i=j;
      break;
    }
    key[j -1]=key[i -1];
    j=j-1;
  }
  jstack=jstack+2;
  if (jstack > nstack) fprintf(stderr,"***ERROR: qcksrt: NSTACK must be made larger.\n");
  if ((ir-i) >= (i-ll)) {
    istack[jstack -1]=ir;
    istack[jstack-1 -1]=i+1;
    ir=i-1;
  } else {
    istack[jstack -1]  = i-1;
    istack[jstack-1 -1]= ll;
    ll=i+1;
  }
}     /* if block */
}     /* while loop */
return;
}



/* ----------------------------------------------------------------------
  Find median of an array of float (4 byte real) numbers.
  Input: arr[] (array of values which will NOT be sorted).
         nkey  (number of elements in the key array).
         key[] (key array containing array element positions within arr[]).
  Returns the median value of the key'ed subset of values.
*/
/*@@*/
float cfind_median_key(float arr[], int nkey, int key[] )
{
/**/
float median;
int kk;
/**/
/* Check. */
if (nkey < 1) { return(0.); }
/* Sort the key array. */
cqcksrtkey4(arr, nkey, key );
kk = (nkey/2);
if ((nkey%2) == 0) {
  median = ( arr[key[kk]] + arr[key[(kk-1)]] ) / 2.;   /* Even number. */
} else {
  median = arr[key[kk]];                               /* Odd number. */
}
return(median);
}



/* ----------------------------------------------------------------------
  Find median of an array of float (4 byte real) numbers.
  Input: arr[] (array of values which will be sorted).
         nn   (number of elements in the array).
  Returns the median value.
*/
/*@@*/
float cfind_median(int nn, float arr[] )
{
/**/
float median;
int kk;
/**/
/* Check. */
if (nn < 1) { return(0.); }
if (nn == 1) { return(arr[0]); }
/* Sort the array. */
cqcksrt4(nn,arr);
kk = (nn/2);
if ((nn%2) == 0) {
  median = ( arr[kk] + arr[(kk-1)] ) / 2.;   /* Even number of elements. */
} else {
  median = arr[kk];                          /* Odd number of elements. */
}
return(median);
}


/* ----------------------------------------------------------------------
  Find median of an array of float (4 byte real) numbers and then compute
    an average excluding "nrej" points from high and low end.
  Input: arr[]: array of values which will be sorted.
         nn   : number of elements in the array.
         nrej : number of elements to reject from top and bottom of sorted list.
  Returns the median'ed mean value.
*/
/*@@*/
float cfind_median_average( int nn, float arr[], int nrej )
{
/**/
float median,sum,num,mean;
int ii;
/**/
median = cfind_median( nn, arr );
/* Average. */
sum = 0.;
num = 0.;
for (ii=0; ii<nn; ++ii) {
  if ((ii >= nrej)&&(ii < (nn-nrej))) {
    sum = sum + arr[ii];
    num = num + 1.;
  }
}
if (num > 0.) { mean = sum / num; } else { mean = median; }
return(mean);
}


/* ----------------------------------------------------------------------
  Find median of an array of double (8 byte real) numbers.
  Input: arr[] (array of values which will be sorted).
         nn   (number of elements in the array).
  Returns the median value.
*/
/*@@*/
double cfind_median8(int nn, double arr[] )
{
/**/
double median;
int kk;
/**/
/* Check. */
if (nn < 1) { return(0.); }
/* Sort the array. */
cqcksrt8(nn,arr);
kk = (nn/2);
if ((nn%2) == 0) {
  median = ( arr[kk] + arr[(kk-1)] ) / 2.;   /* Even number of elements. */
} else {
  median = arr[kk];                          /* Odd number of elements. */
}
return(median);
}


/* ----------------------------------------------------------------------
  Uses double precision... -tab 01dec2009
  Find median of an array of float (4 byte real) numbers and then compute
    an average excluding "nrej" points from high and low end.
  Input: arr[]: array of values which will be sorted.
         nn   : number of elements in the array.
         nrej : number of elements to reject from top and bottom of sorted list.
  Returns the median'ed mean value.
*/
/*@@*/
double cfind_median_average8( int nn, double arr8[], int nrej )
{
/**/
double median,sum,num,mean;
int ii;
/**/
median = cfind_median8( nn, arr8 );
/* Average. */
sum = 0.;
num = 0.;
for (ii=0; ii<nn; ++ii) {
  if ((ii >= nrej)&&(ii < (nn-nrej))) {
    sum = sum + arr8[ii];
    num = num + 1.;
  }
}
if (num > 0.) { mean = sum / num; } else { mean = median; }
return(mean);
}


/* ----------------------------------------------------------------------
  Find median of an array of float (4 byte real) numbers and reject bad points.
  Input:  npt    : number of elements in the array.
  Input:  dat[]  : data values.
  In/Out: wt[]   : weight values (rejections set to 0.).
 Scratch: arr[]  : scratch array.
  Input:  rejsig : rejection sigmas (relative to RMS).
  Input:  maxrej : maximum number of rejections.
  Returns the mean value, post-rejections.
*/
/*@@*/
float cfind_median_rejections( int npt, float dat[], float wt[], 
                               float arr[], float rejsig, int maxrej )
{
/**/
float rejsig2,hi_dev2,dev2;
float rms2,mean,median,sum,num;
/**/
int rejflag,nrej,ii,jj,hi_ii,nn;
/**/

/* Check. */
if ((npt < 1)||(npt <= maxrej)) { return(0.); }

/* Set. */
rejsig2 = rejsig * rejsig;
nrej    = 0;
rejflag = 1;
median  = 0.;

/* Loop. */
while ((rejflag == 1)&&(nrej < maxrej)) {

/* Load array. */
  nn=0;
  for (ii=0; ii<npt; ++ii) { if (wt[ii] > 0.) { arr[nn] = dat[ii]; ++nn; } }
  
/* Compute median. */
  median = cfind_median( nn, arr );

/* Compute RMS relative to median (using abreviated scratch array). */
  sum = 0.;
  num = 0.;
  for (jj=0; jj<nn; ++jj) {
    sum = sum + ( (arr[jj] - median) * (arr[jj] - median) );
    num = num + 1.;
  }
  rms2 = sum / num;
  
/* Highest deviation. */
  hi_ii   = 0;
  hi_dev2 = 0.;
  for (ii=0; ii<npt; ++ii) {  if (wt[ii] > 0.) {
    dev2 = (dat[ii] - median) * (dat[ii] - median);
    if (dev2 > hi_dev2) {
      hi_dev2 = dev2;
      hi_ii   = ii;
    }
  }}
  
/* Is highest deviation greater than rejsig? */
  if ((hi_dev2 / rms2) > rejsig2) {
    rejflag = 1;
    wt[hi_ii] = 0.;
    ++nrej;
  } else { rejflag = 0; }

}

/* Load array. */
nn=0;
for (ii=0; ii<npt; ++ii) { if (wt[ii] > 0.) { arr[nn] = dat[ii]; ++nn; } }

/* Find mean of final list. */
sum = 0.;
for (ii=0; ii<nn; ++ii) { sum = sum + arr[ii]; }
if (nn > 0) { mean = sum / (float)nn; } else { mean = 0.; }

return(mean);
}



/* ----------------------------------------------------------------------
  Find simple mean and RMS.
  Input: arr[]  : array of values.
         narr   : number of elements in the array.
 Output: rms    : RMS of points used to make the average.
  Returns the average value.
*/
/*@@*/
float cfind_mean_rms( int narr, float arr[], float *rms )
{
/**/
float sum,mean;
int ii;
/**/
if (narr < 1) { *rms=0.; return(0); }
sum=0.;
for (ii=0; ii<narr; ++ii) { sum=sum+arr[ii]; }
mean = sum / (float)narr;
sum=0.;
for (ii=0; ii<narr; ++ii) {
  sum = sum + ((arr[ii] - mean) * (arr[ii] - mean));
}
*rms = sqrt( sum / (float)narr );
return(mean);
}


/* ----------------------------------------------------------------------
  Find median of an array of float (4 byte real) numbers and then throw 
    out the nrej most deviant points, and then average.
  Input: arr[]  : array of values which will be sorted.
         nn     : number of elements in the array.
         nrej   : number of most deviant elements to reject.
Scratch: mask[] : scratch array to mask points.
 Output: rms    : RMS of points used to make the average.
  Returns the median'ed mean value.
*/
/*@@*/
float cfind_median_deviant_average( int nn, float arr[], int nrej, int mask[],
                                    float *rms )
{
/**/
float median,sum,num,mean,high,dev;
int ii,nj,iihigh;
/**/
median = cfind_median( nn, arr );
for (ii=0; ii<nn; ++ii) { mask[ii]=0; }
nj=0;
while (nj < nrej) {
  high  = 0.;
  iihigh= 0;
  for (ii=0; ii<nn; ++ii) { 
  if (mask[ii] == 0) {
    dev = ABS((median - arr[ii]));
    if (dev > high) {
      high  = dev;
      iihigh= ii;
    }
  }}
  mask[iihigh] = 1;
  ++nj;
}
/* Average un-masked points. */
sum = 0.;
num = 0.;
for (ii=0; ii<nn; ++ii) { if (mask[ii] == 0) {
  sum = sum + arr[ii];
  num = num + 1.;
}}
if (num > 0.) { mean = sum / num; } else { mean = median; }
/* RMS */
sum = 0.;
for (ii=0; ii<nn; ++ii) { if (mask[ii] == 0) {
  sum = sum + ( (arr[ii] - mean) * (arr[ii] - mean) );
}}
if (num > 0.) { *rms = sqrt(( sum / num )); } else { *rms = 0.; }
return(mean);
}


/* ----------------------------------------------------------------------
  Convert an ASCII file to a PostScript file using Courier fonts.
  "land" : Use landscape mode.  "pt=n" : Specify pointsize n (default 10).
  "bold" : Use Courier-Bold instead of Courier (warning: different aspect).
  "head" : Put header at top of file, i.e. filename and date.
  "left=n" : Shift left n points (n = 70 = one inch).

  ptsz char/line line/page   LANDSCAPE: char/line line/page'
   12 ..  79  .....  55  ................  105  ...  40'
   10 ..  92  .....  66  ................  126  ...  48'
    9 .. 107  .....  73  ................  142  ...  53'
    8 .. 117  .....  82  ................  160  ...  60'
    7 .. 131  .....  94  ................  183  ...  68'
    6 .. 170  ..... 110  ................  210  ...  80'

  If psfile is blank, send to standard output.
*/
/*@@*/
void cASCII_to_PS(char afile[], char psfile[],
        int ptsz, int left, int bold, int land, int head, char font[] )
{
/**/
FILE *infu;
FILE *oufu;
int nlpp,hpos,lno,ii,kk;
char line[300], header[300];
char line1[300], line2[300];
float lss = 1.1;    /* Line skip scaling factor. */
/**/

/* Form header. */
strcpy(header,":: ");
strcat(header,afile);
strcat(header," ::");

/* Open ASCII file for reading. */
infu = fopen(afile,"r");

/* Open PostScript file for writing. */
oufu = fopen(psfile,"w");

/* Number of lines per page. */
if (land == 1) {
  nlpp = cint(530. / (ptsz * lss));
} else {
  nlpp = cint(730. / (ptsz * lss));
}

/* Write beginning stuff. */
fprintf(oufu,"%c!PS-Adobe-2.0\n",'%');
fprintf(oufu,"/ptsz %d def\n",ptsz);
fprintf(oufu,"/lskip ptsz %f mul def\n",lss);
fprintf(oufu,"/landscape{90 rotate 000 -800 translate}def\n");
fprintf(oufu,"/vinit 754 def\n");

/* Horizontal positioning */
if (ptsz < 10) {
  hpos = 20 + left;
} else {
  hpos = 30 + left;
}
fprintf(oufu,"/hpos %d def\n",hpos);

/* More stuff. */
fprintf(oufu,"/newline{/vpos vpos lskip sub def hpos vpos moveto}def\n");

/* Land scape? */
if (land == 1) {
  fprintf(oufu,"/newpage{/vpos vinit def hpos vpos moveto landscape newline}def\n");
} else {
  fprintf(oufu,"/newpage{/vpos vinit def hpos vpos moveto}def\n");
}

/* More stuff. */
fprintf(oufu,"/l{show newline}def\n");

/* Bold? */
if (clc(font) > 1) {
  fprintf(oufu,"/%s findfont\n",font);
} else {
  if (bold == 1) {
    fprintf(oufu,"/Courier-Bold findfont\n");
  } else {
    fprintf(oufu,"/Courier findfont\n");
  }
}

fprintf(oufu,"ptsz scalefont\n");
fprintf(oufu,"setfont\n");
fprintf(oufu,"newpage\n");

/* start text */
lno=1;

/* Add header to top. */
if (head == 1) {
  fprintf(oufu,"( %s )l\n",header);
  ++lno;
}

/* Print lines */
while (fgetline(line,infu) == 1) {

/* Correct for special characters ), (, and \ .  */
  ii=0;
  while (ii <= clc(line)) {
    kk = strlen(line) - 1;
    if ((line[ii] == ')') || (line[ii] == '(') || (line[ii] == '\\')) {
      substrcpy_terminate(line,0,ii-1,line1,0);
      substrcpy_terminate(line,ii,kk,line2,0);
      strcpy(line,line1);
      strcat(line,"\\");
      strcat(line,line2);
      ++ii;
    }
    ++ii;
  }
  fprintf(oufu,"(%s)l\n",line);
  if (lno%nlpp == 0) {
    fprintf(oufu,"showpage\n");
    fprintf(oufu,"newpage\n");
    lno=1;
  } else {
    ++lno;
  }
}

/* Finish it off. */
fprintf(oufu,"showpage\n");

/* Close files. */
fclose(infu);
fclose(oufu);

return;
}


/* ----------------------------------------------------------------------
  Read a CSV file (.csv) ( comma separated values ).
  Input: file     : ASCII CSV file.
         xxtc     : table column for xx[]
         yytc     : table column for yy[]
 Output: xx[]     : x values.
         yy[]     : y values.
 Returns number of points (returns 0 if there is a problem).
*/
/*@@*/
int misc_read_CSV( char file[], int xxtc, int yytc, float xx[], float yy[] )
{
/**/
int npt;
FILE *infu;
char line[200];
/**/
/* Open CSV ASCII file. */
infu = fopen_read( file );
npt=0;
while (fgetline(line,infu) == 1) {
  xx[npt] = GLV(line,xxtc);
  yy[npt] = GLV(line,yytc);
  if ((xx[npt] < -1.e+30)||(yy[npt] < -1.e+30)) { fclose(infu); return(0); }
  ++npt;
}
fclose(infu);
return(npt);
}


/* ----------------------------------------------------------------------
  Reads in all of the column names of an IPAC style ASCII table file.

  Input: infile   : ASCII IPAC file.
 Output: *names[] : Table column names.
         *forms[] : The form types, e.g. I,J,E,D etc.
         nc       : number of columns to be read.

 NOTE: Allocate the names and forms array like this:
    const int maxnss  = 100;  /x Maximum string size of column names. x/
    const int maxfss  = 10;   /x Maximum string size of column names. x/
    const int maxname = 200;  /x Maximum number of column names. x/
    char *names[maxname];
    for (ii=0; ii<maxname; ++ii) { names[ii] = (char *)malloc(maxnss*1); }
    for (ii=0; ii<maxname; ++ii) { forms[ii] = (char *)malloc(maxfss*1); }
*/
/*@@*/
void misc_read_IPAC_column_names( char infile[], char *names[], char *forms[], int *nc )
{
/**/
int ii,jj,kk,nn,ptr;
/**/
char line[9000];
char wrd[200];
char wrd2[200];
/**/
FILE *infu;
/**/

/* Open input file. */
infu = fopen_read( infile );

/* Find column name header line. */
ii=1; strcpy(line,"");
while ((line[0] != '|')&&(ii == 1)) { ii = fgetline(line,infu); }
if (ii == 0) {
  fprintf(stderr,"***ERROR: mrIcn: Could not find header line in file %s.\n",infile);
  *nc=0;
  return;
}

/* Dissect column names. */
nn = 0;
ptr= 0;
while (ptr >= 0) {
  line[ptr] = ' ';
  kk = cindex(line,"|");
  if (kk > -1) {
    if ((ptr+1) > (kk-1)) {
      fprintf(stderr,"***ERROR: mrIcn: Found a zero character column name.\n");
      *nc = 0;
      return;
    }
    substrcpy_terminate(line,ptr+1,kk-1,wrd,0);
    ii = cfc(wrd);
    jj = clc(wrd);
    substrcpy_terminate(wrd,ii,jj,names[nn],0);
    strcpy(forms[nn],"D");  /* default */
    ++nn;
  }
  ptr = kk;
}
*nc = nn;
  
/* Find column forms header line. */
ii=1; strcpy(line,"");
while ((line[0] != '|')&&(ii == 1)) { ii = fgetline(line,infu); }
if (ii == 0) {
  printf("NOTE: Could not find forms header line in file %s.\n",infile);
  return;
}

/* Dissect column forms. */
nn = 0;
ptr= 0;
while (ptr >= 0) {
  line[ptr] = ' ';
  kk = cindex(line,"|");
  if (kk > -1) {
    if ((ptr+1) > (kk-1)) {
      fprintf(stderr,"***ERROR: mrIcn: Found a zero character column name.\n");
      *nc = 0;
      return;
    }
    substrcpy_terminate(line,ptr+1,kk-1,wrd,0);
    ii = cfc(wrd);
    jj = clc(wrd);
    substrcpy_terminate(wrd,ii,jj,wrd2,0);
    cupper_case(wrd2);
    strcpy(forms[nn],"D");
    if (cindex(wrd2,"DOUB") > -1) { strcpy(forms[nn],"D"); } else {
    if (cindex(wrd2,"REAL") > -1) { strcpy(forms[nn],"E"); } else {
    if (cindex(wrd2,"FLOA") > -1) { strcpy(forms[nn],"E"); } else {
    if (cindex(wrd2,"INT" ) > -1) { strcpy(forms[nn],"J"); } else {
    if (cindex(wrd2,"LONG") > -1) { strcpy(forms[nn],"J"); } else {
    if (cindex(wrd2,"SHOR") > -1) { strcpy(forms[nn],"I"); } else {
    if (cindex(wrd2,"J"   ) > -1) { strcpy(forms[nn],"J"); } else {
    if (cindex(wrd2,"D"   ) > -1) { strcpy(forms[nn],"D"); } else {
    if (cindex(wrd2,"E"   ) > -1) { strcpy(forms[nn],"E"); } else {
    if (cindex(wrd2,"I"   ) > -1) { strcpy(forms[nn],"I"); } else {
    if (cindex(wrd2,"R"   ) > -1) { strcpy(forms[nn],"E"); }      }}}}} }}}}}
    ++nn;
  }
  ptr = kk;
}
  
/* Error. */
if (nn != *nc) {
  fprintf(stderr,"***ERROR: mrIcn: Number of column names(%d) does not match number of column forms(%d).\n",*nc,nn);
  *nc=0;
  return;
}

return;
}



/* ----------------------------------------------------------------------
  Reads in an IPAC style ASCII file given the column names.
  Input: file     : ASCII IPAC file.
         *names[] : Table column names.
         nc       : number of columns to be read.
         maxpt    : maximum number of points in img[].
 Output: nr       : number of rows.
         img[]    : image array (CAREFUL! NOTE that the fastest varying 
                                   indice is row#, not col#).

 NOTE: Allocate the names array like this:
    for (ii=0; ii<maxname; ++ii) { names[ii] = (char *)malloc(maxnss*1); }
*/
/*@@*/
void misc_read_IPAC_float( char file[], char *names[], int nc, int maxpt, int *nr, float img[], int verbose )
{
/**/
int ii,jj,kk,max_nc,pixno;
char line[1000];
FILE *infu;
ITCtype *ITC;
int *colnum;
/**/

/* Open IPAC ASCII file. */
infu = fopen_read( file );

/* First line. */
ii = fgetline(line,infu);
max_nc = misc_get_number_of_IPAC_columns( line );

/* Allocate and get column names. */
ITC    = (ITCtype *)malloc((max_nc+1)*sizeof(ITCtype));
colnum = (int *)malloc((max_nc+1)*sizeof(int));
misc_get_IPAC_columns( line, ITC );

/* Which columns? */
for (ii=0; ii<nc; ++ii) {
  colnum[ii] = -1;
  for (kk=0; kk<ITC[0].ncol; ++kk) {
    if (strcmp(names[ii],ITC[kk].name) == 0) { colnum[ii] = kk; }
  }
}

/* Count non-header rows, then close and reopen file.. */
jj=0;
while (fgetline(line,infu) == 1) { if (line[0] != '|') { ++jj; } }
*nr = jj;
fclose(infu);
infu = fopen_read( file );

/* Load data. */
jj=0;
while (fgetline(line,infu) == 1) { if (line[0] != '|') {
  for (ii=0; ii<nc; ++ii) {  if (colnum[ii] > -1) {
    pixno = jj + (ii * (*nr));
    if (pixno > maxpt) {
      printf("=== WARNING: Exceeded limit of image array: maxpt=%d\n",maxpt);
      *nr = jj;
      return;
    }
    img[pixno] = cvalread( line, ITC[colnum[ii]].b1, ITC[colnum[ii]].b2 );
  }}
  ++jj;
}}

/* Close and free. */
fclose(infu);
free(ITC);    ITC   =NULL;
free(colnum); colnum=NULL;

return;
}




/* ----------------------------------------------------------------------
  Reads in an IPAC style ASCII file given the column names.
  Uses all double values.
  Input: file     : ASCII IPAC file.
         *names[] : Table column names.
         nc       : number of columns to be read.
         maxpt    : maximum number of points in img[].
 Output: nr       : number of rows.
         img[]    : image array (CAREFUL! NOTE that the fastest varying 
                                   indice is row#, not col#).

 NOTE: Allocate the names array like this:
    for (ii=0; ii<maxname; ++ii) { names[ii] = (char *)malloc(maxnss*1); }
*/
/*@@*/
void misc_read_IPAC_double( char file[], char *names[], int nc, int maxpt, int *nr, double img[], int verbose )
{
/**/
int ii,jj,kk,max_nc,pixno;
char line[1000];
FILE *infu;
ITCtype *ITC;
int *colnum;
char wrd[100];
/**/

/* Open IPAC ASCII file. */
infu = fopen_read( file );

/* First line. */
ii = fgetline(line,infu);
max_nc = misc_get_number_of_IPAC_columns( line );

/* Allocate and get column names. */
ITC    = (ITCtype *)malloc((max_nc+1)*sizeof(ITCtype));
colnum = (int *)malloc((max_nc+1)*sizeof(int));
misc_get_IPAC_columns( line, ITC );

/* Echo. */
if (verbose == 1) {
  for (kk=0; kk<ITC[0].ncol; ++kk) {
    printf("ITC:%3d:'%s'.\n",kk+1,ITC[kk].name);
  }
}

/* Change column numbers to names. */
for (ii=0; ii<nc; ++ii) {
  for (jj=0; jj<max_nc+1; ++jj) {
    sprintf(wrd,"%d",jj+1);
    if (strcmp(names[ii],wrd) == 0) {
      strcpy(names[ii],ITC[jj].name);
      if (verbose == 1) { printf("NOTE: Change column number request '%s' to name '%s'.\n",wrd,ITC[jj].name); }
    }
  }
}

/* Which columns?  The column name '0' is special. */
for (ii=0; ii<nc; ++ii) {
  colnum[ii] = -1;
  for (kk=0; kk<ITC[0].ncol; ++kk) {
    if (strcmp(names[ii],ITC[kk].name) == 0) { colnum[ii] = kk; }
  }
  if (strcmp(names[ii],"0") != 0) {
    if (colnum[ii] < 0) { printf("***error: cannot find column name '%s'.\n",names[ii]); exit(1); }
  }
}

/* Count non-header rows, then close and reopen file.. */
jj=0;
while (fgetline(line,infu) == 1) { if (line[0] != '|') { ++jj; } }
*nr = jj;
fclose(infu);
infu = fopen_read( file );

/* Load data. */
jj=0;
while (fgetline(line,infu) == 1) { if (line[0] != '|') {
  for (ii=0; ii<nc; ++ii) {  
    if (colnum[ii] == -1) {
      pixno = jj + (ii * (*nr));
      img[pixno] = jj;
    } else {
      pixno = jj + (ii * (*nr));
      if (pixno > maxpt) {
        printf("=== WARNING: Exceeded limit of image array: maxpt=%d\n",maxpt);
        *nr = jj;
        return;
      }
      img[pixno] = cvalread( line, ITC[colnum[ii]].b1, ITC[colnum[ii]].b2 );
    }
  }
  ++jj;
}}

/* Close and free. */
fclose(infu);
free(ITC);    ITC   =NULL;
free(colnum); colnum=NULL;

return;
}






/* -----------------------------------------------------------------------
  Load (integer) data from a column in IDD structure.  -tab 08jan2012
  Input: IDD      : IPAC Double Data structure.
         column[] : column name.
 Output: iaa[]    : allocate this to 'IDD.nr' size.
*/
/*@@*/
void misc_load_IDD_int( IDDtype IDD, char column[], int iaa[] )
{
/**/
int ii,uc,pixno;
/**/
/* Find column number. */
uc=-1;
for (ii=0; ii<IDD.nc; ++ii) {
  if (strcmp(column,IDD.names[ii]) == 0) uc = ii;
}
if (uc < 0) { printf("***error:ml_IDD_int: could not find column '%s'.\n",column); exit(1); }
/* Load data. */
for (ii=0; ii<IDD.nr; ++ii) {
  pixno = ii + (IDD.nr * uc);   /* rows are columns, and columns are rows. */
  iaa[ii] = cnint(( IDD.img[pixno] ));
}
return;
}

/* -----------------------------------------------------------------------
  Load data from a column in IDD structure.  -tab 08jan2012
  Input: IDD      : IPAC Double Data structure.
         column[] : column name.
 Output: aa[]     : allocate this to 'IDD.nr' size.
*/
/*@@*/
void misc_load_IDD( IDDtype IDD, char column[], double aa[] )
{
/**/
int ii,uc,pixno;
/**/
/* Find column number. */
uc=-1;
for (ii=0; ii<IDD.nc; ++ii) {
  if (strcmp(column,IDD.names[ii]) == 0) uc = ii;
}
if (uc < 0) { printf("***error:ml_IDD: could not find column '%s'.\n",column); exit(1); }
/* Load data. */
for (ii=0; ii<IDD.nr; ++ii) {
  pixno = ii + (IDD.nr * uc);   /* rows are columns, and columns are rows. */
  aa[ii] = IDD.img[pixno];
}
return;
}
/* -----------------------------------------------------------------------
  Read IPAC Float Data from IPAC table file. -tab 08jan2012
  Input: ipacfile[]  : filename of IPAC table double data.
         columns[]   : string of column names separated by commas.
 Output: *IDD        : IPAC Float Data structure.
*/
/*@@*/
void misc_read_IDD( char ipacfile[], char columns[], IDDtype *IDD, int verbose )
{
/**/
int maxpt,ii,nn,pp1,pp2,hicn;
char line[1000];
FILE *infu;
/**/
/* Store IPAC table filename. */
strcpy(IDD[0].file,ipacfile);
/* Count column names. */
pp1=0; pp2=-1; nn=0; hicn=1;
while (pp2 < clc(columns)) {
  pp1 = pp2 + 1;
  pp2 = cindex2(columns,",",pp1);
  if (pp2 < 0) pp2 = clc(columns) + 1;
  substrcpy_terminate(columns,pp1,pp2-1,line,0);
  if ((clc(line)+1) > hicn) hicn = (clc(line)+1);
  ++nn;
}
IDD[0].nc = nn;
/* Allocate column name strings. */
IDD[0].names = (char **)calloc(IDD[0].nc,sizeof(char *));
for (ii=0; ii<IDD[0].nc; ++ii) {
  IDD[0].names[ii] = (char *)calloc(hicn,1);
}
/* Load column names. */
pp1=0; pp2=-1; nn=0;
while (pp2 < clc(columns)) {
  pp1 = pp2 + 1;
  pp2 = cindex2(columns,",",pp1);
  if (pp2 < 0) pp2 = clc(columns) + 1;
  substrcpy_terminate(columns,pp1,pp2-1,IDD[0].names[nn],0);
  ++nn;
}
/* Count lines in ipac table file and allocate .img */
ii=0; infu=fopen_read(IDD[0].file);
while (fgetline(line,infu) == 1) { ++ii; }
fclose(infu);
maxpt = ii * IDD[0].nc;
IDD[0].img = (double *)calloc(maxpt,sizeof(double));
/* Read IPAC data. */
if (verbose == 1) printf("Reading %d columns of double data from '%s'.\n",IDD[0].nc,IDD[0].file);
misc_read_IPAC_double( IDD[0].file, IDD[0].names, IDD[0].nc, maxpt, &ii, IDD[0].img, verbose );
IDD[0].nr = ii;
if (verbose == 1) printf("Found %d rows of data.\n",IDD[0].nr);
return;
}
/* -----------------------------------------------------------------------
  Free allocations in IDD.
*/
/*@@*/
void misc_free_IDD( IDDtype IDD )
{
int ii;
for (ii=0; ii<IDD.nc; ++ii) { free(IDD.names[ii]); IDD.names[ii]=NULL; }
free(IDD.names); IDD.names = NULL;
free(IDD.img); IDD.img = NULL;
return;
}
/* -----------------------------------------------------------------------
  Load data from a column in IFD structure.  -tab 08jan2012
  Input: IFD      : IPAC Float Data structure.
         column[] : column name.
 Output: aa[]     : allocate this to 'IFD.nr' size.
*/
/*@@*/
void misc_load_IFD( IFDtype IFD, char column[], float aa[] )
{
/**/
int ii,uc,pixno;
/**/
/* Find column number. */
uc=-1;
for (ii=0; ii<IFD.nc; ++ii) {
  if (strcmp(column,IFD.names[ii]) == 0) uc = ii;
}
if (uc < 0) { printf("***error:ml_IFD: could not find column '%s'.\n",column); exit(1); }
/* Load data. */
for (ii=0; ii<IFD.nr; ++ii) {
  pixno = ii + (IFD.nr * uc);   /* rows are columns, and columns are rows. */
  aa[ii] = IFD.img[pixno];
}
return;
}
/* -----------------------------------------------------------------------
  Read IPAC Float Data from IPAC table file. -tab 08jan2012
  Input: ipacfile[]  : filename of IPAC table float data.
         columns[]   : string of column names separated by columns.
 Output: *IFD        : IPAC Float Data structure.
*/
/*@@*/
void misc_read_IFD( char ipacfile[], char columns[], IFDtype *IFD, int verbose )
{
/**/
int maxpt,ii,nn,pp1,pp2,hicn;
char line[1000];
FILE *infu;
/**/
/* Store IPAC table filename. */
strcpy(IFD[0].file,ipacfile);
/* Count column names. */
pp1=0; pp2=-1; nn=0; hicn=1;
while (pp2 < clc(columns)) {
  pp1 = pp2 + 1;
  pp2 = cindex2(columns,",",pp1);
  if (pp2 < 0) pp2 = clc(columns) + 1;
  substrcpy_terminate(columns,pp1,pp2-1,line,0);
  if ((clc(line)+1) > hicn) hicn = (clc(line)+1);
  ++nn;
}
IFD[0].nc = nn;
/* Allocate column name strings. */
IFD[0].names = (char **)calloc(IFD[0].nc,sizeof(char *));
for (ii=0; ii<IFD[0].nc; ++ii) {
  IFD[0].names[ii] = (char *)calloc(hicn,1);
}
/* Load column names. */
pp1=0; pp2=-1; nn=0;
while (pp2 < clc(columns)) {
  pp1 = pp2 + 1;
  pp2 = cindex2(columns,",",pp1);
  if (pp2 < 0) pp2 = clc(columns) + 1;
  substrcpy_terminate(columns,pp1,pp2-1,IFD[0].names[nn],0);
  ++nn;
}
/* Count lines in ipac table file and allocate .img */
ii=0; infu=fopen_read(IFD[0].file);
while (fgetline(line,infu) == 1) { ++ii; }
fclose(infu);
maxpt = ii * IFD[0].nc;
IFD[0].img = (float *)calloc(maxpt,sizeof(float));
/* Read IPAC data. */
if (verbose == 1) printf("Reading %d columns of float data from '%s'.\n",IFD[0].nc,IFD[0].file);
misc_read_IPAC_float( IFD[0].file, IFD[0].names, IFD[0].nc, maxpt, &ii, IFD[0].img, verbose );
IFD[0].nr = ii;
if (verbose == 1) printf("Found %d rows of data.\n",IFD[0].nr);
return;
}
/* -----------------------------------------------------------------------
  Free allocations in IFD.
*/
/*@@*/
void misc_free_IFD( IFDtype IFD )
{
int ii;
for (ii=0; ii<IFD.nc; ++ii) { free(IFD.names[ii]); IFD.names[ii]=NULL; }
free(IFD.names); IFD.names = NULL;
free(IFD.img); IFD.img = NULL;
return;
}


/* -----------------------------------------------------------------------
  Gregorian calendar date to Julian date.
  The date is universal time at noon.
  So, for example, 245115.0 would be January 14, 2007 at 12:00 (UT).
*/
/*@@*/
int misc_gregorian_to_julian( int year, int month, int day )
{
int I,J,K,JD;
I= year;
J= month;
K= day;
JD= K-32075+1461*(I+4800+(J-14)/12)/4+367*(J-2-(J-14)/12*12)/12-3*((I+4900+(J-14)/12)/100)/4;
return(JD);
}

/* -----------------------------------------------------------------------
  Julian day to Gregorian date.
  The date is universal time at noon.
  So, for example, 245115.0 would be January 14, 2007 at 12:00 (UT).
*/
/*@@*/
void misc_julian_to_gregorian( int jd, int *year, int *month, int *day )
{
int I,J,K,L,N;
L= jd+68569;
N= 4*L/146097;
L= L-(146097*N+3)/4;
I= 4000*(L+1)/1461001;
L= L-1461*I/4+31;
J= 80*L/2447;
K= L-2447*J/80;
L= J/11;
J= J+2-12*L;
I= 100*(N-49)+I+L;
*year = I;
*month= J;
*day  = K;
return;
}

/* -----------------------------------------------------------------------
  Convert zulu time in the form: 2012-01-16T20:38:17.046 (for example)
                                 0123 56 89 12 45 78
  to a double precision julian day.  The seconds can be an integer
  or any number of decimal places.
*/
/*@@*/
double misc_zulu_to_julian( char zdate[] )
{
/**/
int year,month,day,hour,minute,jd;
double djd,seconds;
char wrd[200];
/**/
year   = cvalread0( zdate, 0, 3);
month  = cvalread0( zdate, 5, 6);
day    = cvalread0( zdate, 8, 9);
hour   = cvalread0( zdate,11,12);
minute = cvalread0( zdate,14,15);
substrcpy_terminate( zdate, 17, clc(zdate), wrd, 0);
seconds= GLV(wrd,1);
jd   = misc_gregorian_to_julian( year, month, day );
djd  = ((double)jd - 0.5) +
     ((((seconds / 60.) + (double)minute) / 60.) + (double)hour) / 24.;
return(djd);
}


/* -----------------------------------------------------------------------
  Convert double precision julian day to zulu time in the form: 
  2012-01-16T20:38:17.046 (for example)
  0123 56 89 12 45 78
*/
/*@@*/
void misc_julian_to_zulu( double djd, char zdate[] )
{
/**/
int year,month,day,hour,minute,jd;
double seconds,dhour,dminute;
/**/
jd        = djd + 0.5;
misc_julian_to_gregorian( jd, &year, &month, &day );
dhour     = 24. * ((djd + 0.5) - (double)jd);
hour      = dhour;
dminute   = 60. * (dhour - (double)hour);
minute    = dminute;
seconds   = 60. * (dminute - (double)minute);
sprintf(zdate,"%4.4d-%2.2d-%2.2dT%2.2d:%2.2d:%6.3f",
        year,month,day,hour,minute,seconds);
if (zdate[17] == ' ') zdate[17]='0';
return;
}


/* -----------------------------------------------------------------------
  Convert double precision julian day to a UT hours within that UT day.
*/
/*@@*/
double misc_julian_to_uth( double djd )
{
/**/
int ijd;
double uth;
/**/
ijd    = djd + 0.5;
uth = 24. * ((djd + 0.5) - (double)ijd);
return(uth);
}


/* -----------------------------------------------------------------------
  Convert zulu UT date/time to (approx.) decimal year.
  Assumes 365.25 days in one year.  
  zdate is in form: 2012-01-16T20:38:17.046 (for example)
*/
/*@@*/
double misc_zulu_to_decimal_year( char zdate[] )
{
/**/
int ii;
double jd0,jd,rr;
char wrd[100];
/**/
strcpy(wrd,zdate); ii=cindex(wrd,"-"); wrd[ii]='\0'; strcat(wrd,"-01-01T00:00:00.000");
jd0 = misc_zulu_to_julian( wrd );
jd  = misc_zulu_to_julian( zdate );
rr = cvalread0(zdate,0,3);
rr = rr + ( (jd - jd0) / 365.25 );
return(rr);
}


/* ----------------------------------------------------------------------
  Return the current GMT julian date/time.  -tab 18may2012
*/
/*@@*/
double current_julian()
{
/**/
char ts[100];
double jd;
/**/
timegetstring_gmt(ts);
jd = misc_zulu_to_julian( ts );
return(jd);
}


/* ----------------------------------------------------------------------
  Convert Year + Day-of-Year to a julian day. 
  ('doy' may have a fraction, returns JD with same fraction)
*/
/*@@*/
double misc_doy_to_julian( int year, double doy )
{
/**/
char wrd[100];
double jd,jd0;
/**/
sprintf(wrd,"%4d-01-01T00:00:00.000",year);
jd0 = misc_zulu_to_julian( wrd );
jd = jd0 + (double)(doy -1);
return(jd);
}


/* ----------------------------------------------------------------------
  Convert Year + Day-of-Year to calendar date.
  Returns date-string 'ds', e.g. '2013-03-20'
*/
/*@@*/
void misc_doy_to_calendar( int year, int doy, char ds[] )
{
/**/
char wrd[100];
double jd,jd0;
/**/
sprintf(wrd,"%4d-01-01T00:00:00.000",year);
jd0 = misc_zulu_to_julian( wrd );
jd = jd0 + (double)(doy -1);
misc_julian_to_zulu( jd, wrd );
substrcpy_terminate(wrd,0,9,ds,0);
return;
}


/* ----------------------------------------------------------------------
  Convert hours/degrees, minutes, seconds into decimal degrees.
    Input: hh  : hours or degrees
           mm  : minutes
           ss  : seconds
          mode : 1=ra, 0=dec.
  Returns decimal degrees.
*/
/*@@*/
double misc_hms2deci( int hh, int mm, double ss, int mode )
{
/**/
double dd,thesign;
/**/
thesign = 1.;
if (hh < 0) { thesign = -1.;  hh = -1 * hh; }
if (mm < 0) { thesign = -1.;  mm = -1 * mm; }
if (ss < 0) { thesign = -1.;  ss = -1.* ss; }
if ((thesign < 0.)&&(mode == 1)) {  fprintf(stderr,"***ERROR: No negative RAs.\n"); exit(1); }
dd = (double)hh + ((double)mm + (ss / 60.)) / 60.;
if (mode == 1) dd = dd * 15.;
dd = thesign * dd;
return(dd);
}


/* ----------------------------------------------------------------------
  Convert a string containing hours/degrees, minutes, seconds as 3 readable
  values, and convert to decimal degrees.  If any '-' is found, then it
  assumes it is a negative decimal degrees.
    Input: hms : string such as 'Telescope_HA=-02:22:18.15'
          mode : 1=ra, 0=dec.
  Returns decimal degrees.
*/
/*@@*/
double misc_hmss2deci( char hms[], int mode )
{
/**/
int hh,mm;
double ss,dd,thesign;
/**/
if (cindex(hms,"-") >= 0) { thesign = -1.; } else { thesign = 1.; }
hh=ABS((GLV(hms,1))); mm=ABS((GLV(hms,2))); ss=ABS((GLV(hms,3)));
dd = thesign * misc_hms2deci( hh, mm, ss, mode );
return(dd);
}


/* ----------------------------------------------------------------------
  Convert decimal degrees or hours into hours,minutes,seconds.
    Input: dcm : decimal degrees or hours.
   Output: hh  : hours or degrees
           mm  : minutes
           ss  : seconds
       thesign : +1. or -1.
*/
/*@@*/
void misc_deci2hms( double dcm, int *hh, int *mm, double *ss, double *thesign )
{
/**/
double dd;
/**/
if (dcm < 0.) { *thesign = -1.; } else { *thesign = 1.; }
dd  = *thesign * dcm;
*hh = (int)dd;
dd  = dd - (double)(*hh);
*mm = (int)(dd * 60.);
dd  = dd - ((double)(*mm)/60.);
*ss = dd * 3600.;
if (*ss > 59.9999999) { *ss = 0.; *mm = *mm + 1; }
return;
}


/* ----------------------------------------------------------------------
  Create coordinate name from ra and dec.
    Input: ra  : right ascension in decimal degrees.
           dec : declination in decimal degrees.
          mode : 1 = HHMM+DDD
                 2 = HHMM+DDDD
                 3 = HHMM.S+DDDD
                 4 = HH MM SS.S +DD MM SS
                     ----+----1----+----2----+
                 5 = HH MM SS.SS +DD MM SS.S
                 6 = HH MM SS.SSS +DD MM SS.SS
                 7 = HHMMS+DDMM
                 8 = FHHMMS+DDMM
                     ----+----1----+----2----+
   Output:  ss : string with coordinate name.
*/
/*@@*/
void misc_write_radec(double ra, double dec, int mode, char ss[])
{
/**/
double rs,ds,thesign;
int rh,rm,dd,dm,kk;
char cc;
char wrd[40];
/**/

/* Default and checks. */
strcpy(ss,"");
if ((ra  <   0.)||(ra  >360.)) return;
if ((dec < -90.)||(dec > 90.)) return;
misc_deci2hms( (ra / 15.), &rh, &rm, &rs, &thesign );
misc_deci2hms( (  dec   ), &dd, &dm, &ds, &thesign );
if (thesign > 0.) { cc = '+'; } else { cc = '-'; }
if (mode == 1) {
  dm = (((double)dm + (ds/60.))/6.);
  dm = MIN((9),MAX((0),(dm)));
  sprintf(ss,"%2.2d%2.2d%c%2.2d%1d",rh,rm,cc,dd,dm);
} else {
if (mode == 2) {
  sprintf(ss,"%2.2d%2.2d%c%2.2d%2.2d",rh,rm,cc,dd,dm);
} else {
if (mode == 3) {
  kk = rs / 6.;
  kk = MIN((9),MAX((0),(kk)));
  sprintf(ss,"%2.2d%2.2d%c%1d%c%2.2d%2.2d",rh,rm,'.',kk,cc,dd,dm);
} else {
if ((mode == 7)||(mode == 8)) {
  kk = rs / 6.;
  kk = MIN((9),MAX((0),(kk)));
  sprintf(ss,"%2.2d%2.2d%1d%c%2.2d%2.2d",rh,rm,kk,cc,dd,dm);
  if (mode == 8) { strcpy(wrd,"F"); strcat(wrd,ss); strcpy(ss,wrd); }
} else {
if (mode == 4) {
  rs = MIN((59.9),(rs));
  ds = MIN((59.0),(ds));
  kk = ds + 0.5;
  sprintf(ss,"%2.2d %2.2d %4.1f %c%2.2d %2.2d %2.2d",rh,rm,rs,cc,dd,dm,kk);
  if (ss[6] == ' ') ss[6] = '0';
} else {
if (mode == 5) {
  rs = MIN((59.99),(rs));
  ds = MIN((59.9) ,(ds));
  sprintf(ss,"%2.2d %2.2d %5.2f %c%2.2d %2.2d %4.1f",rh,rm,rs,cc,dd,dm,ds);
  if (ss[6] == ' ') ss[6] = '0';
  if (ss[20]== ' ') ss[20]= '0';
} else {
if (mode == 6) {
  rs = MIN((59.999),(rs));
  ds = MIN((59.99) ,(ds));
  sprintf(ss,"%2.2d %2.2d %6.3f %c%2.2d %2.2d %5.2f",rh,rm,rs,cc,dd,dm,ds);
  if (ss[6] == ' ') ss[6] = '0';
  if (ss[21]== ' ') ss[21]= '0';
}}}}}}}

return;
}


/* ---------------------------------------------------------------------- */
/* Convert epoch to Julian day.  */
/*@@*/
double misc_ast_julday(double epoch)
{
/**/
  int year,century,ii,jj;
  double jd;
/**/
  year    = epoch - 1;
  century = year/100;
  ii = (365 * year) - century + (year/4) + (century/4);
  jd = 1721425.5 + ii;
  jj = epoch;
  jd = jd + ( (epoch - jj) * 365.25 );
  return(jd);
}
/* ---------------------------------------------------------------------- */
/*
  Compute the precession rotation matrix from the standard epoch
  J2000.0 to the specified epoch.
*/
/*@@*/
void misc_ast_rotmatrix(double epoch, double p[3][3])
{
/**/
  double t, a, b, c, ca, cb, cc, sa, sb, sc;
/**/
/*
The rotation matrix coefficients are polynomials in time measured in Julian
centuries from the standard epoch.  The coefficients are in degrees.
*/
  t = ( misc_ast_julday(epoch) - 2451545.0 ) / 36525.0;
  a = t * (0.6406161 + ( t * (0.0000839 + (t * 0.0000050)) ));
  b = t * (0.6406161 + ( t * (0.0003041 + (t * 0.0000051)) ));
  c = t * (0.5567530 - ( t * (0.0001185 + (t * 0.0000116)) ));
/*
Compute the cosines and sines once for efficiency.
*/
  ca = cos(degtorad(a));
  sa = sin(degtorad(a));
  cb = cos(degtorad(b));
  sb = sin(degtorad(b));
  cc = cos(degtorad(c));
  sc = sin(degtorad(c));
/*
Compute the rotation matrix from the sines and cosines.
*/
  p[0][0] =  ( ca * cb * cc)      - (sa * sb);
  p[1][0] =  (-1. * sa * cb * cc) - (ca * sb);
  p[2][0] =  (-1. * cb * sc);
  p[0][1] =  ( ca * sb * cc)      + (sa * cb);
  p[1][1] =  (-1. * sa * sb * cc) + (ca * cb);
  p[2][1] =  (-1. * sb * sc);
  p[0][2] =  ( ca * sc);
  p[1][2] =  (-1. * sa * sc);
  p[2][2] =  ( cc);
  return;
}
/* ---------------------------------------------------------------------- */
/*
  Precess coordinates from epoch1 to epoch2.

  The method used here is based on the new IAU system described in the
  supplement to the 1984 Astronomical Almanac.  The precession is done in two
  steps; precess epoch1 to the standard epoch J2000.0 and then precess from the
  standard epoch to epoch2.  The precession between any two dates is done this
  way because the rotation matrix coefficients are given relative to the
  standard epoch.  RA and Dec are given in decimal degrees.
     Input: ra1,dec1,epoch1,epoch2.   Output: ora2,odec2.
*/
/*@@*/
void misc_ast_precess(double  ra1, double  dec1, double epoch1,
                  double *ora2, double *odec2, double epoch2)
{
/**/
  const double tiny = 1.e-30;     /* A very small number. */
  double  ra2, dec2;              /* Scratch second coordinates */
  double  r0[3], r1[3], p[3][3];
  int ok;
/**/
/*
If the input epoch is 0, assume the input epoch is the same as the output
epoch. If the two epochs are the same, return the coordinates from epoch1.
*/
  ok=1;
  if (epoch1 < 1.e-20) ok=0;
  if (ABS((epoch1-epoch2)) < 1.e-20) ok=0;
  if (ok == 0) {
    *ora2 = ra1;
    *odec2= dec1;
    return;
  }
/* Rectangular equatorial coordinates (direction cosines). */
  ra2  = degtorad(ra1);
  dec2 = degtorad(dec1);
  r0[0]= cos(ra2) * cos(dec2);
  r0[1]= sin(ra2) * cos(dec2);
  r0[2]= sin(dec2);
/* If epoch1 is not the standard epoch then precess to the standard epoch.*/
  if (ABS((epoch1-2000.)) > tiny) {
/* Compute rotation matrix. */
    misc_ast_rotmatrix(epoch1,p);
/* Note that we multiply by the inverse of p which is the transpose of p. */
    r1[0] = (p[0][0] * r0[0]) + (p[0][1] * r0[1]) + (p[0][2] * r0[2]);
    r1[1] = (p[1][0] * r0[0]) + (p[1][1] * r0[1]) + (p[1][2] * r0[2]);
    r1[2] = (p[2][0] * r0[0]) + (p[2][1] * r0[1]) + (p[2][2] * r0[2]);
    r0[0] = r1[0];
    r0[1] = r1[1];
    r0[2] = r1[2];
  }
/*
If epoch2 is not the standard epoch then precess from the standard
epoch to the desired epoch.
*/
  if (ABS((epoch2-2000.)) > tiny) {
/* Compute rotation matrix. */
    misc_ast_rotmatrix(epoch2,p);
/* Multiply by p. */
    r1[0] = (p[0][0] * r0[0]) + (p[1][0] * r0[1]) + (p[2][0] * r0[2]);
    r1[1] = (p[0][1] * r0[0]) + (p[1][1] * r0[1]) + (p[2][1] * r0[2]);
    r1[2] = (p[0][2] * r0[0]) + (p[1][2] * r0[1]) + (p[2][2] * r0[2]);
    r0[0] = r1[0];
    r0[1] = r1[1];
    r0[2] = r1[2];
  }
/* Convert from radians to degrees. */
  ra2 = radtodeg( atan2( r0[1], r0[0] ) );
  dec2= radtodeg( asin( r0[2] ) );
  if (ra2 < 0.) ra2 = ra2 + 360.;
/* Final result. */
  *ora2 = ra2;
  *odec2= dec2;
  return;
}


/* ----------------------------------------------------------------------
  Convert a julian date to a Local Sidereal Time (GMST) for
  a given longitude  (taken from 'ast_mst' routine).  -tab 21may2012
  Result is in decimal hours (0-24).
*/
/*@@*/
double local_sidereal( double jd, double longitude )
{
/**/
double ut,t,st,ast_mst,xx,yy;
/**/
/* Determine JD and UT, and T (JD in centuries from J2000.0). */
ut = (jd - (double)(( (int)jd )) - 0.5) * 24.0;
t  = (jd - 2451545.0) / 36525.0;
/* The GMST at 0 UT in seconds is a power series in T. */
st = 24110.54841 +  ( t * (8640184.812866 + t * (0.093104 - (t * 6.2e-6))) );
/* Correct for longitude and convert to standard hours. */
xx = (st / 3600.) + ut - (longitude / 15.);
yy = 24.; 
st = xx - ( yy * (double)(( (int)(xx/yy) ))  );
if (st < 0.) st = st + 24.;
ast_mst = st;
return(ast_mst);
}


/* ----------------------------------------------------------------------
  Return LST for # degree twilight (where '#' is given by 'twilight' parameter
  and would be '-12.0' for 12 degree twilight), given a julian date and longitude
  and latitude..  Evening is 'lst12a' and morning is 'lst12b'.  
  Result is in decimal degrees (0-360).  -tab 06feb2013
  Precision is ~10 seconds, but neglects elevation of site.
*/
/*@@*/
void misc_twilight_times( double twilight, double ujd, double longitude, double latitude, 
                          double *jd12a, double *jd12b, double *lst12a, double *lst12b )
{
/**/
double jd,alt,lst,ra,dec,ha,tzd,taz;
/**/

/* Evening # degree twilight. */
jd=ujd+0.002; alt=0.;
while (alt > twilight) {
  jd=jd+0.0001;
  lst = local_sidereal( jd, longitude ); lst = lst * 15.;     /* decimal degrees. */
  gal_Sun_RA_Dec( jd, &ra, &dec );       ra = ra * 15.;       /* decimal degrees. */
  ha = ra - lst;
  gal_true_altaz( ha, dec, latitude, &tzd, &taz );
  if (tzd < 0.) { alt = -90. - tzd; } else { alt = 90. - tzd; }
}
*jd12a  = jd;
*lst12a = local_sidereal( jd, longitude ) * 15.;

/* Morning 12 degree twilight. */
jd=jd+0.2; alt=-999.;
while (alt < twilight) {
  jd=jd+0.0001;
  lst = local_sidereal( jd, longitude );   lst = lst * 15.;     /* decimal degrees. */
  gal_Sun_RA_Dec( jd, &ra, &dec );  ra = ra * 15.;            /* decimal degrees. */
  ha = ra - lst;
  gal_true_altaz( ha, dec, latitude, &tzd, &taz );
  if (tzd < 0.) { alt = -90. - tzd; } else { alt = 90. - tzd; }
}
*jd12b  = jd;
*lst12b = local_sidereal( jd, longitude ) * 15.;

return;
}



/* ----------------------------------------------------------------------
  Evaluate polynomial: y = f(x) = co[0] + c[1]*x + c[2]*x*x + ...
    Input: nc (number of coeffecients)
    Input: co (coeffecient array)
    Input: xv (x value)
  Returns function value.
*/
/*@@*/
double cpolyval(int nc, double co[], double xv)
{
/**/
int ii;
double rr;
/**/
rr = 0.;
for (ii=0; ii < (nc-1); ++ii) { rr = ( rr + co[(nc-1)-ii] ) * xv; }
rr = rr + co[0];
return(rr);
}

/* ----------------------------------------------------------------------
   gaussj...  (for line fitting).
*/ 
/*@@*/
void GJ_gaussj( const int np, double a[np][np], int n, double b[] )
{
/**/
const int m = 1;
/**/
const int NMAX = 100;
/**/
int ipiv[NMAX],indxr[NMAX],indxc[NMAX];
/**/
double big,dum,pivinv;
/**/
int icol,irow,i,j,k,l,ll;
/**/
icol=0;
irow=0;
for (j=0; j<n; ++j) {
  ipiv[j]=0;
}
for (i=0; i<n; ++i) {
  big=0.;
  for (j=0; j<n; ++j) {
    if (ipiv[j] != 1) {
      for (k=0; k<n; ++k) {
        if (ipiv[k] == 0) {
          if (ABS((a[j][k])) >= big) {
            big = ABS((a[j][k]));
            irow = j;
            icol = k;
          }
        } else { if (ipiv[k] > 1) {
          fprintf(stderr,"***ERROR: GJ_gaussj: Singular matrix- 001.\n");
          exit(1);
        }}
      }
    }
  }
  ipiv[icol] = ipiv[icol] + 1;
  if (irow != icol) {
    for (l=0; l<n; ++l) {
      dum = a[irow][l];
      a[irow][l] = a[icol][l];
      a[icol][l] = dum;
    }
    for (l=0; l<m; ++l) {
      dum = b[irow];
      b[irow] = b[icol];
      b[icol] = dum;
    }
  }
  indxr[i] = irow;
  indxc[i] = icol;
  if (a[icol][icol] == 0.) {
    fprintf(stderr,"***ERROR: GJ_gaussj: Singular matrix- 002.\n");
    exit(1);
  }
  pivinv = 1. / a[icol][icol];
  a[icol][icol] = 1.;
  for (l=0; l<n; ++l) {
    a[icol][l] = a[icol][l] * pivinv;
  }
  for (l=0; l<m; ++l) {
    b[icol] = b[icol] * pivinv;
  }
  for (ll=0; ll<n; ++ll) {
    if (ll != icol) {
      dum = a[ll][icol];
      a[ll][icol] = 0.;
      for (l=0; l<n; ++l) {
        a[ll][l] = a[ll][l] - (a[icol][l] * dum);
      }
      for (l=0; l<m; ++l) {
        b[ll] = b[ll] - (b[icol] * dum);
      }
    }
  }
}
for (l=(n-1); l>=0; --l) {
  if (indxr[l] != indxc[l]) {
    for (k=0; k<n; ++k) {
      dum = a[k][(indxr[l])];
      a[k][(indxr[l])] = a[k][(indxc[l])];
      a[k][(indxc[l])] = dum;
    }
  }
}
return;
}
/* ----------------------------------------------------------------------
   covsrt...  (for line fitting).
*/ 
/*@@*/
void GJ_covsrt( const int ncvm, int ma, int mfit,
                double covar[ncvm][ncvm], int lista[] )
{
/**/
double swap;
int i,j;
/**/
for (j=0; j<(ma-1); ++j) {
  for (i=((j+1)-1); i<ma; ++i) {
    covar[i][j] = 0.;
  }
}
for (i=0; i<(mfit-1); ++i) {
  for (j=((i+1)-1); j<mfit; ++j) {
    if (lista[j] > lista[i]) {
      covar[lista[j]][lista[i]] = covar[i][j];
    } else {
      covar[lista[i]][lista[j]] = covar[i][j];
    }
  }
}
swap = covar[0][0];
for (j=0; j<ma; ++j) {
  covar[0][j] = covar[j][j];
  covar[j][j] = 0.;
}
covar[lista[0]][lista[0]] = swap;
for (j=1; j<mfit; ++j) {
  covar[lista[j]][lista[j]] = covar[0][j];
}
for (j=1; j<ma; ++j) {
  for (i=0; i<=(j-1); ++i) {
    covar[i][j] = covar[j][i];
  }
}
return;
}
/* ----------------------------------------------------------------------
   Fitting routine for a polynomial of degree np-1, with np coeffecients.
*/ 
/*@@*/
void GJ_polyfuncs( int np, double x, double p[] )
{
/**/
int j;
/**/
p[0] = 1.0;
for (j=1; j<np; ++j) {
  p[j] = p[(j-1)] * x;
}
return;
}
/* ----------------------------------------------------------------------
   GJ_lfit...  (Gauss-Jordan elimination line fitting)...
*/ 
/*@@*/
void GJ_lfit( int ndata, int ma, const int ncvm,
              double x[], double y[], double sig[],
              double a[], int lista[], int mfit,
              double covar[ncvm][ncvm], double *chisq )
{
/**/
int kk,j,ihit,k,i;
/**/
const int NMAX = 100;
/**/
double beta[NMAX],afunc[NMAX];
double ym,sig2i,wt,sum;
/**/
kk = mfit + 1;
for (j=0; j<ma; ++j) {
  ihit = 0;
  for (k=0; k<mfit; ++k) {
    if (lista[k] == j) { ihit = ihit + 1; }
  }
  if (ihit == 0) {
    lista[kk] = j;
    kk = kk + 1;
  } else { if (ihit > 1) {
    fprintf(stderr,"***ERROR: GJ_lfit:1:Improper set in LISTA.\n");
    exit(1);
  }}
}
if (kk != (ma+1)) {
  fprintf(stderr,"***ERROR: GJ_lfit:2:Improper set in LISTA.\n");
  exit(1);
}
for (j=0; j<mfit; ++j) {
  for (k=0; k<mfit; ++k) {
    covar[j][k] = 0.;
  }
  beta[j] = 0.;
}
for (i=0; i<ndata; ++i) {
  GJ_polyfuncs( ma, x[i], afunc );
  ym = y[i];
  if (mfit < ma) {
    for (j=((mfit+1)-1); j<ma; ++j) {
      ym = ym - ( a[lista[j]] * afunc[lista[j]] );
    }
  }
  sig2i = 1. / (sig[i] * sig[i]);
  for (j=0; j<mfit; ++j) {
    wt = afunc[lista[j]] * sig2i; 
    for (k=0; k<=j; ++k) {
      covar[j][k] = covar[j][k] + (wt * afunc[lista[k]]);
    }
    beta[j] = beta[j] + ( ym * wt );
  }
}
if (mfit > 0) {
  for (j=1; j<mfit; ++j) {
    for (k=0; k<=(j-1); ++k) {
      covar[k][j] = covar[j][k];
    }
  }
}
GJ_gaussj(ncvm,covar,mfit,beta);
for (j=0; j<mfit; ++j) {
  a[lista[j]] = beta[j];
}
*chisq = 0.;
for (i=0; i<ndata; ++i) {
  GJ_polyfuncs( ma, x[i], afunc );
  sum = 0.;
  for (j=0; j<ma; ++j) {
    sum = sum + (a[j] * afunc[j]);
  }
  *chisq = *chisq + (((y[i] - sum) / sig[i]) * ((y[i] - sum) / sig[i]));
}
GJ_covsrt( ncvm, ma, mfit, covar, lista );
return;
}
/* ----------------------------------------------------------------------
   GJ_polyfit_residuals  (Compute residuals after GJ polynomial fitting).

   Input: npt      : Number of points (npt must be > order+1).
          x[]      : x values.
          y[]      : y values.
          wt[]     : weight values.
          order    : Polynomial order (0 to 99).
          xoff     : X value offset. Subtract this value from x values
                        before using coeffecients.
          coef[]   : User polynomial coeffecients (from fit # >= order+1).
  Output: high     : high residual
          wrms     : weighted root-mean-square
          wppd     : weighted average point-to-point deviation
*/ 
/*@@*/
void GJ_polyfit_residuals( int npt, double x[], double y[], double wt[], 
                           int order, double xoff, double coef[], 
                           double *high, double *wrms, double *wppd )
{
/**/
int i;
double xv,yv,diff,sum1,sum2,wsum1,wsum2;
/**/
*high = 0.;
*wrms = 0.;
*wppd = 0.;
sum1  = 0.;
sum2  = 0.;
wsum1 = 0.;
wsum2 = 0.;
for (i=0; i<npt; ++i) { if (wt[i] > 0.) {
  xv = x[i] - xoff;
  yv = cpolyval( order+1, coef, xv );
  diff = yv - y[i];
  if (ABS((diff)) > ABS((*high))) *high = diff;
  sum1 = sum1 + (wt[i] * diff * diff);
  wsum1= wsum1+ wt[i];
  if (i > 0) {
    xv = x[(i-1)] - xoff;
    yv = cpolyval( order+1, coef, xv );
    sum2 = sum2 + ( wt[i] * wt[i] * ABS((diff - (yv - y[(i-1)]))) );
    wsum2= wsum2+ ( wt[i] * wt[i] );
  }
}}
if (wsum1 > 0.) *wrms = sqrt((sum1/wsum1));
if (wsum2 > 0.) *wppd = sqrt((sum2/wsum2));
return;
}

/* ----------------------------------------------------------------------
   GJ_polyfit...  (Polynomial fitting with Gauss-Jordan elimination)...

   Input: npt      : Number of points (npt must be > order+1).
          x[]      : x values.
          y[]      : y values.
          wt[]     : weight values (usually 1/sig^2).
                       (Entries with wt=0 are excluded from the data set.)
          order    : Polynomial order (0 to 99).
          forcexoff: Force routine to use the given "xoff".
                       0 = no force, xoff determined by this routine.
                       1 = force, use user xoff value for offset.
  In/Out: xoff     : X value offset. Subtract this value from x values
                        before using coeffecients.
  Output: ucoef[]  : User polynomial coeffecients (from fit # >= order+1).

    Returns "1" if successful, otherwise returns "0".
*/ 
/*@@*/
int GJ_polyfit( int npt, double x[], double y[], double wt[], int order,
                int forcexoff, double *xoff, double ucoef[] )
{
/**/
const int np = 100;
/**/
double chisq,yoff,yscale,xscale,rr,wsum;
double a[np],covar[np][np],coef[np];
double *xx,*yy,*sig;
/**/
int ncoef,ndata,mp,i,ma,mfit,ok,npt_nzw;
int lista[np];
/**/

/* Check order. */
if ((order > 99)||(order < 0)) {
  fprintf(stderr,"***ERROR:GJ_polyfit:Order(%d) out of range (0-99).\n",order);
  return(0);
}

/* Allocate. */
mp = npt + 1000;
xx = (double *)calloc(mp,sizeof(double));
yy = (double *)calloc(mp,sizeof(double));
sig= (double *)calloc(mp,sizeof(double));

/* Check size of data set. */
if (npt > mp) {
  fprintf(stderr,"***ERROR:GJ_polyfit: Too many points: %d. Max: %d.\n",npt,mp);
  free(xx); free(yy); free(sig);
  return(0);
}
/* Check weights, count number of non-zero weight points. */
ok=1;
npt_nzw=0;
for (i=0; i<npt; ++i) { 
  if (wt[i] < 0.) ok=0;
  if (wt[i] > 0.) ++npt_nzw;
}
if (ok == 0) {
  fprintf(stderr,"***ERROR:GJ_polyfit: Weights must all be >= 0.\n");
  free(xx); free(yy); free(sig);
  return(0);
}
/* Check polynomial order vs. number of non-zero weight points. */
if (npt_nzw <= order+1) {
  fprintf(stderr,
    "***ERROR:GJ_polyfit:Not enough (>0 weight) points(%d) for order(%d).\n",
    npt_nzw,order);
  free(xx); free(yy); free(sig);
  return(0);
}
/* Zero coeffecients, set lista[]. */
ncoef = order + 1;
for (i=0; i<np; ++i) {
  coef[i] = 0.;
  lista[i]= i;
}
/* Determine xoff. */
if (forcexoff == 0) {
  *xoff= 0.;
  wsum = 0.;
  for (i=0; i<npt; ++i) {
    if (wt[i] > 0.) {
      *xoff= *xoff + (wt[i] * x[i]);
      wsum = wsum + wt[i];
    }
  }
  *xoff = *xoff / wsum;    /* wsum must be >0 (see above). */
}
/* Set xscale and yoff. */
xscale = 1.;
yscale = 1.;
yoff   = 0.;
/* Load data arrays. */
ndata = 0;
for (i=0; i<npt; ++i) {
  if (wt[i] > 0.) {
    xx[ndata] = (x[i] - *xoff) / xscale;
    yy[ndata] = (y[i] - yoff) / yscale;
    sig[ndata]= 1.0 / sqrt(wt[i]);
    ++ndata;
  }
}
/* Number of parameters. */
ma  = ncoef;
mfit= ncoef;

/* Call least squares fitting routine (Gauss-Jordan elimination). */
GJ_lfit( ndata, ma, np, xx, yy, sig, a, lista, mfit, covar, &chisq );

/* Load coeffecients, adjusting for y scaling and offset. */
for (i=0; i<ncoef; ++i) {
  coef[i] = (a[i] * yscale);
}
coef[0] = coef[0] + yoff;
/* Adjust for x scaling. */
rr = 1.;
for (i=1; i<ncoef; ++i) {
  rr = rr / xscale;
  coef[i] = rr * coef[i];
}
/* Transfer coeffecients to user array. */
for (i=0; i<ncoef; ++i) {
  ucoef[i] = coef[i];
}
/* Success. */
free(xx); free(yy); free(sig);
return(1);
}


/* ----------------------------------------------------------------------
 Smooth an image using a boxcar.
 Input:   nc,nr  : number of columns and rows.
          img[]  : image to be smoothed (changed on output).
          ref[]  : scratch image of same size to be used as reference.
          colrad : radius of box (-colrad..0..colrad).  1,1 is a 3x3 box.
          rowrad : radius of box (-rowrad..0..rowrad).
 (Warning: the edge pixels of colrad and rowrad are ignored (for speed).)
*/
/*@@*/
void misc_Smooth_Image( int nc, int nr, float img[], float ref[], int colrad, int rowrad )
{
/**/
int pixno,ii,jj,iii,jjj,ii1,ii2,jj1,jj2;
float sum,size;
/**/
/* Copy. */
for (ii=0; ii<(nc*nr); ++ii) { ref[ii] = img[ii]; }
/* Smooth. */
size = (float)((colrad + colrad + 1) * (rowrad + rowrad + 1));
for (ii=colrad; ii<(nc-colrad); ++ii) {
  ii1=ii-colrad; ii2=ii+colrad;
  for (jj=rowrad; jj<(nr-rowrad); ++jj) {
    jj1=jj-rowrad; jj2=jj+rowrad;
    sum=0.;
    for (iii=ii1; iii<=ii2; ++iii) {
    for (jjj=jj1; jjj<=jj2; ++jjj) {
      pixno = iii + (jjj * nc);
      sum = sum + ref[pixno];
    }}
    pixno = ii + (jj * nc);
    img[pixno] = sum / size;
  }
}
return;
}


